local BlockingInfo = {}
local lg = love.graphics

-- coarse grid to check if there can potentially be a path at all (pre-check)
-- fine grid to determine a specific detailed path

-- 0 = free, no blocking
-- 1 = blocked (wall)
-- 2 = blocked (water)
-- 3 = blocked (lava)
-- 4 = blocked (building, fine grid at least)

function BlockingInfo:create()
	local blocking = {}
	
	-- coarse grid: one entry per tile
	blocking.coarse = {
		width = 0,
		height = 0,
		data = {},
	}
	
	-- fine grid: one tile = 5x5 entries
	blocking.fine = {
		width = 0,
		height = 0,
		data = {},
	}
	
	----------------------------------------------------------------------------------------------------
	
	-- size in tiles
	function blocking:setSize( width, height )
		
		local coarse = self.coarse
		local fine = self.fine
		
		coarse.width = width
		coarse.height = height
		
		fine.width = width * 5
		fine.height = height * 5
		
		-- prepare the blocking info: all empty
		for y = 1, coarse.height do
			coarse.data[y] = {}
			for x = 1, coarse.width do
				coarse.data[y][x] = 0
			end
		end
		
		for y = 1, fine.height do
			fine.data[y] = {}
			for x = 1, fine.width do
				fine.data[y][x] = 0
			end
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- dimensions must match!
	-- terrain = object
	-- wallData = simple 2d table
	function blocking:fullBlockingUpdate( terrain, wallData )
		
		local coarse = self.coarse
		local fine = self.fine
		
		for yi = 1, coarse.height do
			for xi = 1, coarse.width do
				self:_singleTileUpdate( xi, yi, terrain, wallData )
			end
		end
		
		self:updatePathfinding()
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- update a single full sized tile with tile index
	-- see fullBlockingUpdate
	function blocking:singleTileUpdate( xi, yi, terrain, wallData )
		
		self:_singleTileUpdate( xi, yi, terrain, wallData )
		self:updatePathfinding()
		
	end
	
	-- internal only, do not call updatePathfinding
	function blocking:_singleTileUpdate( xi, yi, terrain, wallData )
		
		local coarse = self.coarse
		local fine = self.fine
		
		local function updateFine( xi, yi, t )
			local x2 = xi*5
			local y2 = yi*5
			for y = y2-4, y2 do
				for x = x2-4, x2 do
					fine.data[y][x] = t
				end
			end
		end
		
		local t = terrain:getTerrainTypeAtTile( xi, yi )
		local w = wallData[yi][xi]
		-- any wall type?
		if w > 0 then
			coarse.data[yi][xi] = 1	-- wall (type does not matter here)
			updateFine( xi, yi, 1 )
		else
			-- check terrain
			if t == 1 then	-- dirt
				coarse.data[yi][xi] = 0
				updateFine( xi, yi, 0 )
			else
				coarse.data[yi][xi] = t	-- water (2) or lava (3)
				updateFine( xi, yi, t )
			end
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- add the blocking info for a new building
	-- only updates the fine grid
	-- for now each building can only occupy one full tile!
	function blocking:addEntityBlockingAtTile( xi, yi, blockingInfo )
		
		local fine = self.fine
		
		-- tile index to first mini tile index
		local x = (xi-1) * 5 + 1
		local y = (yi-1) * 5 + 1
		
		-- _x and _y = offset inside blockingInfo
		for _y = 0, 4 do
			for _x = 0, 4 do
				if blockingInfo[_y+1][_x+1] == 1 then
					fine.data[y+_y][x+_x] = 4	-- blocked by a building
					-- print(x+_x,y+_y)
				end
			end
		end
		
		self:updatePathfinding()
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- remove the blocking info for a building
	-- only updates the fine grid
	-- for now each building can only occupy one full tile!
	function blocking:removeEntityBlockingAtTile( xi, yi, blockingInfo )
		
		local fine = self.fine
		
		-- tile index to first mini tile index
		local x = (xi-1) * 5 + 1
		local y = (yi-1) * 5 + 1
		
		-- _x and _y = offset inside blockingInfo
		for _y = 0, 4 do
			for _x = 0, 4 do
				if blockingInfo[_y+1][_x+1] == 1 then
					fine.data[y+_y][x+_x] = 0
				end
			end
		end
		
		self:updatePathfinding()
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- check the coarse grid only
	function blocking:isTileBlocked( xi, yi )
		if not self.coarse.data[yi] or not self.coarse.data[yi][xi] then
			return false
		end
		return self.coarse.data[yi][xi] > 0
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- fine grid tile index! not the normal tile index
	function blocking:isMiniTileBlocked( xi, yi )
		if not self.fine.data[yi] or not self.fine.data[yi][xi] then
			return false
		end
		return self.fine.data[yi][xi] > 0
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- fine grid tile index! not the normal tile index
	function blocking:isMiniTileFlightBlocked( xi, yi )
		if not self.fine.data[yi] or not self.fine.data[yi][xi] then
			return false
		end
		local b = self.fine.data[yi][xi]
		return b == 1 or b == 4
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- debug only
	function blocking:draw()
		local grid = self.fine
		
		local tilesize = 25	-- full tile = 125
		
		if grid.width < 1
		or grid.height < 1 then
			return
		end
		
		for yi = 1, grid.height do
			for xi = 1, grid.width do
				-- note: the additional -2 for the index is because the normal tiles have their origin at the center
				local x = (xi-3) * tilesize
				local y = (yi-3) * tilesize
				x, y = convertToDimetric( x, y )
				
				if self:isMiniTileFlightBlocked( xi, yi ) then
					lg.setColor( 1, 0, 0, 0.5 )
				else
					lg.setColor( 0, 0.8, 0.4, 0.5 )
				end
				lg.circle( "fill", x, y, 4 )
				-- lg.print( xi..":"..yi, x, y )
			end
		end
		lg.setColor( 1, 1, 1 )
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- 2x2 different maps:
	-- coarse and fine
	-- for ground units and flying units
	function blocking:updatePathfinding()
		
		local coarse = self.coarse
		local fine = self.fine
		
		-- maps
		local ground_coarse = { width = coarse.width, height = coarse.height, blockingData = {}, }
		local ground_fine = { width = fine.width, height = fine.height, blockingData = {}, }
		local flying_coarse = { width = coarse.width, height = coarse.height, blockingData = {}, }
		local flying_fine = { width = fine.width, height = fine.height, blockingData = {}, }
		
		-- update the blocking data (boolean entries per tile where true = traversable)
		
		-- ground_coarse
		for yi = 1, ground_coarse.height do
			ground_coarse.blockingData[yi] = {}
			for xi = 1, ground_coarse.width do
				ground_coarse.blockingData[yi][xi] = coarse.data[yi][xi] == 0	-- not blocked
			end
		end
		
		-- ground_fine
		for yi = 1, ground_fine.height do
			ground_fine.blockingData[yi] = {}
			for xi = 1, ground_fine.width do
				ground_fine.blockingData[yi][xi] = fine.data[yi][xi] == 0	-- not blocked
			end
		end
		
		-- flying_coarse
		for yi = 1, flying_coarse.height do
			flying_coarse.blockingData[yi] = {}
			for xi = 1, flying_coarse.width do
				flying_coarse.blockingData[yi][xi] = coarse.data[yi][xi] ~= 1
			end
		end
		
		-- flying_fine
		for yi = 1, flying_fine.height do
			flying_fine.blockingData[yi] = {}
			for xi = 1, flying_fine.width do
				flying_fine.blockingData[yi][xi] = not self:isMiniTileFlightBlocked( xi, yi )
			end
		end
		
		-- pass info to A* algo
		Pathfinding.setMap( "ground_coarse", ground_coarse )
		Pathfinding.setMap( "ground_fine", ground_fine )
		Pathfinding.setMap( "flying_coarse", flying_coarse )
		Pathfinding.setMap( "flying_fine", flying_fine )
		
		-- also reset stored paths
		Pathfinding.clearPaths( "ground_coarse" )
		Pathfinding.clearPaths( "ground_fine" )
		Pathfinding.clearPaths( "flying_coarse" )
		Pathfinding.clearPaths( "flying_fine" )
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- debug only!
	-- Renderer:add( blocking, "terrain" )
	
	return blocking
end

return BlockingInfo
