local ConstructionSite = {}
local lg = love.graphics


function ConstructionSite:create( data )
	local entity = Building:create( data )
	
	entity.entityType = ENTITIES.constructionSite
	
	entity.isConstructionComplete = true
	
	-- the entity type of the building that will be created at this site
	entity.buildingType = data.buildingType
	
	-- range in full tiles
	entity.explorationRange = 2
	
	entity.image = Images["construction_site_2"]
	entity.rendererLayer = "ground_entities"
	
	-- blocking info in mini tiles
	-- for now each building can only occupy one full tile
	entity.blocking = {
		{0,0,0,0,0},
		{0,1,1,1,0},
		{0,1,1,1,0},
		{0,1,1,1,0},
		{0,0,0,0,0},
	}
	
	entity.numberOfPlacedConstructionFences = 0
	
	entity.missingResources = Tools.copyTable( data.costs )
	
	-- see getStoredResources
	entity.storedResources = {}
	
	----------------------------------------------------------------------------------------------------
	
	-- there are 4 positions ...
	function entity:getFencePosition( index )
		local x, y = self:getPosition()
		if index == 1 then
			return x, y + 50
		elseif index == 2 then
			return x - 50, y
		elseif index == 3 then
			return x, y - 50
		elseif index == 4 then
			return x + 50, y
		end
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- a fence has been placed
	function entity:onFencePlaced()
		self.numberOfPlacedConstructionFences = self.numberOfPlacedConstructionFences + 1
	end
	
	
	function entity:areAllFencesPlaced()
		return self.numberOfPlacedConstructionFences == 4
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:getMissingResources()
		return self.missingResources
	end
	
	
	function entity:isMissingResource( resourceType )
		return self.missingResources[resourceType] and self.missingResources[resourceType] > 0
	end
	
	
	function entity:addResourceProgress( resourceType, amount )
		local amount = amount or 1
		if not self:isMissingResource( resourceType ) then
			return false
		end
		self.missingResources[resourceType] = self.missingResources[resourceType] - amount
		-- print( "construction site - addResourceProgress, missing", self.missingResources[resourceType] )
		-- print( "construction site id:", self.id )
		self.storedResources[resourceType] = (self.storedResources[resourceType] or 0) + amount
		
		self.ohw:updateMissingResources( self:getMissingResources() )
		
		-- TODO: all resources brought here?
		local ready = true
		for k, v in pairs ( self:getMissingResources() ) do
			if v > 0 then
				ready = false
				break
			end
		end
		if ready then
			-- remove self and the fences + create the building (with 0 construction progress)
			Game:replaceConstructionSite( self, self.buildingType )
		end
	end
	
	
	-- makes it easier to create needed resourc entities upon destroying the site
	function entity:getStoredResources()
		return self.storedResources
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:load( id )
		self.id = id
		
		Gameloop:add( self )
		Renderer:add( self, self.rendererLayer )
		
		local body = self.body
		self.ohw = OhwConstructionSite:create{
			x = body.x,
			y = body.y,
			resources = self:getMissingResources(),
			icon = getIconByEntityType( self.buildingType ),
		}
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:tick( dt )
		
		if self.selected then
			self.animationTick = self.animationTick + dt
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:draw()
		local body = self.body
		local x, y = convertToDimetric( body.x, body.y )
		
		local imgData = self.image
		local ox = imgData.ox or 0
		local oy = imgData.oy or 0
		
		if self.selected then
			-- a simple but decent looking "animation"; value goes from 0.6 to 1.0
			local v = 0.2 * math.sin( 3 * self.animationTick ) + 0.8
			-- lg.setColor( 0.8, 0.8, v )
			lg.setColor( v, v, 1 )
		end
		lg.draw( imgData.image, x, y, 0, 1, 1, ox, oy )
		lg.setColor( 1, 1, 1 )
		
		
		self.ohw:setVisible( self.selected == true )
		--[[
		if self.selected then
			local missingResources = self:getMissingResources()
			if missingResources[ENTITIES.ore] and missingResources[ENTITIES.ore] > 0 then
				local img = Images["icon_collect_ore"]
				local scale = 32 / img.image:getWidth()
				lg.draw( img.image, x, y, 0, scale, scale, img.ox * scale, img.oy * scale )
				lg.print( string.format( "%dx", missingResources[ENTITIES.ore] ), x, y )
			end
		end
		--]]
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:destroy()
		-- remove fences
		local fences = EntityManager:getEntitiesFilter( function(e)
			return e.entityType == ENTITIES.constructionFence
			and e.constructionSiteId == self.id
		end )
		for i, fence in pairs( fences ) do
			-- fence:destroy()
			EntityManager:remove( fence )
		end
		
		Gameloop:remove( self )
		Renderer:remove( self, self.rendererLayer )
		self.ohw:destroy()
		self = nil
	end
	
	
	return entity
end


return ConstructionSite
