local SupportStation = {}
local lg = love.graphics


function SupportStation:create( data )
	local entity = Building:create( data )
	
	entity.entityType = ENTITIES.supportStation
	
	-- also store the tile index (avoid to calculate them over and over again)
	entity.xi = data.xi
	entity.yi = data.yi
	
	-- range in full tiles
	entity.explorationRange = 4
	
	-- entity.image = Images["support_station_1"]
	entity.rendererLayer = "entities"
	
	-- animations
	entity.animations = {}
	entity.animations.normal = Animation.new( "animations/support station/normal.lua" )
	entity.animations.normal:setRelativeOrigin( 0.5, 0.75 )
	
	-- set initial animation set
	entity.currentAnimation = "normal"
	
	-- selection frame image
	entity.imageWidth = 180
	entity.imageHeight = 140
	entity.outlineOffsetY = 28
	local frameImageData = Images["entity_selection_frame_1"]
	entity.selectionFrameImage = frameImageData
	entity.selectionFrameScaleX = entity.imageWidth / frameImageData.image:getWidth()
	entity.selectionFrameScaleY = entity.imageHeight / frameImageData.image:getHeight()
	
	-- blocking info in mini tiles
	-- for now each building can only occupy one full tile
	entity.blocking = {
		{0,0,0,0,0},
		{0,1,1,1,0},
		{0,1,1,0,0},
		{0,1,1,1,0},
		{0,0,0,0,0},
	}
	
	----------------------------------------------------------------------------------------------------
	
	function entity:load( id )
		self.id = id
		
		Gameloop:add( self )
		Renderer:add( self, self.rendererLayer )
		
		-- for buildings the position will not be changed
		local body = self.body
		local x, y = convertToDimetric( body.x, body.y )
		self.animations[self.currentAnimation]:setPosition( x, y )
		
		self.ohwSelection = OhwEntitySelection:create{
			entityId = self.id,
			width = self.imageWidth,
			height = self.imageHeight,
			offsetY = self.outlineOffsetY,
		}
		
		Game:onBuildingCreated( self )
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- set currently selected - true / false
	function entity:setSelected( selected )
		self.selected = selected
		self.animationTick = 0
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:tick( dt )
		
		if self.selected then
			self.animationTick = self.animationTick + dt
		end
		
		self.timeSinceLastHealthChange = self.timeSinceLastHealthChange + dt
		
		
		-- animation
		if self.currentAnimation and self.animations[self.currentAnimation] then
			local anim = self.animations[self.currentAnimation]
			anim:update( dt )
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:draw()
		local body = self.body
		local x, y = convertToDimetric( body.x, body.y )
		
		-- animation
		if self.currentAnimation and self.animations[self.currentAnimation] then
			local anim = self.animations[self.currentAnimation]
			-- anim:setPosition( x, y )
			-- anim:draw( 0.8, 0.8 )
			anim:draw()
		end
		
		-- draw "light"
		-- TODO: do not draw when the building is not operating at the moment
		if true then
			local liSize = 800
			local alpha = 0.8
			local color = { 1, 0.7, 0.7, alpha }
			Renderer:drawCircularLight( x, y, liSize, color )
		end
		
		self.ohwSelection:setVisible( self.selected == true )
		
		-- healthbar
		-- should be about 4 to 6 times as wide as high and above the entity
		-- the width MIGHT be larger for entities with more health (?)
		if Game.drawEntityStatusOverhead or self.selected
		or self.timeSinceLastHealthChange < 4 then
			local hb_y_offset = body.z + 100	-- draw above the entity
			self.health:draw( x, y - hb_y_offset, 50, 8 )
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:destroy()
		Gameloop:remove( self )
		Renderer:remove( self, self.rendererLayer )
		self.ohwSelection:destroy()
		self = nil
	end
	
	
	return entity
end


return SupportStation
