local TeleportPad = {}
local lg = love.graphics

-- can only teleport a unit every seconds
local teleportUnitCooldown = 2


function TeleportPad:create( data )
	local entity = Building:create( data )
	
	entity.entityType = ENTITIES.teleportPad
	-- entity.entityCategory = ENTITYCATEGORY.building
	
	-- also store the tile index (avoid to calculate them over and over again)
	entity.xi = data.xi
	entity.yi = data.yi
	
	-- range in full tiles
	entity.explorationRange = 4
	
	entity.image = Images["teleport_pad_1"]
	entity.rendererLayer = "entities"
	
	-- selection frame image
	entity.imageWidth = 120
	entity.imageHeight = 120
	entity.outlineOffsetY = 24
	local frameImageData = Images["entity_selection_frame_1"]
	entity.selectionFrameImage = frameImageData
	entity.selectionFrameScaleX = entity.imageWidth / frameImageData.image:getWidth()
	entity.selectionFrameScaleY = entity.imageHeight / frameImageData.image:getHeight()
	
	-- blocking info in mini tiles
	-- for now each building can only occupy one full tile
	entity.blocking = {
		{0,0,0,0,0},
		{0,1,1,1,0},
		{0,1,1,1,0},
		{0,1,1,1,0},
		{0,0,0,0,0},
	}
	
	entity.teleportUnitCooldown = 0
	
	----------------------------------------------------------------------------------------------------
	
	-- entity (Rock Raider) spawn location
	-- and position where they move to after being teleported down
	
	-- important: position is inside the building blocking
	function entity:getTeleportUnitPosition()
		local x, y = self:getPosition()
		return x + 10, y + 1
	end
	
	-- important: position must not be blocked!
	function entity:getLeaveBuildingPosition()
		local x, y = self:getPosition()
		return x + 50, y
	end
	
	-- called when the player tries to teleport a unit here (button callback)
	function entity:isReadyToTeleportUnit()
		return self.teleportUnitCooldown == 0
	end
	
	function entity:teleportUnit()
		self.teleportUnitCooldown = teleportUnitCooldown
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:getBlocking()
		return self.blocking
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- range in full tiles
	function entity:getExplorationRange()
		return self.explorationRange
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:load( id )
		self.id = id
		
		Gameloop:add( self )
		Renderer:add( self, self.rendererLayer )
		
		self.ohwSelection = OhwEntitySelection:create{
			entityId = self.id,
			width = self.imageWidth,
			height = self.imageHeight,
			offsetY = self.outlineOffsetY,
		}
		
		Game:onBuildingCreated( self )
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- set currently selected - true / false
	function entity:setSelected( selected )
		self.selected = selected
		self.animationTick = 0
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:tick( dt )
		
		if self.selected then
			self.animationTick = self.animationTick + dt
		end
		
		self.timeSinceLastHealthChange = self.timeSinceLastHealthChange + dt
		
		
		if self.teleportUnitCooldown > 0 then
			self.teleportUnitCooldown = Tools.max( 0, self.teleportUnitCooldown - dt )
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:draw()
		local body = self.body
		local x, y = convertToDimetric( body.x, body.y )
		local imgData = self.image
		local ox = imgData.ox or 0
		local oy = imgData.oy or 0
		
		lg.draw( imgData.image, x, y, 0, 1, 1, ox, oy )
		
		-- draw "light"
		-- TODO: do not draw when the building is not operating at the moment
		if true then
			local liSize = 800
			local alpha = 0.8
			local color = { 1, 0.7, 0.7, alpha }
			Renderer:drawCircularLight( x, y, liSize, color )
		end
		
		self.ohwSelection:setVisible( self.selected == true )
		
		-- healthbar
		if self.selected or self.timeSinceLastHealthChange < 4 then
			local hb_y_offset = body.z + 100	-- draw above the entity
			self.health:draw( x, y - hb_y_offset, 50, 8 )
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:destroy()
		Gameloop:remove( self )
		Renderer:remove( self, self.rendererLayer )
		self.ohwSelection:destroy()
		self = nil
	end
	
	
	return entity
end

return TeleportPad
