local TeslaTower = {}
local lg = love.graphics
local random = math.random


function TeslaTower:create( data )
	local entity = Building:create( data )
	
	entity.entityType = ENTITIES.teslaTower
	entity.isDefendableBuilding = true
	
	-- range in full tiles
	entity.explorationRange = 4
	
	entity.image = Images["tesla_tower_1"]
	entity.rendererLayer = "entities"
	
	entity.maxNumberOfCharges = 10	-- can be increased with upgrades (?)
	entity.remainingNumberOfCharges = 0
	entity.attackDamage = 50				-- damage per charge (bolt)
	entity.attackRange = 350				-- in world coordinates
	entity.totalAttackWait = 0.2		-- time between two attacks in seconds
	entity.currentAttackWait = 0		-- time already waited after an attack
	entity.totalChargeTime = 3			-- it takes X seconds to fully charge up
	entity.currentChargeTime = 0		-- charge up to totalChargeTime
	entity.timeToNextEnemyScan = 0
	entity.timeBetweenScans = 0.5		-- check if attackable enemies are nearby
	entity.totalStayAlertTime = 4		-- when charged, stay up to X seconds ready for attack before going back to idle
	entity.currentStayAlertTime = 0
	
	-- entity state "machine"
	-- idle = do nothing except to scan for attackable enemies in range every timeBetweenScans
	--> if an enemy is nearby, go to charging state
	-- charging = charge until fully charged --> go to charged state
	-- charged = ready to fire
	--> if there are no enemies left, wait for X seconds and eventually go to idle state
	--> at least one enemy there, go to attack state
	-- attack = fire at a nearby enemy as long as charges are available
	--> no charges left (remainingNumberOfCharges), go to idle state
	entity.state = "idle"
	
	-- selection frame image
	entity.imageWidth = 90
	entity.imageHeight = 90
	entity.outlineOffsetY = 24
	local frameImageData = Images["entity_selection_frame_1"]
	entity.selectionFrameImage = frameImageData
	entity.selectionFrameScaleX = entity.imageWidth / frameImageData.image:getWidth()
	entity.selectionFrameScaleY = entity.imageHeight / frameImageData.image:getHeight()
	
	-- blocking info in mini tiles
	entity.blocking = {
		{0,0,0,0,0},
		{0,1,1,1,0},
		{0,1,1,1,0},
		{0,1,1,1,0},
		{0,0,0,0,0},
	}
	
	----------------------------------------------------------------------------------------------------
	
	function entity:load( id )
		self.id = id
		
		Gameloop:add( self )
		Renderer:add( self, self.rendererLayer )
		
		self.ohwSelection = OhwEntitySelection:create{
			entityId = self.id,
			width = self.imageWidth,
			height = self.imageHeight,
			offsetY = self.outlineOffsetY,
		}
		
		Game:onBuildingCreated( self )
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:tick( dt )
		
		-- behavior depending on the internal state
		self:behaviorTick( dt )
		
		
		if self.selected then
			self.animationTick = self.animationTick + dt
		end
		
		self.timeSinceLastHealthChange = self.timeSinceLastHealthChange + dt
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- TODO: only scan/target/attack monsters that are not stealthed (e.g. underground)
	
	function entity:behaviorTick( dt )
		
		local ex, ey = self:getPosition()
		
		if self.state == "idle" then
			
			self.timeToNextEnemyScan = self.timeToNextEnemyScan - dt
			if self.timeToNextEnemyScan <= 0 then
				self.timeToNextEnemyScan = self.timeToNextEnemyScan + self.timeBetweenScans
				
				local enemies = self:getEnemiesInAttackRange()
				
				if #enemies > 0 then
					self:switchToState( "charging" )
				end
			end
			
			
		elseif self.state == "charging" then
			
			self.currentChargeTime = self.currentChargeTime + dt
			if self.currentChargeTime >= self.totalChargeTime then
				self.currentChargeTime = 0
				self:switchToState( "charged" )
			end
			
			
		elseif self.state == "charged" then
			
			local enemies = self:getEnemiesInAttackRange()
			
			if #enemies > 0 then
				self:switchToState( "attack" )
			else
				self.currentStayAlertTime = self.currentStayAlertTime + dt
				if self.currentStayAlertTime >= self.totalStayAlertTime then
					self.currentStayAlertTime = 0
					self:switchToState( "idle" )
				end
			end
			
			
		elseif self.state == "attack" then
			
			local enemies = self:getEnemiesInAttackRange()
			
			if #enemies == 0 or self.remainingNumberOfCharges < 1 then
				self:switchToState( "idle" )
			else
				self.currentAttackWait = self.currentAttackWait + dt
				if self.currentAttackWait >= self.totalAttackWait then
					self.currentAttackWait = self.currentAttackWait - self.totalAttackWait
					
					-- fire at a random enemy in range
					local enemy = enemies[random(1,#enemies)]
					-- print( "tesla fire at enemy", enemy.id )
					
					-- create "projectile"
					-- add a fake offset to make it look like the "projectile" originates from the top of the building
					local projectile = EntityManager:createEntity(
						ENTITIES.teslaBolt,
						{ x = ex-80, y = ey-80, targetId = enemy.id }
						-- { x = ex-80, y = ey-80, target = enemy, damage = self.attackDamage, }
					)
					Game:onEntityHurtEntity( self, enemy, self.attackDamage )
					
					self.remainingNumberOfCharges = self.remainingNumberOfCharges - 1
				end
			end
			
		end
		
	end
	
	
	function entity:switchToState( newState )
		
		self.state = newState
		-- print( "tesla tower: switch to state", newState )
		
		-- TODO: update or reset values?
		if newState == "attack" then
			self.remainingNumberOfCharges = self.maxNumberOfCharges
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:getEnemiesInAttackRange()
		local ex, ey = self:getPosition()
		local enemies = EntityManager:getEntitiesFilter( function(e)
			-- entity must have a healt component
			if not e.health
			or e.entityCategory ~= ENTITYCATEGORY.monster then
				return false
			end
			local x, y = e:getPosition()
			return Tools.getDistance2( x,y, ex, ey ) <= self.attackRange
		end )
		-- print( "#enemies:", #enemies )
		return enemies
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:draw()
		local body = self.body
		local x, y = convertToDimetric( body.x, body.y )
		local imgData = self.image
		local ox = imgData.ox or 0
		local oy = imgData.oy or 0
		
		lg.draw( imgData.image, x, y, 0, 1, 1, ox, oy )
		
		-- draw "light"
		-- TODO: do not draw when the building is not operating at the moment
		if true then
			local liSize = 800
			local alpha = 0.8
			local color = { 1, 0.7, 0.7, alpha }
			Renderer:drawCircularLight( x, y, liSize, color )
		end
		
		self.ohwSelection:setVisible( self.selected == true )
		
		-- healthbar
		if self.selected or self.timeSinceLastHealthChange < 4 then
			local hb_y_offset = body.z + 100	-- draw above the entity
			self.health:draw( x, y - hb_y_offset, 50, 8 )
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:destroy()
		Gameloop:remove( self )
		Renderer:remove( self, self.rendererLayer )
		self.ohwSelection:destroy()
		self = nil
	end
	
	return entity
end

return TeslaTower
