local ToolStore = {}
local lg = love.graphics


function ToolStore:create( data )
	local entity = Building:create( data )
	
	entity.entityType = ENTITIES.toolStore
	-- entity.entityCategory = ENTITYCATEGORY.building
	
	-- also store the tile index (avoid to calculate them over and over again)
	entity.xi = data.xi
	entity.yi = data.yi
	
	-- first create a construction site (image)
	entity.isConstructionComplete = true
	
	-- range in full tiles
	entity.explorationRange = 4
	-- entity.explorationIgnoresWalls = true	-- test only
	
	entity.maxHealth = 500
	
	entity.image = Images["tool_store_1"]
	entity.rendererLayer = "entities"
	
	-- selection frame image
	entity.imageWidth = 120
	entity.imageHeight = 120
	entity.outlineOffsetY = 24
	local frameImageData = Images["entity_selection_frame_1"]
	entity.selectionFrameImage = frameImageData
	entity.selectionFrameScaleX = entity.imageWidth / frameImageData.image:getWidth()
	entity.selectionFrameScaleY = entity.imageHeight / frameImageData.image:getHeight()
	
	-- blocking info in mini tiles
	-- for now each building can only occupy one full tile
	entity.blocking = {
		{0,0,0,0,0},
		{0,1,1,1,0},
		{0,1,1,1,0},
		{0,1,1,1,0},
		{0,0,0,0,0},
	}
	
	-- workers can deposit resources here
	-- need canDepositEntity and getResourceDepositPosition function
	entity.isResourceDeposit = true
	entity.storedEntities = {}	-- list of lists of entity IDs by entity type (e.g. ore)
	
	----------------------------------------------------------------------------------------------------
	
	function entity:canDepositEntity( entityType )
		return entityType == ENTITIES.ore
		or entityType == ENTITIES.energyCrystal
		or entityType == ENTITIES.rockSample
	end
	
	
	-- absolute world position where workers can deposit resources
	-- important: position must not be blocked!
	function entity:getResourceDepositPosition()
		local x, y = self:getPosition()
		return x + 50, y - 25
	end
	
	
	-- store the entity ID
	-- e.g. an energy crystal
	function entity:storeEntity( entity )
		local eType = entity.entityType
		-- print( "toolstore: storeEntity (id, type)", entity.id, eType )
		self.storedEntities[eType] = self.storedEntities[eType] or {}
		table.insert( self.storedEntities[eType], entity.id )
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- absolute world position where construction fences can be spawned
	-- important: position must not be blocked!
	function entity:getConstructionFenceSpawnPosition()
		local x, y = self:getPosition()
		return x + 50, y + 25
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:load( id )
		self.id = id
		
		Gameloop:add( self )
		Renderer:add( self, self.rendererLayer )
		
		self.ohwSelection = OhwEntitySelection:create{
			entityId = self.id,
			width = self.imageWidth,
			height = self.imageHeight,
			offsetY = self.outlineOffsetY,
		}
		
		Game:onBuildingCreated( self )
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- set currently selected - true / false
	function entity:setSelected( selected )
		self.selected = selected
		self.animationTick = 0
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:tick( dt )
		
		if self.selected then
			self.animationTick = self.animationTick + dt
		end
		
		self.timeSinceLastHealthChange = self.timeSinceLastHealthChange + dt
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:draw()
		local body = self.body
		local x, y = convertToDimetric( body.x, body.y )
		
		local imgData = self.image
		local ox = imgData.ox or 0
		local oy = imgData.oy or 0
		
		lg.draw( imgData.image, x, y, 0, 1, 1, ox, oy )
		
		-- draw "light"
		-- TODO: do not draw when the building is not operating at the moment
		if true then
			local liSize = 800
			local alpha = 0.8
			local color = { 1, 0.7, 0.7, alpha }
			Renderer:drawCircularLight( x, y, liSize, color )
		end
		
		self.ohwSelection:setVisible( self.selected == true )
		
		-- healthbar
		if self.selected or self.timeSinceLastHealthChange < 4 then
			local hb_y_offset = body.z + 100	-- draw above the entity
			self.health:draw( x, y - hb_y_offset, 50, 8 )
			-- local healthRatio = self.health:getHealthRatio()
			-- local hb_width = 50
			-- local hb_height = 8
			-- local hb_x = x - hb_width * 0.5
			-- local hb_y = y - hb_height * 0.5 - hb_y_offset
			-- lg.setColor( COLORS.hb_black )
			-- lg.rectangle( "fill", hb_x, hb_y, hb_width, hb_height )
			-- lg.setColor( 0.33, 1, 0, 0.8 )
			-- lg.rectangle( "fill", hb_x, hb_y, hb_width * healthRatio, hb_height )
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:destroy()
		Gameloop:remove( self )
		Renderer:remove( self, self.rendererLayer )
		self.ohwSelection:destroy()
		self = nil
	end
	
	
	return entity
end


function ToolStore:getConstructionCost()
	return {
		[ENTITIES.ore] = 8,
	}
end


return ToolStore
