local ProbeDroid = {}
local lg = love.graphics
local pi = math.pi
local pipi = 2 * pi


-- special: exploration range AND type can be variable
-- idea: small range while moving, normal range when idle
--> higher range while scanning (remain in position for a few seconds) + ability to scan through walls
local explorationRangeIdle = 3
local explorationRangeBusy = 2
local explorationRangeScan = 4
local totalTimeToStartScan = 3	-- time in seconds until the scanners are fully operational (switch to scanning task)

-- images in stud.io are 128x128 with a zoom of 1.2 (and a y offset of about -15?)

function ProbeDroid:create( data )
	local entity = {}
	
	entity.entityType = ENTITIES.probeDroid
	entity.entityCategory = ENTITYCATEGORY.robot
	
	-- by default, this entity can be controlled by the player
	-- deactivate e.g. for entities that should be rescued from a different cave
	entity.isPlayerControlled = true
	
	entity.body = {
		shape = "circle",
		x = data.x or 0,
		y = data.y or 0,
		radius = 4,
		z = 16,	-- draw the entity above the origin of the body
	}
	entity.orientation = 0
	
	-- 25/s = one mini tile per second
	entity.movementSpeed = 100
	
	-- range in full tiles
	entity.explorationRange = 2
	-- explorationIgnoresWalls can change depending on the state: only explore when idle
	entity.timeToStartScan = totalTimeToStartScan
	
	entity.maxHealth = 150
	
	-- current task
	entity.task = "idle"
	
	-- world position and mini tile index
	entity.movementDestination = {}-- x = 0, y = 0, mtx = 0, mty = 0 }
	
	entity.rendererLayer = "entities"
	
	-- animations
	entity.animations = {}
	entity.animations.idle = Animation.new( "animations/probe droid/idle.lua" )
	entity.animations.idle:setRelativeOrigin( 0.5, 1 )
	entity.animations.scan = Animation.new( "animations/probe droid/scan.lua" )
	entity.animations.scan:setRelativeOrigin( 0.5, 1 )
	
	-- set initial animation set
	entity.currentAnimation = "idle"
	
	-- 2d hitbox for mouse clicks
	entity.imageWidth = 64
	entity.imageHeight = 80
	entity.outlineOffsetY = 40
	
	----------------------------------------------------------------------------------------------------
	
	function entity:load( id )
		self.id = id
		
		Gameloop:add( self )
		Renderer:add( self, self.rendererLayer )
		
		self.ohwSelection = OhwEntitySelection:create{
			entityId = self.id,
			width = self.imageWidth,
			height = self.imageHeight,
			offsetY = self.outlineOffsetY,
		}
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- in order to check if a mouse click was issued to select this entity, use the 2d screen coordinates
	function entity:get2dHitbox()
		local x, y = convertToDimetric( self.body.x, self.body.y )
		local x, y = Camera:getTransform():transformPoint( x, y )
		local cameraScale = Camera:getScale()
		return {
			shape = "rect",
			x = Tools.round(x),
			-- do not forget the camera scale here!
			y = Tools.round(y) - self.outlineOffsetY / cameraScale,
			sizeX = self.imageWidth / cameraScale,
			sizeY = self.imageHeight / cameraScale,
		}
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- set currently selected - true / false
	function entity:setSelected( selected )
		self.selected = selected
		self.selectionAnimationTick = 0
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:getPosition()
		return self.body.x, self.body.y
	end
	
	
	function entity:setPosition( x, y )
		self.body.x = x
		self.body.y = y
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- range in full tiles
	function entity:getExplorationRange()
		return self.explorationRange
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- from game
	-- check if this entity can be given a move command at the moment
	function entity:canIssueMoveCommand()
		if self.task == "leaveBuilding" then
			return false
		end
		
		return true
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- get the currently possible movement speed
	-- TODO: can be increased with technologies?
	function entity:getCurrentMovementSpeed()
		local speed = self.movementSpeed
		
		return speed
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- internal only
	-- mini tile coordinates and world position
	function entity:_setMovementDestination( mini_xi, mini_yi, x, y )
		self.movementDestination.x = x
		self.movementDestination.y = y
		self.movementDestination.mtx = mini_xi
		self.movementDestination.mty = mini_yi
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- from extern: move to a specific location (from player move command)
	-- mini tile coordinates and world position
	function entity:moveToDestination( mini_xi, mini_yi, x, y )
		local ex, ey = self:getPosition()
		local start_txi, start_tyi = getTileIndex( ex, ey )
		local target_xi, target_yi = getTileIndex( x, y )
		
		-- extra: pre-check if there can be a path at all
		if not Pathfinding.getPath( start_txi, start_tyi, target_xi, target_yi, "flying_coarse" ) then
			print( "probe droid:moveToDestination - no path found! (1)" )
			return false
		end
		
		local start_xi, start_yi = getMiniTileIndex( ex, ey )
		local path = Pathfinding.getPath( start_xi, start_yi, mini_xi, mini_yi, "flying_fine" )
		if path then
			
			-- if canMove then
				
				-- new movement
				self:_setMovementDestination( mini_xi, mini_yi, x, y )
				self:setMovementPath( path )
				self:setTask( "moveTo" )
				
			-- end
			
		else
			print( "probe droid:moveToDestination - no path found! (2)" )
		end
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:setOrientation( angle )
		self.orientation = angle
		-- note: for some reason the angle sometimes gets larger than 2*pi instead of being negative
		if self.orientation > pipi then
			self.orientation = self.orientation - pipi
		end
		
		-- if self.currentAnimation == "idle" then
			local currName = self.animations[self.currentAnimation]:getCurrentState()
			local animName = self:getAnimationNameByOrientation( self.orientation )
			if currName ~= animName then
				self.animations[self.currentAnimation]:setState( animName )
			end
		-- end
	end
	
	
	function entity:getAnimationNameByOrientation( angle )
		local a = angle
		if a >= 0.25*pi and a < 0.75*pi then
			return "facing_sw"
		elseif a < -0.25*pi and a >= -0.75*pi then
			return "facing_ne"
		elseif math.abs( a ) >= 0.75*pi then
			return "facing_nw"
		else
			return "facing_se"
		end
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:tick( dt )
		
		if self.selected then
			self.selectionAnimationTick = self.selectionAnimationTick + dt
		end
		
		
		-- behavior depending on the internal state
		self:behaviorTick( dt )
		
		
		-- animation
		if self.currentAnimation and self.animations[self.currentAnimation] then
			local anim = self.animations[self.currentAnimation]
			anim:update( dt )
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:behaviorTick( dt )
		
		-- TODO
		
		if self.task == "moveTo" then
			
			-- currently want to move somewhere
			if self.movementPath then
				local nSteps = #self.movementPath
				if nSteps > 0 then
					local ex, ey = self:getPosition()
					local start_xi, start_yi = getMiniTileIndex( ex, ey )
					if start_xi == self.movementPath[nSteps].x
					and start_yi == self.movementPath[nSteps].y then
						table.remove( self.movementPath, nSteps )
					else
						-- move to the next mini tile center
						local goal_x, goal_y = miniTileIndexToPosition( self.movementPath[nSteps].x, self.movementPath[nSteps].y )
						local direction = Tools.vectorSub( self.body, { x = goal_x, y = goal_y } )
						local vec2 = Tools.vectorSetLength( direction, self:getCurrentMovementSpeed() )
						self:move( vec2.x * dt, vec2.y * dt )
					end
				else
					-- already on the final mini tile
					-- move to the exact target location
					local direction = Tools.vectorSub( self.body, self.movementDestination )
					local vec2 = Tools.vectorTrim( direction, self:getCurrentMovementSpeed() * dt )
					self:move( vec2.x, vec2.y )
					
					if math.abs( self.body.x - self.movementDestination.x ) < 2
					and math.abs( self.body.y - self.movementDestination.y ) < 2 then
						self.movementPath = nil
						self:onMovementDestinationReached()
					end
				end
			end
			
		elseif self.task == "leaveBuilding" then
			
			-- entity is likely still in the blocking area of the building
			-- but that why there is a different task for this
			local direction = Tools.vectorSub( self.body, self.leaveBuildingPosition )
			local vec2 = Tools.vectorTrim( direction, self:getCurrentMovementSpeed() * dt )
			self:move( vec2.x, vec2.y )
			
			-- if position reached, set task to idle
			if math.abs( self.body.x - self.leaveBuildingPosition.x ) < 2
			and math.abs( self.body.y - self.leaveBuildingPosition.y ) < 2 then
				self.leaveBuildingPosition = nil
				self:setTask( "idle" )
			end
			
		elseif self.task == "idle" then
			
			self.timeToStartScan = self.timeToStartScan - dt
			if self.timeToStartScan <= 0 then
				self.timeToStartScan = nil
				self:setTask( "scanning" )
			end
			
		end
		
		
		-- special: dynamic exploration properties
		if self.task == "idle" then
			self.explorationRange = explorationRangeIdle
			self.explorationIgnoresWalls = false
		elseif self.task == "scanning" then
			self.explorationRange = explorationRangeScan
			self.explorationIgnoresWalls = true
		else
			self.explorationRange = explorationRangeBusy
			self.explorationIgnoresWalls = false
		end
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- set animation etc.
	function entity:setTask( task )
		
		-- already doing that?
		if self.task == task then
			return
		end
		
		-- update task
		self.task = task
		-- print( "probe droid: set task:", task )
		
		if task == "idle" then
			self.timeToStartScan = totalTimeToStartScan
			self.currentAnimation = "idle"
		elseif task == "scanning" then
			self.currentAnimation = "scan"
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:onMovementDestinationReached()
		
		-- TODO: update task
		self:setTask( "idle" )
		
		self:setOrientation( self.orientation )	-- after updating the animation, keep the last orientation
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:setMovementPath( path )
		self.movementPath = path
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:move( dx, dy )
		self.body.x = self.body.x + (dx or 0)
		self.body.y = self.body.y + (dy or 0)
		self:setOrientation( Tools.getAngleFromVector{ x = dx, y = dy } )
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- teleport down to building, leave the building (inside blocking)
	function entity:setLeaveBuildingState( x, y )
		self.leaveBuildingPosition = { x = x, y = y }
		self:setTask( "leaveBuilding" )
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:draw()
		local body = self.body
		local x, y = convertToDimetric( body.x, body.y )
		-- local scale = self.imageScale
		
		-- animation
		if self.currentAnimation and self.animations[self.currentAnimation] then
			local anim = self.animations[self.currentAnimation]
			anim:setPosition( x, y - body.z )	-- extra: offset makes it look like the entity is in the air
			anim:draw( 0.6, 0.6 )
		end
		
		-- draw "light"
		do
			local liSize = 500
			if self.task == "scanning" then
				liSize = 800
			end
			local alpha = 0.8
			local color = { 1, 0.7, 0.7, alpha }
			Renderer:drawCircularLight( x, y, liSize, color )
		end
		
		self.ohwSelection:setVisible( self.selected == true )
		
		-- "shadow"
		lg.setColor( 0.1, 0.1, 0.1, 0.5 )
		lg.ellipse( "fill", x, y, 8, 4 )
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:destroy()
		Gameloop:remove( self )
		Renderer:remove( self, self.rendererLayer )
		self.ohwSelection:destroy()
		self = nil
	end
	
	
	return entity
end


return ProbeDroid
