local RockRaider = {}
local lg = love.graphics
local pi = math.pi
local pipi = 2 * pi

-- image size is 128x128
-- rendered in studio with a zoom scale of 0.7

--[[
info for part designer (minifig) images
https://docs.google.com/document/d/1aXIMGFmB2IS6BJSBDjoY_g8mZ8bjXM_PGUApxaohTPk/edit#heading=h.cazyd79wmote
face: 416x320 px inner: 193x142 px
torso: 608x512 (top: 422) px
legs: 608x640
--]]


function RockRaider:create( data )
	local entity = {}
	
	entity.entityType = ENTITIES.rockRaider
	entity.entityCategory = ENTITYCATEGORY.rockRaider
	
	-- by default, this entity can be controlled by the player
	-- deactivate e.g. for entities that should be rescued from a different cave
	entity.isPlayerControlled = true
	
	-- Rock Raiders can either automatically be assigned new jobs from the game
	-- or they only perform what the player manually assigns while disableAutomaticJobAssigment is true
	entity.disableAutomaticJobAssigment = false
	
	entity.body = {
		shape = "circle",
		x = data.x or 0,
		y = data.y or 0,
		radius = 4,
		z = 0,
	}
	entity.orientation = 0
	
	-- this entity can perform worker jobs
	entity.isWorker = true
	
	-- 25/s = one mini tile per second
	entity.movementSpeed = 64
	
	-- not implemented yet: modifiers for the movement speed
	entity.movementSpeedModifierDebris = 0.6	-- walk slower when on debris
	entity.movementSpeedModifierFlee = 1.3	-- walk faster when fleeing from a threat
	entity.movementSpeedModifierCarry = 0.7	-- walk slower when carrying something
	entity.movementSpeedModifierLeaveBuilding = 0.5	-- e.g. leaving a teleport pad
	
	-- energy or steadiness of the protective suit
	entity.maxHealth = 200
	
	-- efficiency for drilling walls = reduce their endurance until they collapse
	-- while drilling: wall.drillEndurance -= entity.drillWallEfficiency * dt
	entity.drillWallEfficiency = 40
	
	-- time in seconds it takes to remove mass from debris
	entity.clearDebrisTimePerStep = 1
	
	-- range in full tiles
	entity.explorationRange = 3
	
	
	-- current task
	-- idle
	-- moveTo	-- position from player command or job
	-- drillWall
	-- clearDebris
	-- collectResource
	-- teleport? down/up?
	-- driveVehicle
	entity.task = "idle"
	
	-- world position and mini tile index
	entity.movementDestination = {}-- x = 0, y = 0, mtx = 0, mty = 0 }
	
	entity.rendererLayer = "entities"
	-- for now: one static image
	-- entity.image = Images["rock_raider_1"]
	-- entity.imageScale = 0.2
	
	-- animations
	entity.animations = {}
	entity.animations.idle = Animation.new( "animations/rock raider/idle.lua" )
	entity.animations.idle:setRelativeOrigin( 0.5, 0.8 )
	entity.animations.carryIdle = Animation.new( "animations/rock raider/carry idle.lua" )
	entity.animations.carryIdle:setRelativeOrigin( 0.5, 0.8 )
	entity.animations.walk = Animation.new( "animations/rock raider/walk.lua" )
	entity.animations.walk:setRelativeOrigin( 0.5, 0.8 )
	entity.animations.carryWalk = Animation.new( "animations/rock raider/carry walk.lua" )
	entity.animations.carryWalk:setRelativeOrigin( 0.5, 0.8 )
	entity.animations.shovel = Animation.new( "animations/rock raider/shovel.lua" )
	entity.animations.shovel:setRelativeOrigin( 0.5, 0.8 )
	entity.animations.drill = Animation.new( "animations/rock raider/drill.lua" )
	entity.animations.drill:setRelativeOrigin( 0.5, 0.8 )
	entity.animations.reinforceWall = Animation.new( "animations/rock raider/reinforce wall.lua" )
	entity.animations.reinforceWall:setRelativeOrigin( 0.5, 0.8 )
	
	-- set initial animation set
	entity.currentAnimation = "idle"
	
	
	-- 2d hitbox for mouse clicks
	-- values are derived from the source image (draw by eye)
	entity.imageWidth = 50
	entity.imageHeight = 50
	entity.outlineOffsetY = 20
	
	-- selection frame image
	-- local frameImageData = Images["entity_selection_frame_1"]
	-- entity.selectionFrameImage = frameImageData
	-- entity.selectionFrameScaleX = entity.imageWidth / frameImageData.image:getWidth()
	-- entity.selectionFrameScaleY = entity.imageHeight / frameImageData.image:getHeight()
	
	-- when the unit recently lost or gained health, draw the healthbar for a few seconds
	entity.timeSinceLastHealthChange = 0
	
	----------------------------------------------------------------------------------------------------
	
	function entity:load( id )
		self.id = id
		
		self.health = Health:create( self, self.maxHealth, false, 0 )	-- health component
		Gameloop:add( self )
		Renderer:add( self, self.rendererLayer )
		
		self.ohwSelection = OhwEntitySelection:create{
			entityId = self.id,
			width = self.imageWidth,
			height = self.imageHeight,
			offsetY = self.outlineOffsetY,
		}
		
		self.ohwWorkerTask = OhwWorkerTask:create{
			entityId = self.id,
			width = self.imageWidth,
			height = self.imageHeight,
			offsetY = 70,
		}
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- in order to check if a mouse click was issued to select this entity, use the 2d screen coordinates
	function entity:get2dHitbox()
		local x, y = convertToDimetric( self.body.x, self.body.y )
		local x, y = Camera:getTransform():transformPoint( x, y )
		local cameraScale = Camera:getScale()
		return {
			shape = "rect",
			x = Tools.round(x),
			-- do not forget the camera scale here!
			y = Tools.round(y) - self.outlineOffsetY / cameraScale,
			sizeX = self.imageWidth / cameraScale,
			sizeY = self.imageHeight / cameraScale,
		}
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- set currently selected - true / false
	function entity:setSelected( selected )
		self.selected = selected
		self.selectionAnimationTick = 0
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:getPosition()
		return self.body.x, self.body.y
	end
	
	
	function entity:setPosition( x, y )
		self.body.x = x
		self.body.y = y
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- range in full tiles
	function entity:getExplorationRange()
		return self.explorationRange
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- from game
	-- check if this entity can be given a move command
	-- return false if the entity is in a building or somehow not available right now
	function entity:canIssueMoveCommand()
		if self.task == "leaveBuilding"
		or self.task == "teleport"
		or self.task == "driveVehicle" then
			return false
		end
		
		-- cannot start a move command when in a blocked area
		local x, y = self:getPosition()
		local mini_xi, mini_yi = getMiniTileIndex( x, y )
		if Game.blocking:isMiniTileBlocked( mini_xi, mini_yi ) then
			-- print( "RR cannot move while in blocking!" )
			return false
		end
		
		return true
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- get the currently possible movement speed
	function entity:getCurrentMovementSpeed()
		local speed = self.movementSpeed
		
		-- no further modifications when leaving a building
		if self.task == "leaveBuilding" then
			return self.movementSpeed * self.movementSpeedModifierLeaveBuilding
		end
		
		-- move slower when on debris
		local x, y = self:getPosition()
		local xi, yi = getTileIndex( x, y )
		local debris = EntityManager:getEntitiesFilter( function(e)
			return e.entityCategory == ENTITYCATEGORY.debris
				and e.xi == xi and e.yi == yi
		end )
		if debris[1] then
			speed = speed * self.movementSpeedModifierDebris
		end
		
		-- move slower when carrying something
		if self.carriedEntityId then
			speed = speed * self.movementSpeedModifierCarry
		end
		
		return speed
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- internal only
	-- mini tile coordinates and world position
	function entity:_setMovementDestination( mini_xi, mini_yi, x, y )
		self.movementDestination.x = x
		self.movementDestination.y = y
		self.movementDestination.mtx = mini_xi
		self.movementDestination.mty = mini_yi
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- from extern: move to a specific location (from player move command)
	-- mini tile coordinates and world position
	function entity:moveToDestination( mini_xi, mini_yi, x, y )
		local ex, ey = self:getPosition()
		local start_txi, start_tyi = getTileIndex( ex, ey )
		local target_xi, target_yi = getTileIndex( x, y )
		
		-- extra: pre-check if there can be a path at all
		if not Pathfinding.getPath( start_txi, start_tyi, target_xi, target_yi, "ground_coarse" ) then
			print( "RR:moveToDestination - no path found! (1)" )
			return false
		end
		
		local start_xi, start_yi = getMiniTileIndex( ex, ey )
		local path = Pathfinding.getPath( start_xi, start_yi, mini_xi, mini_yi, "ground_fine" )
		if path then
			-- print( "RR:moveToDestination - path length:", #path )
			
			-- if canMove then
				
				-- important: cancel the current job!
				if self.currentJob then
					self:cancelCurrentJob()
				end
				
				-- new movement
				self:_setMovementDestination( mini_xi, mini_yi, x, y )
				self:setMovementPath( path )
				self:setTask( "moveTo" )
				
			-- end
			
		else
			print( "RR:moveToDestination - no path found!" )
		end
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:cancelCurrentJob()
		if not self.currentJob then
			return false
		end
		
		if self.carriedEntityId then
			self:dropCarriedEntity()
		end
		
		-- do not list jobs again that have been assigned to this worker manually by the player
		if not self.currentJob.manuallyAssigned then
			Game:entityCancelJob( self, self.currentJob )
		end
		self.currentJob = nil
		self:setTask( "idle" )
		-- print( "RR: cancel job" )
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:setOrientation( angle )
		self.orientation = angle
		-- note: for some reason the angle sometimes gets larger than 2*pi instead of being negative
		if self.orientation > pipi then
			self.orientation = self.orientation - pipi
		end
		
		if self.currentAnimation then	-- no animation while driving
			local currName = self.animations[self.currentAnimation]:getCurrentState()
			local animName = self:getAnimationNameByOrientation( self.orientation )
			if currName ~= animName then
				self.animations[self.currentAnimation]:setState( animName )
			end
		end
	end
	
	
	function entity:getAnimationNameByOrientation( angle )
		local a = angle
		if a >= 0.25*pi and a < 0.75*pi then
			return "facing_sw"
		elseif a < -0.25*pi and a >= -0.75*pi then
			return "facing_ne"
		elseif math.abs( a ) >= 0.75*pi then
			return "facing_nw"
		else
			return "facing_se"
		end
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:tick( dt )
		
		if self.selected then
			self.selectionAnimationTick = self.selectionAnimationTick + dt
		end
		
		self.timeSinceLastHealthChange = self.timeSinceLastHealthChange + dt
		
		
		-- current job
		if self.currentJob then
			if Game:isJobDone( self.currentJob ) then
				self:jobDone()
			end
		end
		
		
		-- behavior depending on the internal state
		self:behaviorTick( dt )
		
		
		-- animation
		if self.currentAnimation and self.animations[self.currentAnimation] then
			local anim = self.animations[self.currentAnimation]
			anim:update( dt )
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:behaviorTick( dt )
		
		local job = self.currentJob
		
		if self.task == "moveTo" then
			-- check if arrived at job site?
			
			-- currently want to move somewhere
			if self.movementPath then
				local nSteps = #self.movementPath
				if nSteps > 0 then
					local ex, ey = self:getPosition()
					local start_xi, start_yi = getMiniTileIndex( ex, ey )
					if start_xi == self.movementPath[nSteps].x
					and start_yi == self.movementPath[nSteps].y then
						-- print( "RR: remove step", self.movementPath[nSteps].x, self.movementPath[nSteps].y )
						table.remove( self.movementPath, nSteps )
					else
						-- move to the next mini tile center
						local goal_x, goal_y = miniTileIndexToPosition( self.movementPath[nSteps].x, self.movementPath[nSteps].y )
						local direction = Tools.vectorSub( self.body, { x = goal_x, y = goal_y } )
						local vec2 = Tools.vectorSetLength( direction, self:getCurrentMovementSpeed() )
						self:move( vec2.x * dt, vec2.y * dt )
					end
				else
					-- already on the final mini tile
					-- move to the exact target location
					local direction = Tools.vectorSub( self.body, self.movementDestination )
					local vec2 = Tools.vectorTrim( direction, self:getCurrentMovementSpeed() * dt )
					self:move( vec2.x, vec2.y )
					
					if math.abs( self.body.x - self.movementDestination.x ) < 2
					and math.abs( self.body.y - self.movementDestination.y ) < 2 then
						-- print( "RR: path done" )
						self.movementPath = nil
						self:onMovementDestinationReached()
					end
				end
			end
			
		elseif self.task == "teleport" then
			
			self.remainingTeleportTime = self.remainingTeleportTime - dt
			if self.remainingTeleportTime <= 0 then
				self.remainingTeleportTime = nil
				-- self:setTask( "leaveBuilding" )
				if self.teleporterEntity and self.teleporterEntity.getLeaveBuildingPosition then
					local x2, y2 = self.teleporterEntity:getLeaveBuildingPosition()
					self:setLeaveBuildingState( x2, y2 )
				else
					self:setTask( "idle" )
					self.teleporterEntity = nil
				end
			end
			
		elseif self.task == "driveVehicle" then
			
			-- synch position with vehicle
			local vehicle = self:getVehicle()
			if vehicle then
				local vehicle_x, vehicle_y = vehicle:getPosition()
				self:setPosition( vehicle_x, vehicle_y )
			end
			
		elseif self.task == "leaveBuilding" then
			
			-- entity is likely still in the blocking area of the building
			-- but that why there is a different task for this
			local direction = Tools.vectorSub( self.body, self.leaveBuildingPosition )
			-- note: the movement speed might be higher than what would be achieved when using the full vector length
			-- alternative: see above, using vectorSetLength is ok, but the dt modifier must be considered there already!
			local vec2 = Tools.vectorTrim( direction, self:getCurrentMovementSpeed() * dt )
			-- local vec2 = Tools.vectorSetLength( direction, self:getCurrentMovementSpeed() )
			self:move( vec2.x, vec2.y )
			-- print( Tools.vectorGetLength(vec2) )
			
			-- if position reached, set task to idle
			if math.abs( self.body.x - self.leaveBuildingPosition.x ) < 2
			and math.abs( self.body.y - self.leaveBuildingPosition.y ) < 2 then
				self.leaveBuildingPosition = nil
				self:setTask( "idle" )
				-- print( "RR leaveBuildingPosition reached" )
			end
			
		elseif self.task == "pickUpEntity" then
			
			-- is the item still there and can be picked up?
			local item = EntityManager:getEntityById( job.entityId )
			-- also make sure the entity does not already carry the item
			-- e.g. if there is no toolstore to bring it to, the RR will just stand and wait with the item
			-- which is at least better than picking it up and dropping it over and over again
			if item and self.carriedEntityId ~= item.id then
				-- check position
				local ex, ey = self:getPosition()
				local item_x, item_y = item:getPosition()
				if Tools.getDistance2( ex, ey, item_x, item_y ) > 50 then
					print( "RR: item to pick up is too far away ..." )
					-- easy "solution": set to idle...
					-- avoid double check if the new item's position can be reached etc.
					self:cancelCurrentJob()
					-- self:setTask( "idle" )
					return
				end
				-- check that the item is not already carried by another entity...
				if not item:canBePickedUp() then
					print( "RR: item cannot be picked up now ..." )
					self:cancelCurrentJob()
					-- self:setTask( "idle" )
					return
				end
				
				-- yes, this item can be picked up ... do it
				self:pickUpEntity( item )
				item:setCarrierEntity( self )
				-- Game:onEntityPickedUpEntity( self, item )
				
				-- RR now carries this entity, what now?
				-- look for a deposit (e.g. tool store)
				if item.entityType == ENTITIES.energyCrystal
				or item.entityType == ENTITIES.ore
				or item.entityType == ENTITIES.rockSample then
					
					-- search for the closest place to deposit the resource
					local deposit = Game:getClosestDeposit( self, item )
					if deposit then
						-- local ex, ey = self:getPosition()
						local start_xi, start_yi = getMiniTileIndex( ex, ey )
						local target_x, target_y = deposit:getResourceDepositPosition()
						local target_xi, target_yi = getMiniTileIndex( target_x, target_y )
						local path = Pathfinding.getPath( start_xi, start_yi, target_xi, target_yi, "ground_fine" )
						if path then
							self:_setMovementDestination( target_xi, target_yi, target_x, target_y )
							self:setMovementPath( path )
							-- print( "RR: move to deposit" )
							-- self.targetDeposit = deposit
							self.targetDepositId = deposit.id
							self:setTask( "moveTo" )	-- TODO: set task moveToDeposit?
						else
							-- no path found ... what now?
							-- self:setTask( "idle" )	-- TODO
							-- self:cancelCurrentJob()
						end
					else
						-- no deposit ... what now?
						-- self:setTask( "idle" )	-- TODO
						-- self:cancelCurrentJob()
					end
					
				elseif item.entityType == ENTITIES.constructionFence then
					
					-- go to the construction site where the fence should be placed at
					local fence = EntityManager:getEntityById( job.entityId )
					if fence then
						local constructionSite = EntityManager:getEntityById( fence.constructionSiteId )
						if constructionSite then
							local ex, ey = self:getPosition()
							local start_xi, start_yi = getMiniTileIndex( ex, ey )
							local target_x, target_y = constructionSite:getFencePosition( fence.constructionSitePositionIndex )
							local target_xi, target_yi = getMiniTileIndex( target_x, target_y )
							local path = Pathfinding.getPath( start_xi, start_yi, target_xi, target_yi, "ground_fine" )
							if path then
								self:_setMovementDestination( target_xi, target_yi, target_x, target_y )
								self:setMovementPath( path )
								-- print( "RR: move to construction site with fence" )
								-- self.targetDeposit = constructionSite
								self.targetDepositId = constructionSite.id
								self:setTask( "moveTo" )
							end
						end
					end
					
				end
				
				-- call after looking for a deposit, so the game can override this when assigning a bringResourceToBuilding job
				Game:onEntityPickedUpEntity( self, item )
				
			end
			
		elseif self.task == "drillWall" then
			
			local wall = EntityManager:getEntityById( job.entityId )
			if wall then
				Game:entityDrillWall( self, wall, dt )
			end
			
		elseif self.task == "reinforceWall" then
			
			self.reinforceWallTick = (self.reinforceWallTick or 0) + dt
			-- time to reinforce a wall: 2 seconds
			if self.reinforceWallTick >= 2 then
				self.reinforceWallTick = nil
				local wall = EntityManager:getEntityById( job.entityId )
				if wall then
					Game:entityReinforceWall( self, wall )
				end
				-- self:setTask( "idle" )
			end
			
		elseif self.task == "clearDebris" then
			
			-- ?
			self.clearDebrisTick = (self.clearDebrisTick or 0) + dt
			if self.clearDebrisTick >= self.clearDebrisTimePerStep then
				self.clearDebrisTick = self.clearDebrisTick - self.clearDebrisTimePerStep
				local debris = EntityManager:getEntityById( job.entityId )
				if debris then
					Game:entityRemovesMassFromDebris( self, debris )
				else
					print( "RR: no debris to remove mass from!" )
				end
			end
			
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:onMovementDestinationReached()
		
		-- make sure the orientation when stopping is aligned with x or y axis (see "light" in draw)
		self.orientation = Tools.roundTo( self.orientation, 0.5 * pi )
		
		local job = self.currentJob
		
		if job then
			if job.type == "clearDebris" then
				self:setTask( "clearDebris" )
				-- since the shovel animation only has one orientation...
				self.orientation = 0.5 * pi
				
			elseif job.type == "drillWall" then
				self:setTask( "drillWall" )
				-- look at the wall!
				local wall = EntityManager:getEntityById( job.entityId )
				if wall then
					local ex, ey = self:getPosition()
					local wx, wy = wall:getPosition()
					self:setOrientation( Tools.getAngleFromVector{ x = wx-ex, y = wy-ey } )
				end
				
			elseif job.type == "reinforceWall" then
				self:setTask( "reinforceWall" )
				-- look at the wall!
				local wall = EntityManager:getEntityById( job.entityId )
				if wall then
					local ex, ey = self:getPosition()
					local wx, wy = wall:getPosition()
					self:setOrientation( Tools.getAngleFromVector{ x = wx-ex, y = wy-ey } )
				end
				
			elseif job.type == "placeConstructionFence" then
				local carriedEntity = EntityManager:getEntityById( self.carriedEntityId )
				if carriedEntity then
					carriedEntity.isPlacedAtConstructionSite = true
					Game:onConstructionFencePlaced( carriedEntity )
					self:dropCarriedEntity()
					self:setTask( "idle" )
				else
					self:setTask( "pickUpEntity" )
				end
				
			elseif job.type == "bringResourceToBuilding" then
				if self.carriedEntityId then
					self:depositCarriedEntity( self.targetDepositId )
					self.targetDeposit = nil
					self.targetDepositId = nil
					self:setTask( "idle" )
				end
				
			elseif job.type == "vehicleGetDriver" then
				self:setTask( "driveVehicle" )
				local vehicle = EntityManager:getEntityById( job.entityId )
				Game:vehicleSetDriver( vehicle, self )
				
			elseif job.type == "collectEnergyCrystal"
			or job.type == "collectOre"
			or job.type == "collectRockSample" then
				if self.carriedEntityId then
					-- if self.targetDeposit then
					if self.targetDepositId then
						-- print( "RR: deposit carried entity" )
						-- self:depositCarriedEntity( self.targetDeposit )
						-- self.targetDeposit = nil
						self:depositCarriedEntity( self.targetDepositId )
						self.targetDepositId = nil
					end
					-- print( "RR: drop carried entity" )
					-- self:dropCarriedEntity()
					self:setTask( "idle" )
				else
					self:setTask( "pickUpEntity" )
				end
				
			end
			
		else	-- no job
			
			self:setTask( "idle" )
		end
		
		-- self:setOrientation( self.orientation )	-- after updating the animation, keep the last orientation
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:setMovementPath( path )
		self.movementPath = path
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- teleport down to teleport pad
	-- building should have a getLeaveBuildingPosition function
	function entity:setTeleportToBuildingState( building )
		self:setTask( "teleport" )
		self.animations[self.currentAnimation]:setSpeedMultiplier( 0.2 )	-- slower animation
		self.teleporterEntity = building
	end
	
	-- leave the building (inside blocking)
	function entity:setLeaveBuildingState( x, y )
		self.leaveBuildingPosition = { x = x, y = y }
		self:setTask( "leaveBuilding" )
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:move( dx, dy )
		self.body.x = self.body.x + (dx or 0)
		self.body.y = self.body.y + (dy or 0)
		self:setOrientation( Tools.getAngleFromVector{ x = dx, y = dy } )
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- check if this worker can take on a given job right now
	function entity:checkReadyForJob( job )
		local jobType = job.type
		if not self:canPerformJobType( jobType ) then
			return false
		end
		
		-- most jobs can only be accepted if the worker is idle
		-- except if they require the worker to have (carry) a certain item
		if jobType == "bringResourceToBuilding" then
			local carriedEntity = EntityManager:getEntityById( self.carriedEntityId )
			if carriedEntity then
				return carriedEntity.entityType == job.resourceType
					and not carriedEntity.destinationId
			else
				return false
			end
		else
			return self.task == "idle"
		end
	end
	
	-- check if this worker can perform a job of the given type at all
	function entity:canPerformJobType( jobType )
		return jobType == "drillWall" or jobType == "clearDebris" or jobType == "collectEnergyCrystal"
			or jobType == "collectOre" or jobType == "vehicleGetDriver" or jobType == "placeConstructionFence"
			or jobType == "bringResourceToBuilding" or jobType == "reinforceWall" or jobType == "collectRockSample"
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- drill wall, remove debris...
	function entity:assignJob( job )
		
		if job.type ~= "bringResourceToBuilding" then
			self:cancelCurrentJob()
		end
		
		-- print( "RR: got new job", job.type )
		self.currentJob = job
		
		if job.type == "drillWall" then
			
			local wall = EntityManager:getEntityById( job.entityId )
			if not wall then
				print( "RR: invalid wall to drill!" )
				self.currentJob = nil
				return false
			end
			
			-- move to a neighbour tile! -> search for a free mini tile index
			local ex, ey = self:getPosition()
			local start_xi, start_yi = getMiniTileIndex( ex, ey )
			-- job.targetTileX, job.targetTileY
			-- parameters: full tile index of the primary tile and full tile index of the neighbour
			local target_xi, target_yi = Game:getFreeMiniTileNextToFullTile( wall.xi, wall.yi, job.targetTileX, job.targetTileY )
			local target_x, target_y = miniTileIndexToPosition( target_xi, target_yi )
			local path = Pathfinding.getPath( start_xi, start_yi, target_xi, target_yi, "ground_fine" )
			if path then
				self:_setMovementDestination( target_xi, target_yi, target_x, target_y )
				self:setMovementPath( path )
				-- print( "RR: found path to job!" )
				-- print( "start:", start_xi, start_yi )
				-- print( "target:", target_xi, target_yi )
				self:setTask( "moveTo" )
				
				self.ohwWorkerTask:setImage( Images["icon_drill_wall"] )
				
			else
				print( "RR: no path to drill job!" )
			end
			
		elseif job.type == "reinforceWall" then
			
			local wall = EntityManager:getEntityById( job.entityId )
			if not wall then
				print( "RR: invalid wall to reinforce!" )
				self.currentJob = nil
				return false
			end
			
			-- move to a neighbour tile! -> search for a free mini tile index
			local ex, ey = self:getPosition()
			local start_xi, start_yi = getMiniTileIndex( ex, ey )
			-- parameters: full tile index of the primary tile and full tile index of the neighbour
			local target_xi, target_yi = Game:getFreeMiniTileNextToFullTile( wall.xi, wall.yi, job.targetTileX, job.targetTileY )
			local target_x, target_y = miniTileIndexToPosition( target_xi, target_yi )
			local path = Pathfinding.getPath( start_xi, start_yi, target_xi, target_yi, "ground_fine" )
			if path then
				self:_setMovementDestination( target_xi, target_yi, target_x, target_y )
				self:setMovementPath( path )
				self:setTask( "moveTo" )
				
				self.ohwWorkerTask:setImage( Images["icon_reinforce_wall"] )
				
			else
				print( "RR: no path to reinforce job!" )
			end
			
		elseif job.type == "clearDebris" then
			
			local debris = EntityManager:getEntityById( job.entityId )
			if not debris then
				print( "RR: invalid debris to remove!" )
				self.currentJob = nil
				return false
			end
			
			-- move to the tile center
			local ex, ey = self:getPosition()
			local start_xi, start_yi = getMiniTileIndex( ex, ey )
			local target_x, target_y = debris:getPosition()
			local target_xi, target_yi = getMiniTileIndex( target_x, target_y )
			local path = Pathfinding.getPath( start_xi, start_yi, target_xi, target_yi, "ground_fine" )
			if path then
				self:_setMovementDestination( target_xi, target_yi, target_x, target_y )
				self:setMovementPath( path )
				-- print( "RR: found path to job!" )
				-- print( "start:", start_xi, start_yi )
				-- print( "target:", target_xi, target_yi )
				self:setTask( "moveTo" )
				
				self.ohwWorkerTask:setImage( Images["icon_clear_debris"] )
				
			else
				print( "RR: no path to debris job!" )
			end
			
		elseif job.type == "collectEnergyCrystal"
		or job.type == "collectOre"
		or job.type == "collectRockSample" then
			
			local resourceEntity = EntityManager:getEntityById( job.entityId )
			if not resourceEntity then
				print( "RR: invalid resourceEntity to collect!" )
				self.currentJob = nil
				return false
			end
			
			local ex, ey = self:getPosition()
			local start_xi, start_yi = getMiniTileIndex( ex, ey )
			local target_x, target_y = resourceEntity:getPosition()
			local target_xi, target_yi = getMiniTileIndex( target_x, target_y )
			local path = Pathfinding.getPath( start_xi, start_yi, target_xi, target_yi, "ground_fine" )
			if path then
				self:_setMovementDestination( target_xi, target_yi, target_x, target_y )
				self:setMovementPath( path )
				-- print( "RR: found path to job!", job.type )
				self:setTask( "moveTo" )
				
				if job.type == "collectEnergyCrystal" then
					self.ohwWorkerTask:setImage( Images["icon_collect_energy_crystal"] )
				elseif job.type == "collectOre" then
					self.ohwWorkerTask:setImage( Images["icon_collect_ore"] )
				elseif job.type == "collectRockSample" then
					self.ohwWorkerTask:setImage( Images["icon_collect_rock_sample"] )
				end
				
			else
				print( "RR: no path to resourceEntity job!" )
			end
			
		elseif job.type == "placeConstructionFence" then
			
			local fence = EntityManager:getEntityById( job.entityId )
			if not fence then
				print( "RR: invalid fence to fetch!" )
				self.currentJob = nil
				return false
			end
			
			local ex, ey = self:getPosition()
			local start_xi, start_yi = getMiniTileIndex( ex, ey )
			local target_x, target_y = fence:getPosition()
			local target_xi, target_yi = getMiniTileIndex( target_x, target_y )
			local path = Pathfinding.getPath( start_xi, start_yi, target_xi, target_yi, "ground_fine" )
			if path then
				self:_setMovementDestination( target_xi, target_yi, target_x, target_y )
				self:setMovementPath( path )
				-- print( "RR: found path to job!" )
				self:setTask( "moveTo" )
				
				self.ohwWorkerTask:setImage( Images["icon_construct_building"] )
				
			else
				print( "RR: no path to fence job!" )
			end
			
			
		elseif job.type == "bringResourceToBuilding" then
			
			-- this worker already carries an item
			-- bring it to the target entity (construction site)
			local constructionSite = EntityManager:getEntityById( job.entityId )
			if not constructionSite then
				print( "RR: invalid constructionSite to bring resource to!" )
				self.currentJob = nil
				return false
			end
			
			local ex, ey = self:getPosition()
			local start_xi, start_yi = getMiniTileIndex( ex, ey )
			local target_x, target_y = constructionSite:getPosition()
			local target_xi, target_yi = getMiniTileIndex( target_x, target_y )
			local path = Pathfinding.getPath( start_xi, start_yi, target_xi, target_yi, "ground_fine" )
			if path then
				self:_setMovementDestination( target_xi, target_yi, target_x, target_y )
				self:setMovementPath( path )
				-- self.targetDeposit = constructionSite
				self.targetDepositId = constructionSite.id
				local carriedEntity = EntityManager:getEntityById( self.carriedEntityId )
				carriedEntity.destinationId = constructionSite.id
				self:setTask( "moveTo" )
				
				self.ohwWorkerTask:setImage( Images["icon_construct_building"] )
				
			else
				print( "RR: no path to bringResourceToBuilding job!" )
			end
			
		elseif job.type == "vehicleGetDriver" then
			
			local vehicle = EntityManager:getEntityById( job.entityId )
			if not vehicle then
				print( "RR: invalid vehicle!" )
				self.currentJob = nil
				return false
			end
			
			local ex, ey = self:getPosition()
			local start_xi, start_yi = getMiniTileIndex( ex, ey )
			local target_x, target_y = vehicle:getPosition()
			local target_xi, target_yi = getMiniTileIndex( target_x, target_y )
			local path = Pathfinding.getPath( start_xi, start_yi, target_xi, target_yi, "ground_fine" )
			if path then
				self:_setMovementDestination( target_xi, target_yi, target_x, target_y )
				self:setMovementPath( path )
				-- print( "RR: found path to job!" )
				self:setTask( "moveTo" )
				
				self.ohwWorkerTask:setImage( Images["icon_vehicle_get_driver"] )
				
			else
				print( "RR: no path to vehicle job!" )
			end
			
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:jobDone()
		-- print( "RR: job is done", self.currentJob.type )
		self.currentJob = nil
		
		self.ohwWorkerTask:setImage()
		
		if self.task == "driveVehicle" then
			-- print( "RR is driving a vehicle now..." )
			return
		end
		
		self:setTask( "idle" )
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- set animation etc.
	function entity:setTask( task )
		
		-- already doing that?
		if self.task == task then
			return
		end
		
		-- update task
		self.task = task
		-- print( "RR: set task:", task )
		
		if task == "idle" then
			self.currentAnimation = "idle"
		elseif task == "moveTo" then
			if self.carriedEntityId then
				self.currentAnimation = "carryWalk"
				self.animations[self.currentAnimation]:setSpeedMultiplier( 0.5 )
			else
				self.currentAnimation = "walk"
			end
		elseif task == "clearDebris" then
			self.currentAnimation = "shovel"
		elseif task == "drillWall" then
			self.currentAnimation = "drill"
		elseif task == "reinforceWall" then
			self.currentAnimation = "reinforceWall"
		elseif task == "driveVehicle" then
			self.currentAnimation = nil
		elseif task == "leaveBuilding" then
			self.currentAnimation = "walk"
		elseif task == "teleport" then
			self.currentAnimation = "idle"	-- TODO: teleport animation?
			self.remainingTeleportTime = 2
		elseif task == "pickUpEntity" then
			self.currentAnimation = "carryIdle"	-- TODO: pick up animation
		end
		
		self:setOrientation( self.orientation )	-- after updating the animation, keep the last orientation
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- called from game when a hurt event occured
	function entity:onHurt( damage )
		
	end
	
	
	function entity:onHealthChanged()
		self.timeSinceLastHealthChange = 0
		
		if self.health:getHealth() <= 0 then
			self:cancelCurrentJob()
			-- force teleport up
			Game:onBeforeEntityTeleportedUp( self )
			-- TODO: animation...
			EntityManager:remove( self )
		end
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- RR can pick up resource entities (and more...)
	-- but only one at a time
	function entity:pickUpEntity( entity )
		if self.carriedEntityId then
			print( "RR WARNING: already carrying an entity!", self.carriedEntityId.id )
		end
		self.carriedEntityId = entity.id
	end
	
	
	function entity:depositCarriedEntity( depositId )
		local deposit = EntityManager:getEntityById( depositId )
		local carriedEntity = EntityManager:getEntityById( self.carriedEntityId )
		if not deposit then
			print( "RR:depositCarriedEntity - deposit does not exist!" )
			self:dropCarriedEntity()
			return
		end
		if deposit.entityType == ENTITIES.constructionSite then
			Game:addResourceEntityToConstructionSite( carriedEntity, deposit )
		else
			Game:depositEntity( carriedEntity, deposit )
		end
		self:dropCarriedEntity()
	end
	
	
	function entity:dropCarriedEntity()
		local carriedEntity = EntityManager:getEntityById( self.carriedEntityId )
		if carriedEntity and carriedEntity.setCarrierEntity then
			carriedEntity:setCarrierEntity( nil )
		end
		self.carriedEntityId = nil
		-- print( "RR: drop carried entity" )
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:setVehicle( vehicle )
		if vehicle then
			self.ohwSelection:setVisible( false )
			self.vehicleEntityId = vehicle.id
		else
			self.vehicleEntityId = nil
		end
	end
	
	
	function entity:getVehicle()
		return EntityManager:getEntityById( self.vehicleEntityId )
	end
	
	
	function entity:isDrivingVehicle()
		return self:getVehicle() ~= false
	end
	
	
	-- from button callback this action should always be safe (position is not blocked)
	-- but when the vehicle gets destroyed, this location might not be safe to exit
	function entity:leaveVehicle()
		local vehicle = self:getVehicle()
		if not vehicle then
			return false
		end
		local x, y = vehicle:getPosition()
		if vehicle.getLeaveVehiclePosition then
			local x2, y2 = vehicle:getLeaveVehiclePosition()
			local mini_xi, mini_yi = getMiniTileIndex( x2, y2 )
			if not Game.blocking:isMiniTileBlocked( mini_xi, mini_yi ) then
				-- only use the vehicle leave position if it is not blocked
				-- because it might lay inside a wall, but the vehicle position itself should not
				x = x2
				y = y2
			end
		end
		self:setVehicle( nil )
		self:setPosition( x, y )
		self:setTask( "idle" )
		
		-- update entity orientation
		self:onMovementDestinationReached()
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- these values are in screen coordinates!
	function entity:getCarriedEntityPositionOffset( entity )
		local anim = self.animations[self.currentAnimation]
		local state = anim:getCurrentState()
		if state == "facing_sw" then
			return -16, -20
		elseif state == "facing_ne" then
			return 16, -20
		elseif state == "facing_nw" then
			return -16, -20
		elseif state == "facing_se" then
			return 16, -20
		end
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:draw()
		
		if self:isDrivingVehicle() then
			return
		end
		
		local body = self.body
		local x, y = convertToDimetric( body.x, body.y )
		-- local imgData = self.image
		-- local ox = imgData.ox or 0
		-- local oy = imgData.oy or 0
		-- local scale = self.imageScale
		
		-- animation
		if self.currentAnimation and self.animations[self.currentAnimation] then
			--print( "player draw animation", self.currentAnimation )
			local anim = self.animations[self.currentAnimation]
			anim:setPosition( x, y )
			-- for the animation an extra scale can be given here...
			anim:draw( 0.5, 0.5 )
		end
		
		-- "light"
		do
			local angle = self.orientation + 0.25 * pi
			local liSize = 500
			local alpha = 0.8
			local color = { 1, 0.7, 0.7, alpha }
			-- y offset so that the light appears to come from the helmet
			Renderer:drawConeLight( x, y-30, liSize, angle, color )
		end
		
		self.ohwSelection:setVisible( self.selected == true )
		
		self.ohwWorkerTask:setVisible( self.selected == true or Game.drawEntityStatusOverhead )
		
		-- lg.circle( "fill", x, y, 6 )	-- debug
		
		-- 2d hitbox (mouse click)
		-- lg.setColor( 1, 1, 1, 0.3 )
		-- lg.rectangle( "line", x - self.imageWidth * 0.5, y - self.imageHeight * 0.5 - self.outlineOffsetY, self.imageWidth, self.imageHeight )
		
		-- healthbar
		-- should be about 4 to 6 times as wide as high and above the entity
		-- the width MIGHT be larger for entities with more health (?)
		if Game.drawEntityStatusOverhead or self.selected
		or self.timeSinceLastHealthChange < 4 then
			local hb_y_offset = body.z + 52	-- draw above the entity
			self.health:draw( x, y - hb_y_offset )
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:destroy()
		Gameloop:remove( self )
		Renderer:remove( self, self.rendererLayer )
		self.ohwSelection:destroy()
		self.ohwWorkerTask:destroy()
		self = nil
	end
	
	
	return entity
end

return RockRaider
