local Digger = {}
local lg = love.graphics
local pi = math.pi
local pipi = 2 * pi

local maxHealth = 500


function Digger:create( data )
	local entity = Vehicle:create( data )
	
	entity.entityType = ENTITIES.digger
	
	-- this entity can perform worker jobs
	entity.isWorker = true
	
	-- 25/s = one mini tile per second
	entity.movementSpeed = 100
	
	-- efficiency for drilling walls = reduce their endurance until they collapse
	-- while drilling: wall.drillEndurance -= entity.drillWallEfficiency * dt
	entity.drillWallEfficiency = 100
	
	-- range in full tiles
	entity.explorationRange = 4
	
	-- current task
	-- idle
	-- moveTo	-- position from player command or job
	-- teleport? down/up?
	entity.task = "idle"
	
	entity.rendererLayer = "entities"
	entity.image = Images["digger_driver_1"]	-- static image for ... ?
	entity.imageScale = 1
	
	-- animations
	entity.animations = {}
	entity.animations.idleEmpty = Animation.new( "animations/digger/idle_empty.lua" )
	entity.animations.idleEmpty:setRelativeOrigin( 0.5, 0.7 )
	entity.animations.idleDriver = Animation.new( "animations/digger/idle_driver.lua" )
	entity.animations.idleDriver:setRelativeOrigin( 0.5, 0.7 )
	entity.animations.drill = Animation.new( "animations/digger/drill.lua" )
	entity.animations.drill:setRelativeOrigin( 0.5, 0.7 )
	
	-- set initial animation set
	entity.currentAnimation = "idleEmpty"
	
	-- 2d hitbox for mouse clicks
	entity.imageWidth = 80
	entity.imageHeight = 80
	entity.outlineOffsetY = 16
	
	----------------------------------------------------------------------------------------------------
	
	function entity:load( id )
		self.id = id
		
		self.health = Health:create( self, maxHealth, false, 0 )	-- health component
		Gameloop:add( self )
		Renderer:add( self, self.rendererLayer )
		
		self.ohwSelection = OhwEntitySelection:create{
			entityId = self.id,
			width = self.imageWidth,
			height = self.imageHeight,
			offsetY = self.outlineOffsetY,
		}
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- in order to check if a mouse click was issued to select this entity, use the 2d screen coordinates
	function entity:get2dHitbox()
		local x, y = convertToDimetric( self.body.x, self.body.y )
		local x, y = Camera:getTransform():transformPoint( x, y )
		local cameraScale = Camera:getScale()
		return {
			shape = "rect",
			x = Tools.round(x),
			-- do not forget the camera scale here!
			y = Tools.round(y) - self.outlineOffsetY / cameraScale,
			sizeX = self.imageWidth / cameraScale,
			sizeY = self.imageHeight / cameraScale,
		}
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- from game
	-- check if this entity can be given a move command at the moment
	function entity:canIssueMoveCommand()
		if not self:hasDriver() then
			return false
		end
		if self.task == "leaveBuilding" then
			return false
		end
		
		-- cannot start a move command when in a blocked area
		local x, y = self:getPosition()
		local mini_xi, mini_yi = getMiniTileIndex( x, y )
		if Game.blocking:isMiniTileBlocked( mini_xi, mini_yi ) then
			return false
		end
		
		return true
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- get the currently possible movement speed
	-- TODO: can be increased with technologies?
	function entity:getCurrentMovementSpeed()
		local speed = self.movementSpeed
		
		return speed
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- from extern: move to a specific location (from player move command)
	-- mini tile coordinates and world position
	function entity:moveToDestination( mini_xi, mini_yi, x, y )
		local ex, ey = self:getPosition()
		local start_txi, start_tyi = getTileIndex( ex, ey )
		local target_xi, target_yi = getTileIndex( x, y )
		
		-- extra: pre-check if there can be a path at all
		if not Pathfinding.getPath( start_txi, start_tyi, target_xi, target_yi, "ground_coarse" ) then
			print( "digger:moveToDestination - no path found! (1)" )
			return false
		end
		
		local start_xi, start_yi = getMiniTileIndex( ex, ey )
		local path = Pathfinding.getPath( start_xi, start_yi, mini_xi, mini_yi, "ground_fine" )
		if path then
			
			-- if canMove then
				
				-- new movement
				self:_setMovementDestination( mini_xi, mini_yi, x, y )
				self:setMovementPath( path )
				self:setTask( "moveTo" )
				
			-- end
			
		else
			print( "digger:moveToDestination - no path found! (2)" )
		end
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- check if this worker can take on a given job right now
	function entity:checkReadyForJob( job )
		local jobType = job.type
		if not self:canPerformJobType( jobType ) then
			return false
		end
		return self.task == "idle"
	end
	
	-- can only drill wall
	function entity:canPerformJobType( jobType )
		if not self:hasDriver() then
			return false
		end
		return jobType == "drillWall"
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:assignJob( job )
		
		self:cancelCurrentJob()
		
		if job.type ~= "drillWall" then
			print( "digger can only perform drill wall jobs ..." )
			return
		end
		
		-- print( "digger: got new job", job.type )
		self.currentJob = job
		
		local wall = EntityManager:getEntityById( job.entityId )
		if not wall then
			print( "digger: invalid wall to drill!" )
			self.currentJob = nil
			return false
		end
		
		-- move to a neighbour tile! -> search for a free mini tile index
		local ex, ey = self:getPosition()
		local start_xi, start_yi = getMiniTileIndex( ex, ey )
		-- job.targetTileX, job.targetTileY
		-- parameters: full tile index of the primary tile and full tile index of the neighbour
		local target_xi, target_yi = Game:getFreeMiniTileNextToFullTile( wall.xi, wall.yi, job.targetTileX, job.targetTileY )
		local target_x, target_y = miniTileIndexToPosition( target_xi, target_yi )
		local path = Pathfinding.getPath( start_xi, start_yi, target_xi, target_yi, "ground_fine" )
		if path then
			self:_setMovementDestination( target_xi, target_yi, target_x, target_y )
			self:setMovementPath( path )
			-- print( "digger: found path to job!" )
			-- print( "start:", start_xi, start_yi )
			-- print( "target:", target_xi, target_yi )
			self:setTask( "moveTo" )
		else
			print( "digger: no path to drill job!" )
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:jobDone()
		-- print( "digger: job is done", self.currentJob.type )
		self.currentJob = nil
		
		self:setTask( "idle" )
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:cancelCurrentJob()
		if not self.currentJob then
			return false
		end
		
		-- do not list jobs again that have been assigned to this worker manually by the player
		if not self.currentJob.manuallyAssigned then
			Game:entityCancelJob( self, self.currentJob )
		end
		self.currentJob = nil
		self:setTask( "idle" )
		-- print( "digger: cancel job" )
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:setOrientation( angle )
		self.orientation = angle
		-- note: for some reason the angle sometimes gets larger than 2*pi instead of being negative
		if self.orientation > pipi then
			self.orientation = self.orientation - pipi
		end
		
		local currName = self.animations[self.currentAnimation]:getCurrentState()
		local animName = self:getAnimationNameByOrientation( self.orientation )
		if currName ~= animName then
			self.animations[self.currentAnimation]:setState( animName )
		end
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:tick( dt )
		
		if self.selected then
			self.selectionAnimationTick = self.selectionAnimationTick + dt
		end
		
		self.timeSinceLastHealthChange = self.timeSinceLastHealthChange + dt
		
		
		-- current job
		if self.currentJob then
			if Game:isJobDone( self.currentJob ) then
				self:jobDone()
			end
		end
		
		
		-- behavior depending on the internal state
		self:behaviorTick( dt )
		
		
		-- animation
		if self.currentAnimation and self.animations[self.currentAnimation] then
			local anim = self.animations[self.currentAnimation]
			anim:update( dt )
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:behaviorTick( dt )
		
		local job = self.currentJob
		
		if self.task == "moveTo" then
			
			-- currently want to move somewhere
			if self.movementPath then
				local nSteps = #self.movementPath
				if nSteps > 0 then
					local ex, ey = self:getPosition()
					local start_xi, start_yi = getMiniTileIndex( ex, ey )
					if start_xi == self.movementPath[nSteps].x
					and start_yi == self.movementPath[nSteps].y then
						table.remove( self.movementPath, nSteps )
					else
						-- move to the next mini tile center
						local goal_x, goal_y = miniTileIndexToPosition( self.movementPath[nSteps].x, self.movementPath[nSteps].y )
						local direction = Tools.vectorSub( self.body, { x = goal_x, y = goal_y } )
						local vec2 = Tools.vectorSetLength( direction, self:getCurrentMovementSpeed() )
						self:move( vec2.x * dt, vec2.y * dt )
					end
				else
					-- already on the final mini tile
					-- move to the exact target location
					local direction = Tools.vectorSub( self.body, self.movementDestination )
					local vec2 = Tools.vectorTrim( direction, self:getCurrentMovementSpeed() * dt )
					self:move( vec2.x, vec2.y )
					
					if math.abs( self.body.x - self.movementDestination.x ) < 2
					and math.abs( self.body.y - self.movementDestination.y ) < 2 then
						self.movementPath = nil
						self:onMovementDestinationReached()
					end
				end
			end
			
		elseif self.task == "leaveBuilding" then
			
			-- entity is likely still in the blocking area of the building
			-- but that why there is a different task for this
			local direction = Tools.vectorSub( self.body, self.leaveBuildingPosition )
			local vec2 = Tools.vectorTrim( direction, self:getCurrentMovementSpeed() * dt )
			self:move( vec2.x, vec2.y )
			
			-- if position reached, set task to idle
			if math.abs( self.body.x - self.leaveBuildingPosition.x ) < 2
			and math.abs( self.body.y - self.leaveBuildingPosition.y ) < 2 then
				self.leaveBuildingPosition = nil
				self:setTask( "idle" )
			end
			
		elseif self.task == "drillWall" then
			
			local wall = EntityManager:getEntityById( job.entityId )
			if wall then
				Game:entityDrillWall( self, wall, dt )
			end
			
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- set animation etc.
	function entity:setTask( task )
		
		-- already doing that?
		if self.task == task then
			return
		end
		
		-- update task
		self.task = task
		-- print( "digger: set task:", task )
		
		if task == "drillWall" then
			self.currentAnimation = "drill"
		else
			if self:hasDriver() then
				self.currentAnimation = "idleDriver"
			else
				self.currentAnimation = "idleEmpty"
			end
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:onMovementDestinationReached()
		
		-- make sure the orientation when stopping is aligned with x or y axis (see "light" in draw)
		self.orientation = Tools.roundTo( self.orientation, 0.5 * pi )
		-- when going to idle, keep the last orientation
		self:setOrientation( self.orientation )
		
		local job = self.currentJob
		
		if job and job.type == "drillWall" then
			self:setTask( "drillWall" )
			-- look at the wall!
			local wall = EntityManager:getEntityById( job.entityId )
			if wall then
				local ex, ey = self:getPosition()
				local wx, wy = wall:getPosition()
				self:setOrientation( Tools.getAngleFromVector{ x = wx-ex, y = wy-ey } )
			end
		else
			self:setTask( "idle" )
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:draw()
		local body = self.body
		local x, y = convertToDimetric( body.x, body.y )
		-- local imgData = self.image
		-- local ox = imgData.ox or 0
		-- local oy = imgData.oy or 0
		-- local scale = self.imageScale
		
		-- animation
		if self.currentAnimation and self.animations[self.currentAnimation] then
			local anim = self.animations[self.currentAnimation]
			anim:setPosition( x, y )
			anim:draw( 0.8, 0.8 )
			-- anim:draw()
		end
		
		-- draw "light"
		if self:hasDriver() then
			local angle = self.orientation + 0.25 * pi
			local liSize = 600
			local alpha = 0.8
			local color = { 1, 0.7, 0.7, alpha }
			local offset = 4	-- the light appears to come from the front, not the center
			local vec2 = Tools.vectorSetLength( Tools.getVectorFromAngle( angle ), offset )
			Renderer:drawConeLight( x + vec2.x, y + vec2.y, liSize, angle, color )
		end
		
		self.ohwSelection:setVisible( self.selected == true )
		
		if Game.drawEntityStatusOverhead or self.selected
		or self.timeSinceLastHealthChange < 4 then
			local hb_y_offset = body.z + 64	-- draw above the entity
			self.health:draw( x, y - hb_y_offset )
		end
		
		-- lg.circle( "fill", x, y, 6 )	-- debug
		
		-- 2d hitbox (mouse click)
		-- lg.setColor( 1, 1, 1, 0.3 )
		-- lg.rectangle( "line", x - self.imageWidth * 0.5, y - self.imageHeight * 0.5 - self.outlineOffsetY, self.imageWidth, self.imageHeight )
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:destroy()
		Gameloop:remove( self )
		Renderer:remove( self, self.rendererLayer )
		self.ohwSelection:destroy()
		self = nil
	end
	
	
	return entity
end


return Digger
