local TransportTruck = {}
local lg = love.graphics
local pi = math.pi
local pipi = 2 * pi

local maxHealth = 500

-- TODO: when moving to a tool store to pick up resources for a construction site
-- the tt may already carry and keep on carrying some resources that are not needed

-- behavior:
-- can take multiple jobs at once
-- collect one item as normal (RR)
-- after collecting it...
-- ... if full -> move to deposit (or construction site)
-- ... not full -> wait for a moment to take a new job
-- ... no new job -> move to deposit
-- do not cancel and overwrite the current job when a new one is assigned!

function TransportTruck:create( data )
	local entity = Vehicle:create( data )
	
	entity.entityType = ENTITIES.transportTruck
	
	-- this entity can perform worker jobs
	entity.isWorker = true
	
	-- 25/s = one mini tile per second
	entity.movementSpeed = 100
	
	-- range in full tiles
	entity.explorationRange = 4
	
	-- current task
	-- idle
	-- moveTo	-- position from player command or job
	-- teleport? down/up?
	entity.task = "idle"
	
	-- cargo
	-- can hold multiple items of different types at the same time
	-- supported entities: ore, energyCrystal
	entity.cargo = {}
	entity.currentCargoAmount = 0
	entity.maximumCargoAmount = 3
	
	-- can accept multiple jobs at a time (same number as maximum cargo amount)
	entity.currentJobs = {}
	
	-- entity should wait a moment after picking up an entity if the cargo is not full
	-- maybe the game assigns a new collect resource job during this wait time
	-- after this time or when the cargo hold is full, move to a construction site or deposit
	entity.waitForNextJob = 0
	
	entity.rendererLayer = "entities"
	-- entity.imageScale = 1
	
	-- animations
	entity.animations = {}
	entity.animations.idleEmpty = Animation.new( "animations/transport truck/empty.lua" )
	entity.animations.idleEmpty:setRelativeOrigin( 0.5, 0.7 )
	entity.animations.idleDriver = Animation.new( "animations/transport truck/idle_driver.lua" )
	entity.animations.idleDriver:setRelativeOrigin( 0.5, 0.7 )
	entity.animations.driving = Animation.new( "animations/transport truck/driving.lua" )
	entity.animations.driving:setRelativeOrigin( 0.5, 0.7 )
	
	-- set initial animation set
	entity.currentAnimation = "idleEmpty"
	
	-- 2d hitbox for mouse clicks
	entity.imageWidth = 80
	entity.imageHeight = 80
	entity.outlineOffsetY = 16
	
	----------------------------------------------------------------------------------------------------
	
	function entity:load( id )
		self.id = id
		
		self.health = Health:create( self, maxHealth, false, 0 )	-- health component
		Gameloop:add( self )
		Renderer:add( self, self.rendererLayer )
		
		self.ohwSelection = OhwEntitySelection:create{
			entityId = self.id,
			width = self.imageWidth,
			height = self.imageHeight,
			offsetY = self.outlineOffsetY,
		}
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- in order to check if a mouse click was issued to select this entity, use the 2d screen coordinates
	function entity:get2dHitbox()
		local x, y = convertToDimetric( self.body.x, self.body.y )
		local x, y = Camera:getTransform():transformPoint( x, y )
		local cameraScale = Camera:getScale()
		return {
			shape = "rect",
			x = Tools.round(x),
			-- do not forget the camera scale here!
			y = Tools.round(y) - self.outlineOffsetY / cameraScale,
			sizeX = self.imageWidth / cameraScale,
			sizeY = self.imageHeight / cameraScale,
		}
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- from game
	-- check if this entity can be given a move command at the moment
	function entity:canIssueMoveCommand()
		if not self:hasDriver() then
			return false
		end
		if self.task == "leaveBuilding" then
			return false
		end
		
		-- cannot start a move command when in a blocked area
		local x, y = self:getPosition()
		local mini_xi, mini_yi = getMiniTileIndex( x, y )
		if Game.blocking:isMiniTileBlocked( mini_xi, mini_yi ) then
			return false
		end
		
		return true
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- get the currently possible movement speed
	-- TODO: can be increased with technologies?
	function entity:getCurrentMovementSpeed()
		local speed = self.movementSpeed
		
		return speed
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- from extern: move to a specific location (from player move command)
	-- mini tile coordinates and world position
	function entity:moveToDestination( mini_xi, mini_yi, x, y )
		local ex, ey = self:getPosition()
		local start_txi, start_tyi = getTileIndex( ex, ey )
		local target_xi, target_yi = getTileIndex( x, y )
		
		-- extra: pre-check if there can be a path at all
		if not Pathfinding.getPath( start_txi, start_tyi, target_xi, target_yi, "ground_coarse" ) then
			print( "transport truck:moveToDestination - no path found! (1)" )
			return false
		end
		
		local start_xi, start_yi = getMiniTileIndex( ex, ey )
		local path = Pathfinding.getPath( start_xi, start_yi, mini_xi, mini_yi, "ground_fine" )
		if path then
			
			-- if canMove then
				
				-- new movement
				self:_setMovementDestination( mini_xi, mini_yi, x, y )
				self:setMovementPath( path )
				self.targetDeposit = nil	-- reset current target deposit (if any!)
				self:setTask( "moveTo" )
				
				self:dropAllCarriedEntities()
				self:cancelCurrentJob()
			-- end
			
		else
			print( "transport truck:moveToDestination - no path found! (2)" )
		end
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- check if this worker can take on a job at all right now
	function entity:checkReadyForJob( job )
		local jobType = job.type
		if not self:canPerformJobType( jobType ) then
			return false
		end
		
		if jobType == "collectEnergyCrystal" or jobType == "collectOre" then
			return self.task == "idle" and not self:isCargoHoldFull()
		elseif jobType == "bringResourceToBuilding" then
			-- must have at least one resource of the requested type
			-- that is not already destined for a construction site
			for i = 1, #self.cargo do
				local item = EntityManager:getEntityById( self.cargo[i] )
				if item.entityType == job.resourceType
				and not item.destinationId then
					return true
				end
			end
			return false
		end
	end
	
	-- can only collect resources and bring them to construction sites
	function entity:canPerformJobType( jobType )
		if not self:hasDriver() then
			return false
		end
		return jobType == "collectEnergyCrystal" or jobType == "collectOre"
			or jobType == "bringResourceToBuilding"
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:assignJob( job )
		
		-- collect new resource entity OR bring a carried item to a construction site
		local isCollectNewResourceJob = job.type == "collectEnergyCrystal" or job.type == "collectOre"
		
		if self:isCargoHoldFull() and isCollectNewResourceJob then
			print( "transport truck: cannot accept another job! already full." )
			return false
		end
		
		-- print( "transport truck: got new job", job.type )
		
		-- can collect resources
		if isCollectNewResourceJob then
			local resourceEntity = EntityManager:getEntityById( job.entityId )
			if not resourceEntity then
				print( "transport truck: invalid resourceEntity to collect!" )
				-- self.currentJob = nil
				return false
			end
			
			local ex, ey = self:getPosition()
			local start_xi, start_yi = getMiniTileIndex( ex, ey )
			local target_x, target_y = resourceEntity:getPosition()
			local target_xi, target_yi = getMiniTileIndex( target_x, target_y )
			local path = Pathfinding.getPath( start_xi, start_yi, target_xi, target_yi, "ground_fine" )
			if path then
				self:_setMovementDestination( target_xi, target_yi, target_x, target_y )
				self:setMovementPath( path )
				-- print( "transport truck: clear targetDeposit!" )
				self.targetDeposit = nil
				self:setTask( "moveTo" )
			else
				print( "transport truck: no path to resourceEntity job!" )
			end
			
			self.waitForNextJob = 0
			
		else
			
			-- bringResourceToBuilding job
			local constructionSite = EntityManager:getEntityById( job.entityId )
			if not constructionSite then
				print( "transport truck: invalid constructionSite to bring resource to!" )
				return false
			end
			for i = 1, #self.cargo do
				local item = EntityManager:getEntityById( self.cargo[i] )
				if item.entityType == job.resourceType and not item.destinationId then
					item.destinationId = constructionSite.id
					-- print( "transport truck: carried item will be brought to construction site" )
					if not self.targetDeposit then
						-- print( "transport truck: set new targetDeposit (construction site)" )
						self.targetDeposit = constructionSite
					end
					break
				end
			end
			
		end
		
		-- only add this job if not escaped with false...
		table.insert( self.currentJobs, job )
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- cancel all (?)
	function entity:cancelCurrentJob()
		if #self.currentJobs == 0 then
			return false
		end
		
		for i = #self.currentJobs, 1, -1 do
			-- do not list jobs again that have been assigned to this worker manually by the player
			if not self.currentJobs[i].manuallyAssigned then
				Game:entityCancelJob( self, self.currentJobs[i] )
			end
			table.remove( self.currentJobs, i )
		end
		-- self:setTask( "idle" )
		-- print( "transport truck: cancel all jobs" )
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:setOrientation( angle )
		self.orientation = angle
		-- note: for some reason the angle sometimes gets larger than 2*pi instead of being negative
		if self.orientation > pipi then
			self.orientation = self.orientation - pipi
		end
		-- print( "transport truck: orientation is", angle )
		
			local currName = self.animations[self.currentAnimation]:getCurrentState()
			local animName = self:getAnimationNameByOrientation( self.orientation )
			if currName ~= animName then
				-- print( "currName:", currName )
				-- print( "should be:", animName )
				-- print( "angle:", self.orientation )
				self.animations[self.currentAnimation]:setState( animName )
			end
		-- end
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:tick( dt )
		
		if self.selected then
			self.selectionAnimationTick = self.selectionAnimationTick + dt
		end
		
		self.timeSinceLastHealthChange = self.timeSinceLastHealthChange + dt
		
		
		-- current job
		for i = #self.currentJobs, 1, -1 do
			if Game:isJobDone( self.currentJobs[i] ) then
				-- print( "transport truck: job is done", self.currentJobs[i].type )
				table.remove( self.currentJobs, i )
			end
		end
		
		
		-- behavior depending on the internal state
		self:behaviorTick( dt )
		
		
		-- animation
		if self.currentAnimation and self.animations[self.currentAnimation] then
			local anim = self.animations[self.currentAnimation]
			anim:update( dt )
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:behaviorTick( dt )
		
		local ex, ey = self:getPosition()
		
		if self.task == "idle" then
			
			if self.waitForNextJob > 0 then
				self.waitForNextJob = self.waitForNextJob - dt
				
				if self.waitForNextJob <= 0 then
					self.waitForNextJob = 0
					
					-- got no new job during this wait time?
					
					-- should bring to construction site first?
					local moveToConstructionSite = false
					
					for i = 1, #self.cargo do
						local item = EntityManager:getEntityById( self.cargo[i] )
						if item.destinationId then
							local deposit = EntityManager:getEntityById( item.destinationId )
							local start_xi, start_yi = getMiniTileIndex( ex, ey )
							local target_x, target_y = deposit:getPosition()
							local target_xi, target_yi = getMiniTileIndex( target_x, target_y )
							local path = Pathfinding.getPath( start_xi, start_yi, target_xi, target_yi, "ground_fine" )
							if path then
								self:_setMovementDestination( target_xi, target_yi, target_x, target_y )
								self:setMovementPath( path )
								-- print( "transport truck: bring resource to construction site (deposit)" )
								self.targetDeposit = deposit
								self:setTask( "moveTo" )
								
								moveToConstructionSite = true
								break
							end
						end
					end
					
					if not moveToConstructionSite then
						for i = 1, #self.cargo do
							local item = EntityManager:getEntityById( self.cargo[i] )
							-- search for the closest place to deposit the resource
							local deposit = Game:getClosestDeposit( self, item )
							if deposit then
								-- local ex, ey = self:getPosition()
								local start_xi, start_yi = getMiniTileIndex( ex, ey )
								local target_x, target_y = deposit:getResourceDepositPosition()
								local target_xi, target_yi = getMiniTileIndex( target_x, target_y )
								local path = Pathfinding.getPath( start_xi, start_yi, target_xi, target_yi, "ground_fine" )
								if path then
									self:_setMovementDestination( target_xi, target_yi, target_x, target_y )
									self:setMovementPath( path )
									-- print( "transport truck: move to deposit" )
									self.targetDeposit = deposit
									self:setTask( "moveTo" )
									break
								end
							end
						end
					end
					
				end
			end
			
		elseif self.task == "moveTo" then
			
			-- currently want to move somewhere
			if self.movementPath then
				local nSteps = #self.movementPath
				if nSteps > 0 then
					local start_xi, start_yi = getMiniTileIndex( ex, ey )
					if start_xi == self.movementPath[nSteps].x
					and start_yi == self.movementPath[nSteps].y then
						table.remove( self.movementPath, nSteps )
					else
						-- move to the next mini tile center
						local goal_x, goal_y = miniTileIndexToPosition( self.movementPath[nSteps].x, self.movementPath[nSteps].y )
						local direction = Tools.vectorSub( self.body, { x = goal_x, y = goal_y } )
						local vec2 = Tools.vectorSetLength( direction, self:getCurrentMovementSpeed() )
						self:move( vec2.x * dt, vec2.y * dt )
						self:setOrientation( Tools.getAngleFromVector( vec2 ) )
					end
				else
					-- already on the final mini tile
					-- move to the exact target location
					local direction = Tools.vectorSub( self.body, self.movementDestination )
					local vec2 = Tools.vectorTrim( direction, self:getCurrentMovementSpeed() * dt )
					self:move( vec2.x, vec2.y )
					self:setOrientation( Tools.getAngleFromVector( vec2 ) )
					
					if math.abs( self.body.x - self.movementDestination.x ) < 2
					and math.abs( self.body.y - self.movementDestination.y ) < 2 then
						self.movementPath = nil
						self:onMovementDestinationReached()
					end
				end
			end
			
		elseif self.task == "leaveBuilding" then
			
			-- entity is likely still in the blocking area of the building
			-- but that why there is a different task for this
			local direction = Tools.vectorSub( self.body, self.leaveBuildingPosition )
			local vec2 = Tools.vectorTrim( direction, self:getCurrentMovementSpeed() * dt )
			self:move( vec2.x, vec2.y )
			self:setOrientation( Tools.getAngleFromVector( vec2 ) )
			
			-- if position reached, set task to idle
			if math.abs( self.body.x - self.leaveBuildingPosition.x ) < 2
			and math.abs( self.body.y - self.leaveBuildingPosition.y ) < 2 then
				self.leaveBuildingPosition = nil
				self:setTask( "idle" )
			end
			
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- set animation etc.
	function entity:setTask( task )
		
		-- already doing that?
		if self.task == task then
			return
		end
		
		-- update task
		self.task = task
		-- print( "transport truck: set task:", task )
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- on destination reached...
	-- has targetDeposit?
	--> it is a constructionSite? --> deposit all items that belong there
	--> it is a tool store (or other deposit) --> deposit all items without a destinationId
	-- (this implies that a manual player move command needs to clear any target deposit entry)
	
	function entity:onMovementDestinationReached()
		
		-- update animation
		self.currentAnimation = "idleDriver"
		-- make sure the orientation when stopping is aligned with x or y axis (see "light" in draw)
		self.orientation = Tools.roundTo( self.orientation, 0.5 * pi )
		-- when going to idle, keep the last orientation
		self:setOrientation( self.orientation )
		
		-- print( "transport truck reached destination..." )
		-- print( "#cargo:", #self.cargo )
		-- print( "#jobs:", #self.currentJobs )
		
		local ex, ey = self:getPosition()
		
		-- deposit something
		if self.targetDeposit then
			
			-- can be construction site or permanent deposit (e.g. tool store)
			
			if self.targetDeposit.entityType == ENTITIES.constructionSite then
				
				-- deposit all items that belong there
				for i = #self.cargo, 1, -1 do
					local item = EntityManager:getEntityById( self.cargo[i] )
					if item.destinationId and item.destinationId == self.targetDeposit.id then
						self:depositCarriedEntity( item, self.targetDeposit )
						-- print( "transport truck: put item to construction site", item.id, item.destinationId )
					end
				end
				
			else
				
				-- deposit all items without a destinationId
				for i = #self.cargo, 1, -1 do
					local item = EntityManager:getEntityById( self.cargo[i] )
					if not item.destinationId then
						self:depositCarriedEntity( item, self.targetDeposit )
						-- print( "transport truck: put item to deposit", item.id, self.targetDeposit.id )
					end
				end
				
			end
			
			self.targetDeposit = nil
			self:setTask( "idle" )
			
		else
			
			-- no deposit
			-- might just arrived from move command
			--> pick up an entity or do nothing
			
			for i, job in ipairs( self.currentJobs ) do
				if job.type == "collectOre"
				or job.type == "collectEnergyCrystal" then
					-- is the item still there and can be picked up?
					local item = EntityManager:getEntityById( job.entityId )
					if item then
						-- check position
						local item_x, item_y = item:getPosition()
						if Tools.getDistance2( ex, ey, item_x, item_y ) <= 50 then
							-- check that the item is not already carried by another entity...
							if item:canBePickedUp() then
								if self:pickUpEntity( item ) then
									item:setCarrierEntity( self )
									Game:onEntityPickedUpEntity( self, item )
								end
							else
								-- print( "transport truck: item cannot be picked up now ..." )
							end
						else
							-- print( "transport truck: item to pick up is too far away ..." )
						end
					end
				end
			end
			
			-- wait for X seconds before a new job can be assigned automatically
			if self:isCargoHoldFull() then
				-- must be > 0
				self.waitForNextJob = 0.5
			else
				self.waitForNextJob = 2
			end
			self:setTask( "idle" )
			
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:isCargoHoldFull()
		return self.currentCargoAmount >= self.maximumCargoAmount
	end
	
	
	function entity:pickUpEntity( entity )
		if self:isCargoHoldFull() then
			return false
		end
		
		for i = 1, self.maximumCargoAmount do
			if not self.cargo[i] then
				self.cargo[i] = entity.id
				self.currentCargoAmount = self.currentCargoAmount + 1
				-- print( "transport truck: pick up entity", entity.id )
				return true
			end
		end
	end
	
	
	function entity:depositCarriedEntity( entity, deposit )
		if deposit.entityType == ENTITIES.constructionSite then
			Game:addResourceEntityToConstructionSite( entity, deposit )
		else
			Game:depositEntity( entity, deposit )
		end
		self:dropCarriedEntity( entity )
	end
	
	
	function entity:dropCarriedEntity( entity )
		for i = 1, self.maximumCargoAmount do
			if self.cargo[i] == entity.id then
				entity:setCarrierEntity( nil )
				self.cargo[i] = nil
				self.currentCargoAmount = self.currentCargoAmount - 1
				-- print( "transport truck: drop carried entity", entity.id )
				return entity
			end
		end
		return false
	end
	
	
	function entity:dropAllCarriedEntities()
		for i = self.maximumCargoAmount, 1, -1 do
			if self.cargo[i] then
				local entity = EntityManager:getEntityById( self.cargo[i] )
				entity:setCarrierEntity( nil )
				self.cargo[i] = nil
			end
		end
		-- print( "transport truck: drop all carried entities" )
		self.currentCargoAmount = 0
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- these values are in screen coordinates!
	function entity:getCarriedEntityPositionOffset( entity )
		local anim = self.animations[self.currentAnimation]
		local state = anim:getCurrentState()
		if state == "facing_sw" then	-- sw
			return 20, -30
		elseif state == "facing_ne" then	-- ne
			return -20, -10
		elseif state == "facing_nw" then	-- nw
			return 20, -10
		elseif state == "facing_se" then
			return -20, -30
		end
		
		-- if a >= 0.25*pi and a < 0.75*pi then	-- sw
			-- return 20, -20
		-- elseif a < -0.25*pi and a >= -0.75*pi then	-- ne
			-- return -20, -20
		-- elseif math.abs( a ) >= 0.75*pi then	-- nw
			-- return 20, -20
		-- else	-- se
			-- return -30, -20
		-- end
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:draw()
		local body = self.body
		local x, y = convertToDimetric( body.x, body.y )
		-- local imgData = self.image
		-- local ox = imgData.ox or 0
		-- local oy = imgData.oy or 0
		-- local scale = self.imageScale
		
		-- animation
		if self.currentAnimation and self.animations[self.currentAnimation] then
			local anim = self.animations[self.currentAnimation]
			anim:setPosition( x, y )
			anim:draw( 0.8, 0.8 )
		end
		
		-- draw "light"
		if self:hasDriver() then
			-- use the current orientation = light seems to rotate with the vehicle
			-- or use 4 discrete angles from animation for quick transitions
			-- TODO: the orientation must be updated when the movementDestination is reached!
			local angle = self.orientation + 0.25 * pi
			-- local anim = self.animations[self.currentAnimation]
			-- local state = anim:getCurrentState()
			-- if state == "facing_se" then
				-- angle = 0.25 * pi
			-- elseif state == "facing_sw" then
				-- angle = 0.75 * pi
			-- elseif state == "facing_nw" then
				-- angle = 1.25 * pi
			-- elseif state == "facing_ne" then
				-- angle = 1.75 * pi
			-- end
			local liSize = 600
			local alpha = 0.8
			local color = { 1, 0.7, 0.7, alpha }
			local offset = 22	-- the light appears to come from the front, not the center
			local vec2 = Tools.vectorSetLength( Tools.getVectorFromAngle( angle ), offset )
			Renderer:drawConeLight( x + vec2.x, y + vec2.y, liSize, angle, color )
		end
		
		self.ohwSelection:setVisible( self.selected == true )
		
		if Game.drawEntityStatusOverhead or self.selected
		or self.timeSinceLastHealthChange < 4 then
			local hb_y_offset = body.z + 64	-- draw above the entity
			self.health:draw( x, y - hb_y_offset )
		end
		
		-- lg.circle( "line", x, y, 6 )	-- debug
		-- local dx, dy, dz = self:getCarriedEntityPositionOffset()
		-- lg.circle( "fill", x+dx, y+dy, 6 )
		
		-- 2d hitbox (mouse click)
		-- lg.setColor( 1, 1, 1, 0.3 )
		-- lg.rectangle( "line", x - self.imageWidth * 0.5, y - self.imageHeight * 0.5 - self.outlineOffsetY, self.imageWidth, self.imageHeight )
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:destroy()
		self:dropAllCarriedEntities()
		
		Gameloop:remove( self )
		Renderer:remove( self, self.rendererLayer )
		self.ohwSelection:destroy()
		self = nil
	end
	
	
	return entity
end


return TransportTruck
