local Vehicle = {}

local pi = math.pi
local pipi = 2 * pi


function Vehicle:create( data )
	local entity = {}
	
	entity.entityCategory = ENTITYCATEGORY.vehicle
	
	-- by default, this entity can be controlled by the player
	-- deactivate e.g. for entities that should be rescued from a different cave
	entity.isPlayerControlled = true
	
	entity.body = {
		shape = "circle",
		x = data.x or 0,
		y = data.y or 0,
		radius = 4,
		z = 0,
	}
	entity.orientation = 0
	
	-- 25/s = one mini tile per second
	entity.movementSpeed = 125
	
	-- range in full tiles
	entity.explorationRange = 4
	
	-- world position and mini tile index
	entity.movementDestination = {}
	
	
	----------------------------------------------------------------------------------------------------
	
	-- set currently selected - true / false
	function entity:setSelected( selected )
		self.selected = selected
		self.selectionAnimationTick = 0
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:getPosition()
		return self.body.x, self.body.y
	end
	
	
	function entity:setPosition( x, y )
		self.body.x = x
		self.body.y = y
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- range in full tiles
	function entity:getExplorationRange()
		return self.explorationRange
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- from game
	-- check if this entity can be given a move command at the moment
	function entity:canIssueMoveCommand()
		if not self:hasDriver() then
			return false
		end
		if self.task == "leaveBuilding" then
			return false
		end
		
		return true
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- get the currently possible movement speed
	-- TODO: can be increased with technologies?
	function entity:getCurrentMovementSpeed()
		local speed = self.movementSpeed
		
		return speed
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:setDriver( driver )
		if driver then
			self.driverEntityId = driver.id
		else
			self.driverEntityId = nil
		end
		
		if self:getDriver() then
			self.currentAnimation = "idleDriver"
		else
			self.currentAnimation = "idleEmpty"
			
			if self.cancelCurrentJob then
				self:cancelCurrentJob()
			end
			
			self:stopMovement()
		end
		self:setOrientation( self.orientation )
	end
	
	
	function entity:getDriver()
		return EntityManager:getEntityById( self.driverEntityId )
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:hasDriver()
		if self:getDriver() then
			return true
		end
		return false
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:canBeManned()
		if self:hasDriver() then
			return false
		end
		
		return true
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:getLeaveVehiclePosition()
		local x, y = self:getPosition()
		return x, y + 20
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- internal only
	-- mini tile coordinates and world position
	function entity:_setMovementDestination( mini_xi, mini_yi, x, y )
		self.movementDestination.x = x
		self.movementDestination.y = y
		self.movementDestination.mtx = mini_xi
		self.movementDestination.mty = mini_yi
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- immediately stop the vehicle
	function entity:stopMovement()
		self.movementPath = nil
		self.movementDestination = {}
		self:setTask( "idle" )
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- set animation etc.
	function entity:setTask( task )
		
		-- already doing that?
		if self.task == task then
			return
		end
		
		-- update task
		self.task = task
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- called from game when a hurt event occured
	function entity:onHurt( damage )
		
	end
	
	
	function entity:onHealthChanged()
		self.timeSinceLastHealthChange = 0
		
		if self.health:getHealth() <= 0 then
			-- driver should leave, even if the terrain is blocked (e.g. tile turned to lava)
			if self:hasDriver() then
				Game:vehicleDriverLeave( self )
			end
			
			-- destroy entity
			-- the resources spent to create this entity will be lost
			EntityManager:remove( self )
		end
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:onMovementDestinationReached()
		
		-- make sure the orientation when stopping is aligned with x or y axis (see "light" in draw)
		self.orientation = Tools.roundTo( self.orientation, 0.5 * pi )
		-- when going to idle, keep the last orientation
		self:setOrientation( self.orientation )
		
		-- update task
		self:setTask( "idle" )
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:getAnimationNameByOrientation( angle )
		local a = angle
		if a >= 0.25*pi and a < 0.75*pi then
			return "facing_sw"
		elseif a < -0.25*pi and a >= -0.75*pi then
			return "facing_ne"
		elseif math.abs( a ) >= 0.75*pi then
			return "facing_nw"
		else
			return "facing_se"
		end
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:setMovementPath( path )
		self.movementPath = path
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:move( dx, dy )
		self.body.x = self.body.x + (dx or 0)
		self.body.y = self.body.y + (dy or 0)
		self:setOrientation( Tools.getAngleFromVector{ x = dx, y = dy } )
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- teleport down to building, leave the building (inside blocking)
	function entity:setLeaveBuildingState( x, y )
		self.leaveBuildingPosition = { x = x, y = y }
		self:setTask( "leaveBuilding" )
	end
	
	----------------------------------------------------------------------------------------------------
	
	return entity
end

return Vehicle
