local RockSample = {}
local lg = love.graphics

-- rendered in studio with zoom 0.2 and y offset 2

function RockSample:create( data )
	local entity = {}
	
	entity.entityType = ENTITIES.rockSample
	entity.entityCategory = ENTITYCATEGORY.resource
	
	entity.body = {
		shape = "circle",
		x = data.x or 0,
		y = data.y or 0,
		radius = 4,
	}
	
	entity.rendererLayer = "entities"
	-- randomly select one of X different (colored) images
	local i = math.random(5)
	entity.image = Images["rock_sample_"..i]
	entity.imageScale = 0.3
	
	-- 2d hitbox for mouse clicks
	entity.imageWidth = entity.image.image:getWidth() * entity.imageScale
	entity.imageHeight = entity.image.image:getHeight() * entity.imageScale
	entity.outlineOffsetY = 6	-- either calculate it from the image dimensions and offset or hard code a fixed value
	
	----------------------------------------------------------------------------------------------------
	
	function entity:load( id )
		self.id = id
		
		Gameloop:add( self )
		Renderer:add( self, self.rendererLayer )
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- in order to check if a mouse click was issued to select this entity, use the 2d screen coordinates
	function entity:get2dHitbox()
		local x, y = convertToDimetric( self.body.x, self.body.y )
		local x, y = Camera:getTransform():transformPoint( x, y )
		local cameraScale = Camera:getScale()
		return {
			shape = "rect",
			x = Tools.round(x),
			-- do not forget the camera scale here!
			y = Tools.round(y) - self.outlineOffsetY / cameraScale,
			sizeX = self.imageWidth / cameraScale,
			sizeY = self.imageHeight / cameraScale,
		}
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:getPosition()
		return self.body.x, self.body.y
	end
	
	
	function entity:setPosition( x, y )
		self.body.x = x
		self.body.y = y
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- check if this item can be picked up by a RR or another worker
	-- return false if already carried by an entity or already deposited in a building...
	function entity:canBePickedUp()
		if self.carrierEntity or self.depositEntity then
			return false
		end
		
		return true
	end
	
	-- worker (RR or transport truck)
	function entity:setCarrierEntity( entity )
		self.carrierEntity = entity
		if not entity then
			self.drawOffsetX = nil
			self.drawOffsetY = nil
		end
	end
	
	
	-- building (tool store)
	function entity:setDeposit( deposit )
		self.depositEntity = deposit
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:tick( dt )
		local carrier = self.carrierEntity
		if carrier then
			-- entity position in front of the entity for correct rendering order
			local x, y = carrier:getPosition()
			self:setPosition( x+1, y+1 )
			
			-- fetch correct draw offset depending on the carrier entity
			if carrier.getCarriedEntityPositionOffset then
				self.drawOffsetX, self.drawOffsetY = carrier:getCarriedEntityPositionOffset()
			end
		end
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:draw()
		if self.depositEntity then
			return
		end
		
		local body = self.body
		local x, y = convertToDimetric( body.x, body.y )
		local imgData = self.image
		local ox = imgData.ox or 0
		local oy = imgData.oy or 0
		local scale = self.imageScale
		
		-- when being carried by another entity
		if self.drawOffsetX then x = x + self.drawOffsetX end
		if self.drawOffsetY then y = y + self.drawOffsetY end
		
		lg.draw( imgData.image, x, y, 0, scale, scale, ox, oy )
		
		-- lg.circle( "fill", x, y, 2 )	-- debug
		
		-- 2d hitbox (mouse click)
		-- lg.setColor( 1, 1, 1, 0.2 )
		-- lg.rectangle( "line", x - self.imageWidth * 0.5, y - self.imageHeight * 0.5 - self.outlineOffsetY, self.imageWidth, self.imageHeight )
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:destroy()
		Gameloop:remove( self )
		Renderer:remove( self, self.rendererLayer )
		self = nil
	end
	
	
	return entity
end

return RockSample
