local Game = {}

local lg = love.graphics
local tilesize = TILESIZE
local borderSizeFactor = 0.04	-- for camera movement when mouse is near the screen border

--[[
JOBS
Global Perspective:
- one queue per job type (remove debris, drill wall, ...)
- ordered priorities - can be changed by the player ingame
- if a RR is free (idle) a new job is assigned, search the queue with the highes priority first
- player can manually assign a job to a RR, he will discard his previous job and take on the new task
Rock Raider Perspective:
- has a current job attached
- idle, walking to job or executing job
--]]

--[[
IDEAS FOR MISSION NAMES:
Ärger mit Abraum
Steinreich
Durch das Inferno
Lava-Labyrinth
Sackgasse
Insubordination
--]]


-- IDEAS:
-- idea for jobs: RR have an idle timer, only consider RR for jobs who have been idle for a second or two
--> a worker who just went to idle state can look for jobs directly next to him
-- carrier/deposit robot where rock raiders can deposit resources until it is full and returns to the base


local GAMESTATE_PLAYING = 1
local GAMESTATE_WON = 2
local GAMESTATE_FAILED = 3


local WallSanityChecks = {
	{-1,-1, 0,-1, -1,0},	-- top left
	{0,-1, 1,-1, 1,0},	-- top right
	{-1,0, -1,1, 0, 1},	-- bottom left
	{0,1, 1,1, 1,0}	-- bottom right
}


function Game:create()
	local game = {}
	
	game.id = "game"
	game.gametime = 0
	game.gamespeed = 1	-- factor -> do not change here; see setGameSpeed
	game.cameraScrollSpeed = 1	-- see guiActionSetScrollSpeed
	
	game.workerJobPriorities = {}
	
	game._debugPaused = false
	
	----------------------------------------------------------------------------------------------------
	
	-- store all jobs in a table so they can be stopped when the player leaves the game
	function game:trackJob( job )
		table.insert( self.simpleJobs, job )
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:open()
		-- GUImgr:setScreen( self.id )
		
		self.gametime = 0
		self.playerGlobalResources = {
			[ENTITIES.ore] = 0,
			[ENTITIES.energyCrystal] = 0,
			[ENTITIES.rockSample] = 0,
		}
		self.gamestate = GAMESTATE_PLAYING
		
		self:switchToSelectionState( "global" )
		
		self:setPaused( false )
		self:setGameSpeed( 1.5 )
		
		-- healthbar etc.
		-- drawn when a unit takes damage (or gets healed/repaired) or when the entity is selected
		-- or when this variable is true (for all qualified units)
		self.drawEntityStatusOverhead = false
		
		-- reference to minimap (GUI element)
		self.minimap = GUImgr:getObjectById( "gameMinimap" )
		
		-- aboid that the mouse leaves the window area while ingame
		-- love.mouse.setGrabbed( true )
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- start the game, load a test map
	function game:startTestMap()
		
		self:startLevel( "levels.missions.mission1" )
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- to be called from the menu
	function game:startLevel( levelPath )
		
		self.selectionState = "global"
		
		self.selectedEntities = {}	-- list of currently selected entities
		
		-- see trackJob
		self.simpleJobs = {}
		
		-- jobs for rock raiders (and or machines?) to perform
		-- actual jobs that wait to be worked on
		self.workerJobs = {}
		
		self.workerJobPriorities = {}
		self.workerJobPriorities["placeConstructionFence"] = 1
		self.workerJobPriorities["bringResourceToBuilding"] = 1
		self.workerJobPriorities["vehicleGetDriver"] = 2
		self.workerJobPriorities["drillWall"] = 3
		self.workerJobPriorities["clearDebris"] = 4
		self.workerJobPriorities["collectOre"] = 5
		self.workerJobPriorities["collectEnergyCrystal"] = 6
		self.workerJobPriorities["reinforceWall"] = 7
		-- not listed because not automatically assigned
		self.workerJobPriorities["collectRockSample"] = 8
		
		self:updatePriorityButtons()
		
		self.script = require( levelPath )
		local mapName = self.script.mapName
		if not mapName then
			print( "Game ERROR: missing map name!" )
			return
		end
		local mapData = MapLoader:loadMap( mapName )
		self.terrain = Terrain:create( mapData )
		
		-- mission description (briefing)
		GUImgr:getObjectById( "txtGameMissionBriefing" ):setText( self.script.description )
		GUImgr:getObjectById( "btnGameCloseMissionBriefing" ):setText( getString( "game_start_mission" ) )
		self:openMissionBriefingWindow()
		
		-- set map size
		self.mapSizeX = mapData.sizeX
		self.mapSizeY = mapData.sizeY
		
		self.minimap:setMapSize( self.mapSizeX, self.mapSizeY )
		
		self.energyCrystals = mapData.energyCrystals
		
		-- create walls from map data
		self.wallData = mapData.walls	-- store the walldata
		for yi = 1, mapData.sizeY do
			for xi = 1, mapData.sizeX do
				local wallType = self.wallData[yi][xi]
				if wallType > 0 then
					-- create a wall
					local wall = EntityManager:createEntity(
						wallType,
						-- also store the tile index (avoid to calculate them over and over again)
						{ x = (xi-1) * tilesize, y = (yi-1) * tilesize, xi = xi, yi = yi }
					)
					-- plus, store how many energy crystals can be found inside this wall
					local energyCrystals = self.energyCrystals[yi][xi]
					if energyCrystals > 0 then
						wall.numberOfEnergyCrystals = energyCrystals
					end
				end
			end
		end
		
		-- level generation is complete, now setup the blocking info
		self.blocking = Blocking:create()
		self.blocking:setSize( self.mapSizeX, self.mapSizeY )
		self.blocking:fullBlockingUpdate( self.terrain, self.wallData )
		
		-- fog of war:
		-- terrain tiles and entities are not visible if the area is not discovered yet
		-- terrain is always visible once discovered, entities are only visible if player entities are nearby
		self.fow = FogOfWar:create( self.mapSizeX, self.mapSizeY )
		
		-- after all walls have been created, their images must be changed depending on their neighbours
		-- note: if using "fake walls" for FOW - call this update only AFTER the fog of war has been created!
		local walls = EntityManager:getEntitiesInCategory( ENTITYCATEGORY.wall )
		self:updateWallImageCodes( walls )
		
		-- energy crystals can be inside walls or laying on the ground
		self:createResourcesFromMapData()
		
		-- important for all jobs: do nothing while game is paused!
		
		-- once per second: manage worker jobs
		self:trackJob( Simplejob:addJob{
			intervall = 1,
			callback = function()
				if GUImgr:getCurrentScreenId() ~= game.id then
					return true
				end
				if game.isPaused then return end
				
				game:manageWorkerJobAssignment()
			end,
		} )
		
		-- once per second: update fog of war
		self:trackJob( Simplejob:addJob{
			intervall = 1,
			callback = function()
				if GUImgr:getCurrentScreenId() ~= game.id then
					return true
				end
				if game.isPaused then return end
				
				game:updateFogOfWar()
			end,
		} )
		
		-- manage everything lava related
		self:trackJob( Simplejob:addJob{
			intervall = 0.1,
			callback = function()
				if GUImgr:getCurrentScreenId() ~= game.id then
					return true
				end
				if game.isPaused then return end
				
				game:updateLava( 0.1 )
			end,
		} )
		
		
		-- landslides
		self:trackJob( Simplejob:addJob{
			intervall = 3.2,
			callback = function()
				if GUImgr:getCurrentScreenId() ~= game.id then
					return true
				end
				if game.isPaused then return end
				
				game:manageLandslides()
			end,
		} )
		
		-- minimap (radar)
		--[[
		self:trackJob( Simplejob:addJob{
			intervall = 1,
			callback = function()
				if GUImgr:getCurrentScreenId() ~= game.id then
					return true
				end
				if game.isPaused then return end
				
				game:updateMinimap()
			end,
		} )
		--]]
		
		-- mob spawning
		self.mobSpawner = MobSpawner:create()
		
		self:initScriptCallbacks()
		
		-- everything is set up
		self.scriptCallbacks.onGameStart()
		
		-- also update the FOW once at game start
		self:updateFogOfWar()
		
		Logger:log( "Game started." )
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- called right after game start and from button callback
	function game:openMissionBriefingWindow()
		self:setPaused( true )
		GUImgr:setVisibility( "btnGameOpenMissionBriefing", false )
		GUImgr:setVisibility( "btnGameOpenPauseMenu", false )
		GUImgr:setVisibility( "widGameMissionBriefing", true )
		GUImgr:getObjectById( "txtGameMissionBriefing" ):resetOffset()	-- scroll back to the top
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:initScriptCallbacks()
		
		local script = self.script
		
		self.scriptCallbacks = {}
		
		self.scriptCallbacks.onGameStart = script.onGameStart or function() end
		
		self.scriptCallbacks.onEntityHurtEntity = script.onEntityHurtEntity or function() end
		
		self.scriptCallbacks.onWallDestroyed = script.onWallDestroyed or function() end
		
		self.scriptCallbacks.onBeforeDebrisRemoved = script.onBeforeDebrisRemoved or function() end
		
		-- TODO: add more callbacks
		-- e.g. when entities are created or hurt, when resources are discovered, ...
		--> create dummy callbacks here to avoid plenty of if guards in the game code
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- after map start: generate energy crystals that are lying around
	-- for crystals that are inside walls, they will be created when the wall is drilled
	-- note: store the info inside each wall how many crystals there are supposed to be
	-- to make it easier to implement "scanners" that search for resources in walls -> see wall generation above
	function game:createResourcesFromMapData()
		
		local energyCrystals = self.energyCrystals
		local wallData = self.wallData
		local spread = 25
		local random = math.random
		
		for yi = 1, self.mapSizeY do
			for xi = 1, self.mapSizeX do
				local wallType = wallData[yi][xi]
				-- only generate energy crystals if there is no wall
				if wallType == 0 then
					for i = 1, energyCrystals[yi][xi] do
						local energyCrystal = EntityManager:createEntity(
							ENTITIES.energyCrystal,
							{ x = (xi-1) * tilesize + random(-spread,spread), y = (yi-1) * tilesize + random(-spread,spread) }
						)
					end
				end
			end
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- ore has been found (e.g. from mining a wall) at a certain tile
	-- or an energy crystal (mining a wall or simply in a undiscovered cavern area)
	function game:onResourceDiscovered( entity, xi, yi )
		
		-- if not already queued: add a new collect resource job
		local eType = entity.entityType
		if eType == ENTITIES.energyCrystal then
			self:addCollectEnergyCrystalJob( entity )
			self:newNotificationMessage( getString( "game_notification_energy_crystal_discovered" ) )
		elseif eType == ENTITIES.ore then
			self:addCollectOreJob( entity )
			self:newNotificationMessage( getString( "game_notification_ore_discovered" ) )
		end
		
		-- TODO: + cooldown to avoid spam
		-- print( "Game: resource entity found!" )
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- e.g. a RR picked up an energy crystal or a construction fence
	function game:onEntityPickedUpEntity( carrier, entity )
		
		-- a worker picked up a resource entity -> check if there is a bringResourceToBuilding for this entity type
		if entity.entityCategory == ENTITYCATEGORY.resource
		and carrier.checkReadyForJob then
			for jobIndex, job in ipairs( self.workerJobs ) do
				if job.type == "bringResourceToBuilding" and job.resourceType == entity.entityType then
					local worker_x, worker_y = carrier:getPosition()
					local worker_tx, worker_ty = getTileIndex( worker_x, worker_y )
					local constructionSite = EntityManager:getEntityById( job.entityId )
					local constructionSite_x, constructionSite_y = constructionSite:getPosition()
					local target_tx, target_ty = getTileIndex( constructionSite_x, constructionSite_y )
					
					if carrier:checkReadyForJob( job )
					-- TODO: pathfinding type ground vs flying
					and Pathfinding.getPath( worker_tx, worker_ty, target_tx, target_ty, "ground_coarse" ) then
						carrier:assignJob( job )
						table.remove( self.workerJobs, jobIndex )
						break	-- only assign one job!
					end
				end
			end
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- search a place (entity) to deposit the resource entity for a given worker
	function game:getClosestDeposit( worker, resourceEntity )
		local worker_x, worker_y = worker:getPosition()
		local worker_tx, worker_ty = getTileIndex( worker_x, worker_y )
		local typeToDeposit = resourceEntity.entityType
		
		-- TODO: mobile deposit may be full -> cannot deposit more entities
		local targets = EntityManager:getEntitiesFilter( function(e)
			return e.isResourceDeposit and e:canDepositEntity( typeToDeposit )
		end )
		
		local bestPath
		local bestDeposit
		for _, deposit in pairs( targets ) do
			local x, y = deposit:getPosition()
			local target_tx, target_ty = getTileIndex( x, y )
			local path = Pathfinding.getPath( worker_tx, worker_ty, target_tx, target_ty, "ground_coarse" )
			if path and (not bestPath or #path < bestPath.length) then
				bestPath = {
					path = path,
					length = #path,
					-- targetTileX = target_tx,
					-- targetTileY = target_ty,
				}
				bestDeposit = deposit
			end
		end
		
		return bestDeposit	-- (or nil)
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- e.g. a worker brings a resource entity to a tool store
	function game:depositEntity( carriedEntity, deposit )
		if not deposit then
			print( "game:depositEntity - deposit does not exist!" )
			return
		end
		local carriedEntityCategory = carriedEntity.entityCategory
		local carriedEntityType = carriedEntity.entityType
		deposit:storeEntity( carriedEntity )
		carriedEntity:setDeposit( deposit )
		EntityManager:remove( carriedEntity )
		
		if carriedEntityCategory == ENTITYCATEGORY.resource then
			self:addToPlayersGlobalResource( carriedEntityType, 1 )
		end
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- a worker brings a resource entity to a construction site
	function game:addResourceEntityToConstructionSite( carriedEntity, constructionSite )
		local resourceType = carriedEntity.entityType
		if not constructionSite:isMissingResource( resourceType ) then
			print( "Game: construction site is not missing resource of type", resourceType, "anymore!" )
			print( "construction site id:", constructionSite.id )
			return
		end
		constructionSite:addResourceProgress( resourceType )
		
		carriedEntity:setDeposit( constructionSite )
		EntityManager:remove( carriedEntity )
		-- print( "game: destroy resource entity for construction site" )
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- called from construction site when all resources are there
	-- the building will be created with 0% construction progress? (TODO)
	function game:replaceConstructionSite( constructionSite, buildingType )
		
		local x, y = constructionSite:getPosition()
		local xi, yi = constructionSite.xi, constructionSite.yi
		local wasSelected = constructionSite.selected == true
		
		print( "game:replaceConstructionSite at", xi, yi )
		-- print( "...with type", buildingType )
		-- print( "construction site id:", constructionSite.id )
		
		-- important: remove all entites from the blocking area!
		local entities = EntityManager:getEntitiesFilter( function(e)
			-- if e.entityType ~= ENTITIES.rockRaider
			if not e.getPosition
			or not e.setPosition then
				return false
			end
			local x, y = e:getPosition()
			local exi, eyi = getTileIndex( x, y )
			return exi == xi and eyi == yi
		end )
		-- for now simply set them to the bottom corner of the tile, TODO
		local new_x, new_y = x + 50, y + 50
		for i, entity in pairs( entities ) do
			entity:setPosition( new_x, new_y )
		end
		
		EntityManager:remove( constructionSite )
		
		EntityManager:createEntity(
			buildingType,
			{ x = x, y = y, xi = xi, yi = yi }
		)
		
		if wasSelected then
			self:selectTile( xi, yi )
		end
		
	end
			
	----------------------------------------------------------------------------------------------------
	
	-- resources are globally accessible...
	-- when a worker deposits e.g. an ore entity at a tool store
	function game:addToPlayersGlobalResource( resourceType, amount )
		local amount = amount or 1
		self.playerGlobalResources[resourceType] = self.playerGlobalResources[resourceType] + amount
		-- print( "game add player resource;", resourceType, self.playerGlobalResources[resourceType] )
	end
	
	
	function game:getPlayerGlobalResource( resourceType )
		return self.playerGlobalResources[resourceType] or 0
	end
	
	
	-- check if the player has enough resources
	-- cost is a table with the resource type as indices
	-- e.g. hasPlayerEnoughResources{ [ENTITIES.ore] = 8 }
	function game:hasPlayerEnoughResources( cost )
		for k, v in pairs( cost ) do
			if self:getPlayerGlobalResource( k ) < v then
				-- print( "game:hasPlayerEnoughResources - not enough of", k )
				return false
			end
		end
		return true
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- can be called from the mission script
	function game:victory()
		
		if self.gamestate ~= GAMESTATE_PLAYING then
			return fals
		end
		
		self.gamestate = GAMESTATE_WON
		
		Logger:log( "Set game state to won." )
		
		-- TODO: suspend and teleport all units...
		-- ... and wait for all units to be safely teleported
		
		-- go to post game screen
		self:close()
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- iterate over all of the walls to update the code and their image
	function game:updateWallImageCodes( walls )
		for i, wall in pairs( walls ) do
			local xi, yi = wall.xi, wall.yi	-- get the tile index of the wall
			local code = {0,0,0,0,0,0,0,0}	-- representing the 8 neighbours
			local index = 1
			for dy = -1,1 do
				for dx = -1,1 do
					if dx ~= 0 or dy ~= 0 then
						local x = xi + dx
						local y = yi + dy
						if self:isValidTileIndex( x, y ) then
							-- normal variant:
							code[index] = self.wallData[y][x] > 0 and 1 or 0
							
							-- "fake walls" variant:
							-- if self.wallData[y][x] > 0
							-- or not self.fow:isTileDiscovered( x, y ) then
								-- code[index] = 1
							-- else
								-- code[index] = 0
							-- end
						else
							code[index] = 1	-- assume all borders to be walls
						end
						
						index = index + 1
					end
				end
			end
			
			-- print( xi, yi, table.concat(code) )
			wall:updateImageCode( code )
		end
	end
	
	
	function game:isValidTileIndex( xi, yi )
		if xi < 1 or xi > self.mapSizeX
		or yi < 1 or yi > self.mapSizeY then
			return false
		end
		return true
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:updateFogOfWar()
		self.fow:prepareUpdateTick()
		
		-- fetch and apply exploration for all player entities
		local entities = EntityManager:getEntitiesFilter( function(e)
			return e.getPosition and e.getExplorationRange
		end )
		
		local exploreTiles = {}	-- tile indices to be explored eventually
		local walls = self.wallData
		local function explore( xi, yi, depth )
			if not self:isValidTileIndex( xi, yi ) then
				return false
			end
			table.insert( exploreTiles, xi )
			table.insert( exploreTiles, yi )
			-- stop when hitting a wall
			if walls[yi][xi] > 0 then
				return false
			end
			-- can we look further?
			depth = depth - 1
			if depth > 0 then
				for dy = -1, 1 do
					for dx = -1, 1 do
						if dx ~= 0 or dy ~= 0 then
							explore( xi+dx, yi+dy, depth )
						end
					end
				end
			end
		end
		
		for _, entity in pairs( entities ) do
			local ex, ey = entity:getPosition()
			local exi, eyi = getTileIndex( ex, ey )
			local range = entity:getExplorationRange()
			
			if entity.explorationIgnoresWalls then
				
				-- discover everything in exploration range around the unit, view not blocked by walls
				-- maybe usful later? available via upgrades? for special units or buildings?
				for dy = -range, range do
					for dx = -range, range do
						local xi = exi + dx
						local yi = eyi + dy
						if Tools.getManhattanDistance2( exi, eyi, xi, yi ) <= range then
							local isNewDiscovery = self.fow:setTileExplored( xi, yi )
							if isNewDiscovery then
								self:onTileDiscovered( xi, yi )
							end
						end
					end
				end
				
			else
			
				exploreTiles = {}
				
				-- explore tile directly
				explore( exi,eyi, range )
				
				-- tiles to be explored are stored in a list: exploreTiles with xi and yi entries
				for i = 1, #exploreTiles, 2 do
					local xi = exploreTiles[i]
					local yi = exploreTiles[i+1]
					if Tools.getManhattanDistance2( exi, eyi, xi, yi ) <= range then
						local isNewDiscovery = self.fow:setTileExplored( xi, yi )
						if isNewDiscovery then
							self:onTileDiscovered( xi, yi )
						end
					end
				end
				
			end
			
		end
		
		-- if we include the "fake walls" to work with the FOW ... update the wall image codes
		-- disabled again because it can result in invalid arrangements (no valid image exists...)
		-- local walls = EntityManager:getEntitiesInCategory( ENTITYCATEGORY.wall )
		-- self:updateWallImageCodes( walls )
	end
	
	
	-- a tile is now explored that was not discovered before
	function game:onTileDiscovered( xi, yi )
		
		-- print( "Game:onTileDiscovered", xi, yi )
		
		-- if there are resources on this tile, inform the player
		local resourceEntities = EntityManager:getEntitiesFilter( function(e)
			if e.entityCategory ~= ENTITYCATEGORY.resource or not e:canBePickedUp() then
				return false
			end
			local ex, ey = e:getPosition()
			local exi, eyi = getTileIndex( ex, ey )
			if exi ~= xi or eyi ~= yi then
				return false
			end
			return true
		end )
		for i, resourceEntity in pairs( resourceEntities ) do
			self:onResourceDiscovered( resourceEntity, xi, yi )
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- lava can flood neighbour tiles
	-- lava deals damage to units on top of it (except they are protected...)
	-- dt is a fixed (but real) value here
	function game:updateLava( dt )
		
		local random = math.random
		local lavaDamage = 10
		local terrain = self.terrain
		local fow = self.fow
		
		-- damage units, TODO: filter
		-- entity must have a health component
		-- entities inside vehicles are not damaged
		-- some entities can be immune to damage from lava
		local entities = EntityManager:getEntitiesFilter( function(e)
			if not e.health then return false end
			if e.isDrivingVehicle and e:isDrivingVehicle() then return false end
			if e.immuneToLavaDamage then return false end
			return true
		end )
		for i, entity in pairs( entities ) do
			local ex, ey = entity:getPosition()
			local exi, eyi = getTileIndex( ex, ey )
			if terrain:getTerrainTypeAtTile( exi, eyi ) == 3 then
				-- hurt entity...
				self:onEntityHurtEntity( nil, entity, lavaDamage )
			end
		end
		
		-- spread to neighbour tile every once in a while
		local function lavaSpread( xi, yi )
			
			local neighbours = {}
			-- direct neighbours only
			local neighbourIndexOffsets = { 0,-1, 1,0, 0,1, -1,0 }
			for i = 1, #neighbourIndexOffsets, 2 do
				local x = xi + neighbourIndexOffsets[i]
				local y = yi + neighbourIndexOffsets[i+1]
				local tile = terrain:getTile( x, y )
				-- only spread to ground tiles and not beneath walls
				-- also do not occupy a tile if there is water next to it
				-- TODO: other conditions that prevent lava from spreading there?
				if tile and tile.type == 1
				and self.wallData[y][x] == 0
				and not terrain:hasNeighbourOfType( x, y, 2 ) then
					table.insert( neighbours, tile )
				end
			end
			
			-- per tile chance, TODO: value?
			local chance = 0.05
			for i, tile in pairs( neighbours ) do
				if random() < chance and tile.lavaResistance and tile.lavaResistance > 0 then
					tile.lavaResistance = tile.lavaResistance - 1
					if tile.lavaResistance == 0 then
						terrain:setTerrainTypeAtTile( tile.xi, tile.yi, 3 )
						-- print( "game: set lava at tile", tile.xi, tile.yi )
					end
				end
			end
			
		end
		
		for yi = 1, self.mapSizeY do
			for xi = 1, self.mapSizeX do
				local tile = terrain:getTile( xi, yi )
				if tile and tile.type == 3
				and fow:isTileDiscovered( xi, yi ) then
					lavaSpread( xi, yi )
				end
			end
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- called every few(?) seconds
	-- every wall that is not solid and not reinforced can cause a landslide from time to time
	-- if it has a neighbour tile that is free and has normal ground
	function game:manageLandslides()
		local random = math.random
		local chance = 0.001	-- TODO: value can be adjusted in the mission script
		local walls = EntityManager:getEntitiesFilter( function(e)
			return e.entityCategory == ENTITYCATEGORY.wall
				and e.allowLandslides
		end )
		for i, wall in pairs( walls ) do
			local canSlide, xi, yi = self:checkLandslideConditions( wall )
			if canSlide then
				if random() < chance then
					self:startLandslide( wall, xi, yi )
				end
			end
		end
	end
	
	
	-- for a given wall entity, check if a landslide can occur right now
	-- that is: the last landslide from this wall lies at least X seconds in the past
	-- and there is a valid tile next to it (no wall, normal ground)
	-- return true and the xi and yi tile index of the free neighbour
	function game:checkLandslideConditions( wall )
		
		local random = math.random
		local minTimeBetweenLandslides = 20	-- TODO?
		
		if not wall.getLastLandslideTime or self.gametime - wall:getLastLandslideTime() < minTimeBetweenLandslides then
			return false
		end
		
		if wall.isReinforced then
			return false
		end
		
		-- free and valid neighbour?
		-- x1,y1, x2,y2, ... for all 4 direct neighbours (not diagonal)
		local neighbourIndexOffsets = { 0,-1, 1,0, 0,1, -1,0 }
		local valid_x, valid_y
		for i = 1, #neighbourIndexOffsets, 2 do
			local x = wall.xi + neighbourIndexOffsets[i]
			local y = wall.yi + neighbourIndexOffsets[i+1]
			if self:isValidTileIndex( x, y ) then
				-- no wall there? ok
				if self.wallData[y][x] == 0
				and self.terrain:getTerrainTypeAtTile( x, y ) == 1 then
					-- ensure both possible neighbours are selected equally
					-- this works because there are always at most 2 possible free neighbour tiles
					if not valid_x or random() < 0.5 then
						valid_x, valid_y = x, y
					end
				end
			end
		end
		
		if valid_x then
			return true, valid_x, valid_y
		else
			return false
		end
	end
	
	
	function game:startLandslide( wall, xi, yi )
		
		local random = math.random
		local x, y = (xi-1) * tilesize, (yi-1) * tilesize
		
		-- entity contains some animation or effect + logic to hurt entities on that tile
		local landslide = EntityManager:createEntity(
			ENTITIES.landslide,
			-- add a small offset so it will be in front of any building at this tile
			{ x = x + 1, y = y + 1, xi = xi, yi = yi, }
		)
		
		-- inform the wall entity (set landslide time)
		wall:setLastLandslideTime( self.gametime )
		
		-- player notification if the area is already discovered
		if self.fow:isTileExplored( xi, yi ) then
			self:newNotificationMessage( getString( "game_notification_landslide" ) )
		end
		
		-- create debris next to the wall, only if there is no debris already!
		-- and no building!
		local _debris = EntityManager:getEntitiesFilter( function(e)
			return e.entityType == ENTITIES.debris
				and e.xi == xi and e.yi == yi
		end )
		if #_debris > 0 then
			return
		end
		if self:getBuildingAtTile( xi, yi ) then
			return
		end
		
		local debris = EntityManager:createEntity(
			ENTITIES.debris,
			{ x = x, y = y, xi = xi, yi = yi, createdFromLandslide = true }
		)
		if self.fow:isTileDiscovered( xi, yi ) then
			self:addClearDebrisJob( debris )
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:updateMinimap()
		
		if not GUImgr:isVisible( "gameMinimap" ) then
			return
		end
		
		local terrainData = {}
		for yi = 1, self.mapSizeY do
			terrainData[yi] = {}
			for xi = 1, self.mapSizeX do
				-- if self.fow:isTileExplored( xi, yi ) then
				if self.fow:isTileDiscovered( xi, yi ) then
					terrainData[yi][xi] = self.terrain:getTerrainTypeAtTile( xi, yi )
				end
			end
		end
		
		local walls = EntityManager:getEntitiesFilter( function(e)
			return e.entityCategory == ENTITYCATEGORY.wall
				and self.fow:isTileDiscovered( e.xi, e.yi )
		end )
		
		-- fetch all entities that should be drawn + that are not covered by the FoW
		local entities = EntityManager:getEntitiesFilter( function(e)
			local xi, yi
			if e.entityCategory == ENTITYCATEGORY.building
			or e.entityCategory == ENTITYCATEGORY.monster
			or e.entityCategory == ENTITYCATEGORY.vehicle
			or e.entityCategory == ENTITYCATEGORY.robot
			or e.entityType == ENTITIES.rockRaider
			or e.entityType == ENTITIES.ore
			or e.entityType == ENTITIES.energyCrystal then
				if e.xi then
					xi, yi = e.xi, e.yi
				else
					-- local ex, ey = e:getPosition()
					xi, yi = getTileIndex( e:getPosition() )
				end
				return self.fow:isTileExplored( xi, yi )
			end
		end )
		
		self.minimap:updateMapData( terrainData, walls, entities )
		
	end
	
	
	function game:minimapJumpTo( x, y )
		Camera:setPosition( x, y )
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- called from script
	-- update the slider accordingly
	function game:setGameSpeed( speed )
		self.gamespeed = speed
		-- print( "game:setGameSpeed", speed )
		Logger:log( "(Script) Set game speed to " .. speed )
		GUImgr:getObjectById( "sliGameOptionGamespeed" ):setValue( speed )
	end
	
	
	-- called from the UI
	function game:guiActionSetGamespeed( speed )
		self.gamespeed = speed
		Logger:log( "(GUI) Set game speed to " .. speed )
		-- print( "game:guiActionSetGamespeed", speed )
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:tick( dt )
		
		if GUImgr:getCurrentScreenId() ~= self.id then
			return
		end
		
		GUImgr:getObjectById( "lblGameFps" ):setText( love.timer.getFPS() )
		
		if self.isPaused then return end
		
		if self._debugPaused then return end
		
		-- time that passed in the game world (gamespeed modifier)
		local time = self.gamespeed * dt
		self.gametime = self.gametime + time
		
		-- apply gamespeed modifier here
		Gameloop:update( time )
		
		self:updateCameraPosition( dt )
		
		self:updateResourceOverview()
		
		self:updateMinimap()
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- float value from >0 to 2 or 3
	function game:guiActionSetScrollSpeed( speed )
		self.cameraScrollSpeed = speed
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- camera pan
	function game:updateCameraPosition( dt )
		
		if not WINDOW_HAS_FOCUS then
			return
		end
		
		-- double scroll speed in x direction = good movement when both x AND y movement at the same time
		-- but it feels weird if only x OR y movement occurs
		local cx, cy = 0, 0	-- pan to be applied to the camera
		local mouseX, mouseY = love.mouse.getPosition()
		local width, height = SCREEN.width, SCREEN.height
		
		-- can be moved by mouse...
		-- but only if the mouse does not hover any button(?)
		-- if not GUImgr:isMouseHoveringGuiElement() then
			if mouseX < width * borderSizeFactor then
				-- left
				cx = cx - 2
			elseif mouseX > width * (1 - borderSizeFactor) then
				--right
				cx = cx + 2
			end
			if mouseY < height * borderSizeFactor then
				-- up
				cy = cy - 1
			elseif mouseY > height * (1 - borderSizeFactor) then
				-- down
				cy = cy + 1
			end
		-- end
		
		-- can be moved by keypress...
		if love.keyboard.isDown( Keybinds:getKey("camera_pan_up") ) then
			cy = cy - 1
		elseif love.keyboard.isDown( Keybinds:getKey("camera_pan_down") ) then
			cy = cy + 1
		end
		if love.keyboard.isDown( Keybinds:getKey("camera_pan_left") ) then
			-- cx = cx - 1
			cx = cx - 2
		elseif love.keyboard.isDown( Keybinds:getKey("camera_pan_right") ) then
			-- cx = cx + 1
			cx = cx + 2
		end
		if cx ~= 0 or cy ~= 0 then
			-- make the camera move faster when zoomed in (and slower when zoomed out)
			-- scroll speed can also be adjusted in the pause menu
			local scale = Camera:getScale()
			local cameraSpeed = 128 * self.cameraScrollSpeed / scale
			Camera:move( cx * cameraSpeed * dt, cy * cameraSpeed * dt )
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- an entity is about to hurt another entity
	function game:onEntityHurtEntity( attacker, target, damage )
		
		if not target.health then
			return false
		end
		
		local event = {
			entity1 = attacker,
			entity2 = target,
			damage = damage,
		}
		
		target.health:hurt( damage )
		
		self.scriptCallbacks.onEntityHurtEntity( event )
		return event
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- assign jobs to Rock Raiders (and robots?) and check if they are done (?)
	-- gets called every second
	function game:manageWorkerJobAssignment()
		
		local workerJobs = self.workerJobs
		
		-- check all jobs if they are completed
		for i = #workerJobs, 1, -1 do
			local job = workerJobs[i]
			if self:isJobDone( job ) then
				self:onWorkerJobDone( job )
				-- print( "Game: remove completed job", job.type )
				Logger:log( "Game: remove completed job " .. job.type )
				table.remove( workerJobs, i )
			end
		end
		-- print( "game #workerJobs:", #workerJobs )
		
		
		-- sort all jobs by 1. priority (job type), 2. time and 3. entityId (to make the sorting function stable)
		local sortFunction = function( a, b )
			if self.workerJobPriorities[a.type] < self.workerJobPriorities[b.type] then
				return true
			elseif self.workerJobPriorities[a.type] > self.workerJobPriorities[b.type] then
				return false
			end
			-- same priority... check time; assign the oldest jobs first
			if a.time < b.time then
				return true
			elseif a.time > b.time then
				return false
			end
			-- finally, sort by associated entity ID
			return a.entityId > b.entityId
		end
		table.sort( workerJobs, sortFunction )
		
		
		-- iterate over all jobs (high priority first) and for all workers who could do the job,
		-- pick the one who has the shortest path; then assign the job
		
		-- local freeWorkers = self:getIdleWorkers()
		local getWorkers = function()
			return EntityManager:getEntitiesFilter( function(e)
				return e.isWorker == true and e.isPlayerControlled
					and not e.disableAutomaticJobAssigment
					-- workers who are currently selected can still perform their jobs,
					-- but will not be assigned a new one
					--> removed, this is irritating, especially for vehicles (transport truck)
					-- and not e.selected
			end )
		end
		
		local workers = getWorkers()
		for jobIndex, job in ipairs( workerJobs ) do
			
			if job.type == "collectEnergyCrystal" then
				local resourceEntity = EntityManager:getEntityById( job.entityId )
				local resource_x, resource_y = resourceEntity:getPosition()
				local target_tx, target_ty = getTileIndex( resource_x, resource_y )
				local bestPath
				local bestWorker
				
				-- TODO: for both resource types - check if the worker can deposit the resource somewhere?
				
				for i, worker in pairs( workers ) do
					if worker:checkReadyForJob( job ) then
						local worker_x, worker_y = worker:getPosition()
						local worker_tx, worker_ty = getTileIndex( worker_x, worker_y )
						local path = Pathfinding.getPath( worker_tx, worker_ty, target_tx, target_ty, "ground_coarse" )
						-- path may be false! (no path found)
						if path and (not bestPath or #path < bestPath.length) then
							-- this is the first path OR the second path is better than the first...
							bestPath = {
								path = path,
								length = #path,
								targetTileX = target_tx,
								targetTileY = target_ty,
							}
							bestWorker = worker
						end
					end
				end
				
				if bestWorker and bestPath then
					Logger:log( "Game: assign collect energy crystal job to worker " .. bestWorker.id )
					bestWorker:assignJob( job )
					-- remove the job from the queue
					table.remove( self.workerJobs, jobIndex )
					jobAssignedCondition = true
				end
				
			elseif job.type == "collectOre" then
				local resourceEntity = EntityManager:getEntityById( job.entityId )
				local resource_x, resource_y = resourceEntity:getPosition()
				local target_tx, target_ty = getTileIndex( resource_x, resource_y )
				local bestPath
				local bestWorker
				
				for i, worker in pairs( workers ) do
					if worker:checkReadyForJob( job ) then
						local worker_x, worker_y = worker:getPosition()
						local worker_tx, worker_ty = getTileIndex( worker_x, worker_y )
						local path = Pathfinding.getPath( worker_tx, worker_ty, target_tx, target_ty, "ground_coarse" )
						-- path may be false! (no path found)
						if path and (not bestPath or #path < bestPath.length) then
							-- this is the first path OR the second path is better than the first...
							bestPath = {
								path = path,
								length = #path,
								targetTileX = target_tx,
								targetTileY = target_ty,
							}
							bestWorker = worker
						end
					end
				end
				
				if bestWorker and bestPath then
					Logger:log( "Game: assign collect ore job to worker " .. bestWorker.id )
					bestWorker:assignJob( job )
					-- remove the job from the queue
					table.remove( self.workerJobs, jobIndex )
					jobAssignedCondition = true
				end
				
			elseif job.type == "drillWall" then
				
				-- first: check if there is a path from the worker to the target
				-- do not target the wall tile but one neighbour because the wall itself is obviously blocked
				-- check all 4 direct neighbours (not diagonal)
				-- one or two of them should be free
				local wall = EntityManager:getEntityById( job.entityId )
				local xi, yi = wall.xi, wall.yi
				local neighbourIndexOffsets = { 0,-1, 1,0, 0,1, -1,0 }
				local bestPath
				local bestWorker
				
				-- one or two paths possible
				for i = 1, #neighbourIndexOffsets, 2 do
					local target_tx = xi + neighbourIndexOffsets[i]
					local target_ty = yi + neighbourIndexOffsets[i+1]
					-- empty neighbour tile from where a worker could drill this wall
					if self:isValidTileIndex( target_tx, target_ty ) and self.wallData[target_ty][target_tx] == 0 then
						
						for i, worker in pairs( workers ) do
							if worker:checkReadyForJob( job ) then
								local worker_x, worker_y = worker:getPosition()
								local worker_tx, worker_ty = getTileIndex( worker_x, worker_y )
								
								local path = Pathfinding.getPath( worker_tx, worker_ty, target_tx, target_ty, "ground_coarse" )
								-- path may be false! (no path found)
								if path and (not bestPath or #path < bestPath.length) then
									-- this is the first path OR the second path is better than the first...
									bestPath = {
										path = path,
										length = #path,
										targetTileX = target_tx,
										targetTileY = target_ty,
									}
									bestWorker = worker
								end
							end
						end
						
					end
					
				end
				
				if bestWorker and bestPath then
					Logger:log( "Game: assign drill wall job to worker " .. bestWorker.id )
					-- give the info which tile is the best to the worker
					job.targetTileX = bestPath.targetTileX
					job.targetTileY = bestPath.targetTileY
					bestWorker:assignJob( job )
					-- remove the job from the queue
					table.remove( self.workerJobs, jobIndex )
					jobAssignedCondition = true
				end
				
			elseif job.type == "reinforceWall" then
				
				local wall = EntityManager:getEntityById( job.entityId )
				local xi, yi = wall.xi, wall.yi
				local neighbourIndexOffsets = { 0,-1, 1,0, 0,1, -1,0 }
				local bestPath
				local bestWorker
				
				-- one or two paths possible
				for i = 1, #neighbourIndexOffsets, 2 do
					local target_tx = xi + neighbourIndexOffsets[i]
					local target_ty = yi + neighbourIndexOffsets[i+1]
					-- empty neighbour tile from where a worker could reach this wall
					if self:isValidTileIndex( target_tx, target_ty ) and self.wallData[target_ty][target_tx] == 0 then
						
						for i, worker in pairs( workers ) do
							if worker:checkReadyForJob( job ) then
								local worker_x, worker_y = worker:getPosition()
								local worker_tx, worker_ty = getTileIndex( worker_x, worker_y )
								
								local path = Pathfinding.getPath( worker_tx, worker_ty, target_tx, target_ty, "ground_coarse" )
								-- path may be false! (no path found)
								if path and (not bestPath or #path < bestPath.length) then
									-- this is the first path OR the second path is better than the first...
									bestPath = {
										path = path,
										length = #path,
										targetTileX = target_tx,
										targetTileY = target_ty,
									}
									bestWorker = worker
								end
							end
						end
						
					end
					
				end
				
				if bestWorker and bestPath then
					Logger:log( "Game: assign reinforce wall job to worker " .. bestWorker.id )
					-- give the info which tile is the best to the worker
					job.targetTileX = bestPath.targetTileX
					job.targetTileY = bestPath.targetTileY
					bestWorker:assignJob( job )
					-- remove the job from the queue
					table.remove( self.workerJobs, jobIndex )
					jobAssignedCondition = true
				end
				
			elseif job.type == "clearDebris" then
				
				local debris = EntityManager:getEntityById( job.entityId )
				local target_tx, target_ty = debris.xi, debris.yi
				local bestPath
				local bestWorker
				
				for i, worker in pairs( workers ) do
					if worker:checkReadyForJob( job ) then
						local worker_x, worker_y = worker:getPosition()
						local worker_tx, worker_ty = getTileIndex( worker_x, worker_y )
						local path = Pathfinding.getPath( worker_tx, worker_ty, target_tx, target_ty, "ground_coarse" )
						-- path may be false! (no path found)
						if path and (not bestPath or #path < bestPath.length) then
							-- this is the first path OR the second path is better than the first...
							bestPath = {
								path = path,
								length = #path,
								targetTileX = target_tx,
								targetTileY = target_ty,
							}
							bestWorker = worker
						end
					end
				end
				
				if bestWorker and bestPath then
					Logger:log( "Game: assign clear debris job to worker " .. bestWorker.id )
					bestWorker:assignJob( job )
					-- remove the job from the queue
					table.remove( self.workerJobs, jobIndex )
					jobAssignedCondition = true
				end
				
			elseif job.type == "vehicleGetDriver" then
				
				local vehicle = EntityManager:getEntityById( job.entityId )
				local vehicle_x, vehicle_y = vehicle:getPosition()
				local target_tx, target_ty = getTileIndex( vehicle_x, vehicle_y )
				local bestPath
				local bestWorker
				
				for i, worker in pairs( workers ) do
					if worker:checkReadyForJob( job ) then
						local worker_x, worker_y = worker:getPosition()
						local worker_tx, worker_ty = getTileIndex( worker_x, worker_y )
						local path = Pathfinding.getPath( worker_tx, worker_ty, target_tx, target_ty, "ground_coarse" )
						-- path may be false! (no path found)
						if path and (not bestPath or #path < bestPath.length) then
							bestPath = {
								path = path,
								length = #path,
							}
							bestWorker = worker
						end
					end
				end
				
				if bestWorker and bestPath then
					Logger:log( "Game: assign get driver job to worker " .. bestWorker.id )
					bestWorker:assignJob( job )
					-- remove the job from the queue
					table.remove( self.workerJobs, jobIndex )
					jobAssignedCondition = true
				end
				
			elseif job.type == "placeConstructionFence" then
				
				local fence = EntityManager:getEntityById( job.entityId )
				local vehicle_x, vehicle_y = fence:getPosition()
				local target_tx, target_ty = getTileIndex( vehicle_x, vehicle_y )
				local bestPath
				local bestWorker
				
				for i, worker in pairs( workers ) do
					if worker:checkReadyForJob( job ) then
						local worker_x, worker_y = worker:getPosition()
						local worker_tx, worker_ty = getTileIndex( worker_x, worker_y )
						local path = Pathfinding.getPath( worker_tx, worker_ty, target_tx, target_ty, "ground_coarse" )
						-- path may be false! (no path found)
						if path and (not bestPath or #path < bestPath.length) then
							bestPath = {
								path = path,
								length = #path,
							}
							bestWorker = worker
						end
					end
				end
				
				if bestWorker and bestPath then
					Logger:log( "Game: assign place fence job to worker " .. bestWorker.id )
					bestWorker:assignJob( job )
					-- remove the job from the queue
					table.remove( self.workerJobs, jobIndex )
					jobAssignedCondition = true
				end
				
			elseif job.type == "bringResourceToBuilding" then
				
				local constructionSite = EntityManager:getEntityById( job.entityId )
				local constructionSite_x, constructionSite_y = constructionSite:getPosition()
				local target_tx, target_ty = getTileIndex( constructionSite_x, constructionSite_y )
				local bestPath
				local bestWorker
				
				for i, worker in pairs( workers ) do
					if worker:checkReadyForJob( job ) then
						local worker_x, worker_y = worker:getPosition()
						local worker_tx, worker_ty = getTileIndex( worker_x, worker_y )
						local path = Pathfinding.getPath( worker_tx, worker_ty, target_tx, target_ty, "ground_coarse" )
						-- path may be false! (no path found)
						if path and (not bestPath or #path < bestPath.length) then
							bestPath = {
								path = path,
								length = #path,
							}
							bestWorker = worker
						end
					end
				end
				
				if bestWorker and bestPath then
					Logger:log( "Game: assign bringResourceToBuilding job to worker " .. bestWorker.id )
					bestWorker:assignJob( job )
					-- remove the job from the queue
					table.remove( self.workerJobs, jobIndex )
					jobAssignedCondition = true
				end
				
			end-- job type
			
			-- scan for workers again (one might have a new job already)
			workers = getWorkers()
			
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- cannot directly search a path to a wall tile (because it is blocked)
	-- instead, check all four direct neighbours
	-- return the shortest path or false
	function game:getShortestPathToWall( wall_xi, wall_yi, start_xi, start_yi )
		
		local neighbourIndexOffsets = { 0,-1, 1,0, 0,1, -1,0 }
		local bestPath
		
		for i = 1, #neighbourIndexOffsets, 2 do
			local target_tx = wall_xi + neighbourIndexOffsets[i]
			local target_ty = wall_yi + neighbourIndexOffsets[i+1]
			-- empty neighbour tile from where a worker could drill or reinforce this wall
			if self:isValidTileIndex( target_tx, target_ty ) and self.wallData[target_ty][target_tx] == 0 then
				local path = Pathfinding.getPath( start_xi, start_yi, target_tx, target_ty, "ground_coarse" )
				if path and (not bestPath or #path < bestPath.length) then
					bestPath = {
						path = path,
						length = #path,
						-- targetTileX = target_tx,
						-- targetTileY = target_ty,
					}
					bestWorker = worker
				end
			end
		end
		
		if bestPath then
			return bestPath.path
		else
			return false
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:getIdleWorkers()
		-- check Rock Raiders and vehicles (with drivers)
		local workers = EntityManager:getEntitiesFilter( function(e)
			-- if not e.isWorker or not e.checkReadyForJob then
				-- return false
			-- end
			-- return e:checkReadyForJob()
			return e.task == "idle"
		end )
		return workers
	end
	
	-- deprecated
	function game:tryToAssignJobToWorker( worker )
		
		
		-- print( "no job to assign" )
		return false
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:guiActionRockRaiderToggleAutomaticJobAssignment()
		local entity = self:getSelectedEntity()
		if not entity
		or entity.entityType ~= ENTITIES.rockRaider then
			return false
		end
		
		entity.disableAutomaticJobAssigment = not entity.disableAutomaticJobAssigment
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- called after job is done
	function game:onWorkerJobDone( workerJob )
		local type = workerJob.type
		-- nothing so far
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- e.g. a RR had a job but was given a different player command (e.g. walk somewhere)
	function game:entityCancelJob( worker, workerJob )
		-- job is done anyway? fine
		if self:isJobDone( workerJob ) then
			return
		end
		
		-- job still needs to be done ... put it back in the right queue
		-- also skip the queuedJobs check because this job has already been queued, was not removed
		-- because it was not done yet, thus it must be queued again
		local type = workerJob.type
		if type == "clearDebris" then
			local debris = EntityManager:getEntityById( workerJob.entityId )
			self:addClearDebrisJob( debris, true )
		elseif type == "drillWall" then
			local wall = EntityManager:getEntityById( workerJob.entityId )
			self:addDrillWallJob( wall, true )
		elseif type == "reinforceWall" then
			local wall = EntityManager:getEntityById( workerJob.entityId )
			self:addReinforceWallJob( wall, true )
		elseif type == "collectEnergyCrystal" then
			local energyCrystal = EntityManager:getEntityById( workerJob.entityId )
			self:addCollectEnergyCrystalJob( energyCrystal, true )
		elseif type == "collectOre" then
			local ore = EntityManager:getEntityById( workerJob.entityId )
			self:addCollectOreJob( ore, true )
		elseif type == "placeConstructionFence" then
			local fence = EntityManager:getEntityById( workerJob.entityId )
			self:addPlaceConstructionFenceJob( fence )
		end
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:isJobDone( workerJob )
		
		if not workerJob then
			return true
		end
		
		if workerJob.type == "drillWall" then
			
			return not EntityManager:isValidEntityId( workerJob.entityId )
			
		elseif workerJob.type == "reinforceWall" then
			
			local wall = EntityManager:getEntityById( workerJob.entityId )
			return not wall or wall.isReinforced == true
			
		elseif workerJob.type == "clearDebris" then
			
			return not EntityManager:isValidEntityId( workerJob.entityId )
			
		elseif workerJob.type == "collectEnergyCrystal" or workerJob.type == "collectOre" then
			
			local resourceEntity = EntityManager:getEntityById( workerJob.entityId )
			-- resource entity does not exist anymore (because used for building?)
			if not resourceEntity then
				return true
			end
			-- resource entity is now at a deposit
			-- or carried by some entity?
			return resourceEntity.depositEntity ~= nil-- or resourceEntity.carrierEntity ~= nil
			
		elseif workerJob.type == "placeConstructionFence" then
			
			local fence = EntityManager:getEntityById( workerJob.entityId )
			if not fence then
				return true
			end
			-- the fence has been placed at the construction site
			return fence.isPlacedAtConstructionSite == true
			
		elseif workerJob.type == "bringResourceToBuilding" then
			
			local constructionSite = EntityManager:getEntityById( workerJob.entityId )
			if not constructionSite then
				return true
			end
			-- job is done when the construction site is gone
			return false
			
		elseif workerJob.type == "vehicleGetDriver" then
			
			local vehicle = EntityManager:getEntityById( workerJob.entityId )
			if not vehicle then
				return true
			end
			return vehicle:hasDriver()
			
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- make sure to not add a new job for the same thing multiple times!
	-- skipQueuedCheck allows to force queue a job again, see entityCancelJob -> not needed anymore?
	
	function game:addDrillWallJob( wall, skipQueuedCheck )
		-- check if the job is already queued
		for i = 1, #self.workerJobs do
			local job = self.workerJobs[i]
			if job.type == "drillWall"
			and job.entityId == wall.id then
				print( "Game: wall is already queued to be drilled", wall.id )
				return false
			end
		end
		
		Logger:log( "Game: new drill wall job " .. wall.id )
		local job = {
			type = "drillWall",
			entityId = wall.id,
			time = self.gametime,
		}
		table.insert( self.workerJobs, job )
		return true
	end
	
	
	function game:addReinforceWallJob( wall )
		for i = 1, #self.workerJobs do
			local job = self.workerJobs[i]
			if job.type == "reinforceWall"
			and job.entityId == wall.id then
				print( "Game: wall is already queued to be reinforced", wall.id )
				return false
			end
		end
		
		Logger:log( "Game: new reinforce wall job " .. wall.id )
		local job = {
			type = "reinforceWall",
			entityId = wall.id,
			time = self.gametime,
		}
		table.insert( self.workerJobs, job )
		return true
	end
	
	
	function game:addClearDebrisJob( debris, skipQueuedCheck )
		for i = 1, #self.workerJobs do
			local job = self.workerJobs[i]
			if job.type == "clearDebris"
			and job.entityId == debris.id then
				print( "Game: debris is already queued to be cleared", debris.id )
				return false
			end
		end
		
		Logger:log( "Game: new clear debris job " .. debris.id )
		local job = {
			type = "clearDebris",
			entityId = debris.id,
			time = self.gametime,
		}
		table.insert( self.workerJobs, job )
		return true
	end
	
	
	-- separate for energy crystal and ore because the player can change the priority individually
	function game:addCollectEnergyCrystalJob( resource )
		for i = 1, #self.workerJobs do
			local job = self.workerJobs[i]
			if job.type == "collectEnergyCrystal"
			and job.entityId == resource.id then
				print( "Game: energy crystal is already queued to be collected", resource.id )
				return false
			end
		end
		
		Logger:log( "Game: new collect energy crystal job " .. resource.id )
		local job = {
			type = "collectEnergyCrystal",
			entityId = resource.id,
			time = self.gametime,
		}
		table.insert( self.workerJobs, job )
		return true
	end
	
	
	function game:addCollectOreJob( resource )
		for i = 1, #self.workerJobs do
			local job = self.workerJobs[i]
			if job.type == "collectOre"
			and job.entityId == resource.id then
				print( "Game: ore is already queued to be collected", resource.id )
				return false
			end
		end
		
		Logger:log( "Game: new collect ore job " .. resource.id )
		local job = {
			type = "collectOre",
			entityId = resource.id,
			time = self.gametime,
		}
		table.insert( self.workerJobs, job )
		return true
	end
	
	-- wrapper to collect ore or energy crystals
	function game:addCollectResourceJob( resource )
		if resource.entityType == ENTITIES.ore then
			return self:addCollectOreJob( resource )
		elseif resource.entityType == ENTITIES.energyCrystal then
			return self:addCollectEnergyCrystalJob( resource )
		end
	end
	
	
	function game:addVehicleGetDriverJob( vehicle )
		for i = 1, #self.workerJobs do
			local job = self.workerJobs[i]
			if job.type == "vehicleGetDriver"
			and job.entityId == vehicle.id then
				print( "Game: vehicle is already queued to be manned", vehicle.id )
				return false
			end
		end
		
		Logger:log( "Game: new get driver job " .. vehicle.id )
		local job = {
			type = "vehicleGetDriver",
			entityId = vehicle.id,
			manuallyAssigned = true,	-- for now this is always true for vehicles
			time = self.gametime,
		}
		table.insert( self.workerJobs, job )
		return true
	end
	
	
	-- for a new building (construction site actually)
	function game:addPlaceConstructionFenceJob( fence )
		Logger:log( "Game: new place construction fence job " .. fence.id )
		local job = {
			type = "placeConstructionFence",
			entityId = fence.id,
			time = self.gametime,
		}
		table.insert( self.workerJobs, job )
		return true
	end
	
	
	-- for each required resource per construction site
	-- no concrete resource entity specified because maybe not exising yet!
	function game:addBringResourceToBuildingJob( constructionSite, resourceType )
		Logger:log( "Game: new bring resource to building job " .. constructionSite.id )
		local job = {
			type = "bringResourceToBuilding",
			entityId = constructionSite.id,
			resourceType = resourceType,
			time = self.gametime,
		}
		table.insert( self.workerJobs, job )
		return true
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- look for a mini tile next to a tile (with a wall)
	-- return the mini tile indices which are both inside the neighbour tile
	-- parameters: full tile index of the primary tile and full tile index of the neighbour (TODO: optional?)
	function game:getFreeMiniTileNextToFullTile( xi, yi, nxi, nyi )
		if not self:isValidTileIndex( xi, yi ) then
			print( "Game:getFreeMiniTileNextToFullTile error - invalid tile index!", xi, yi )
			return false
		end
		if not self:isValidTileIndex( nxi, nyi ) then
			print( "Game:getFreeMiniTileNextToFullTile error - invalid neighbour tile index!", nxi, nyi )
			return false
		end
		
		-- if the neighbour tile is to the left of the primary tile (nxi < xi)
		--> check the mini tiles on the right border of the neighbour tile
		if nxi < xi then
			-- tile index * 5 = mini tile index of the last mini tile (bottom right corner)
			--> continue with -2 to get the central mini tile OR -4 to get the first (top left) mini tile
			-- check the central mini tile first, then move towards the corners if needed
			local miniXi = nxi * 5
			-- for miniYi = nyi * 5 - 4, nyi * 5 do
			local y = nyi * 5 - 2	-- central mini tile index
			for _, miniYi in ipairs{ y, y-1, y+1, y-2, y+2 } do
				if not self.blocking:isMiniTileBlocked( miniXi, miniYi ) then
					return miniXi, miniYi
				end
			end
		
		-- neighbour tile is to the right of the primary tile (nxi > xi)
		elseif nxi > xi then
			local miniXi = nxi * 5 - 4
			-- for miniYi = nyi * 5 - 4, nyi * 5 do
			local y = nyi * 5 - 2	-- central mini tile index
			for _, miniYi in ipairs{ y, y-1, y+1, y-2, y+2 } do
				if not self.blocking:isMiniTileBlocked( miniXi, miniYi ) then
					return miniXi, miniYi
				end
			end
		
		-- neighbour tile is above the primary tile (nyi < yi)
		elseif nyi < yi then
			local x = nxi * 5 - 2	-- central mini tile index
			local miniYi = nyi * 5
			-- for miniXi = nxi * 5 - 4, nxi * 5 do
			for _, miniXi in ipairs{ x, x-1, x+1, x-2, x+2 } do
				if not self.blocking:isMiniTileBlocked( miniXi, miniYi ) then
					return miniXi, miniYi
				end
			end
		
		-- neighbour tile is below the primary tile (nyi > yi)
		elseif nyi > yi then
			local x = nxi * 5 - 2	-- central mini tile index
			-- for miniXi = nxi * 5 - 4, nxi * 5 do
			local miniYi = nyi * 5 - 4
			for _, miniXi in ipairs{ x, x-1, x+1, x-2, x+2 } do
				if not self.blocking:isMiniTileBlocked( miniXi, miniYi ) then
					return miniXi, miniYi
				end
			end
		
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:onKeyPressed( key, scancode )
		
		local action = Keybinds:getAction( key )
		if not action then return end
		
		if action == "toggle_fps_count" then
			GUImgr:toggleVisibility( "lblGameFps" )
			
		elseif action == "toggle_entity_overhead" then
			self.drawEntityStatusOverhead = not self.drawEntityStatusOverhead
			
		elseif action == "debug_pause" then
			-- DEBUG ONLY
			self._debugPaused = not self._debugPaused
			
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:onMousePressed( _x, _y, button )
		
		if GUImgr:getCurrentScreenId() ~= self.id then
			return
		end
		
		if self.isPaused then return end
		
		-- ignore click if there is any UI element in front...
		if GUImgr:isMouseHoveringGuiElement() then
			return
		end
		
		-- print( "game: onMousePressed", _x, _y )
		local wx, wy = Camera:getTransform():inverseTransformPoint( _x, _y )
		-- print( "world:", wx, wy )
		local x, y = convertTo2D( wx, wy )	-- naming for conversion seems confusing...
		-- print( "world:", x, y )
		local xi, yi = getTileIndex( x, y )
		local mini_xi, mini_yi = getMiniTileIndex( x, y )
		-- print( xi, yi )
		
		-- if selection state...
		-- select the tile or the building / wall that is placed there
		if button == 1 then	-- left click
			self:clearEntitySelection()
			
			if not self.fow:isTileDiscovered( xi, yi ) then
				return
			end
			
			-- selecting an entity such as Rock Raiders requires to check for 2d coordinates, not world coordinates!
			local entityAtMousePosition = self:getEntityAtScreenPosition( _x, _y )
			if entityAtMousePosition then
				self:selectEntity( entityAtMousePosition )
				-- local ex, ey = entityAtMousePosition:getPosition()
				-- local start_xi, start_yi = getMiniTileIndex( ex, ey )
				-- print( "start:", start_xi, start_yi )
				return
			end
			
			if self:isValidTileIndex( xi, yi ) then
				self:selectTile( xi, yi )
			-- else
				-- self:selectTile()
			end
			
		elseif button == 2 then	-- right click
			
			-- do nothing when clicking inside an undiscovered area
			if not self.fow:isTileDiscovered( xi, yi ) then
				return
			end
			
			-- if RR selected -> perform action (walk, drill wall, ...)
			local selectedEntities = self:getSelectedEntities()
			local entityAtMousePosition = self:getEntityAtScreenPosition( _x, _y )
			
			-- click on ground = move there, wall = add drill job, ...
			for i, entity in pairs( selectedEntities ) do
				
				-- interaction with selected entity? (from 2d hitbox)
				-- TODO ...
				
				if entity.isWorker and entity:canIssueMoveCommand() then
					local ex, ey = entity:getPosition()
					local worker_tx, worker_ty = getTileIndex( ex, ey )
					
					-- drill wall
					local walls = EntityManager:getEntitiesFilter( function(e)
						return e.entityCategory == ENTITYCATEGORY.wall
							and e.canBeDrilled
							and e.xi == xi and e.yi == yi
					end )
					if walls[1] and entity:canPerformJobType( "drillWall" ) then
						local wall = walls[1]
						-- has a path?
						local path = self:getShortestPathToWall( wall.xi, wall.yi, worker_tx, worker_ty )
						if not path then
							print( "Game info - no path to wall" )
							return
						end
						-- let the worker drill the wall
						local job = {
							type = "drillWall",
							entityId = wall.id,
							time = self.gametime,
							targetTileX = path[1].x,	-- required for the worker... because there are multiple potential neighbour tiles to go
							targetTileY = path[1].y,
							-- important: mark this job as created from player command
							--> it does not go into the global queue, it will not be listed again if the job is canceled
							manuallyAssigned = true,
						}
						entity:assignJob( job )
						return
					end
					
					-- collect resource entity
					if entityAtMousePosition
					and entity:canPerformJobType( "collectEnergyCrystal" )
					and entityAtMousePosition.entityType == ENTITIES.energyCrystal then
						if Pathfinding.getPath( xi, yi, worker_tx, worker_ty, "ground_coarse" ) then
							local job = {
								type = "collectEnergyCrystal",
								entityId = entityAtMousePosition.id,
								time = self.gametime,
								manuallyAssigned = true,
							}
							entity:assignJob( job )
							return
						end
					end
					
					if entityAtMousePosition
					and entity:canPerformJobType( "collectOre" )
					and entityAtMousePosition.entityType == ENTITIES.ore then
						if Pathfinding.getPath( xi, yi, worker_tx, worker_ty, "ground_coarse" ) then
							local job = {
								type = "collectOre",
								entityId = entityAtMousePosition.id,
								time = self.gametime,
								manuallyAssigned = true,
							}
							entity:assignJob( job )
							return
						end
					end
					
					if entityAtMousePosition
					and entity:canPerformJobType( "collectRockSample" )
					and entityAtMousePosition.entityType == ENTITIES.rockSample then
						if Pathfinding.getPath( xi, yi, worker_tx, worker_ty, "ground_coarse" ) then
							local job = {
								type = "collectRockSample",
								entityId = entityAtMousePosition.id,
								time = self.gametime,
								manuallyAssigned = true,
							}
							entity:assignJob( job )
							return
						end
					end
					
					-- drive vehicle
					if entityAtMousePosition
					and entity.entityType == ENTITIES.rockRaider
					and entityAtMousePosition.canBeManned and entityAtMousePosition:canBeManned() then
					-- and entityAtMousePosition.entityCategory == ENTITYCATEGORY.vehicle then
						if Pathfinding.getPath( xi, yi, worker_tx, worker_ty, "ground_coarse" ) then
							local job = {
								type = "vehicleGetDriver",
								entityId = entityAtMousePosition.id,
								time = self.gametime,
								-- important: mark this job as created from player command
								--> it does not go into the global queue, it will not be listed again if the job is canceled
								manuallyAssigned = true,
							}
							entity:assignJob( job )
							return
						end
					end
					
					-- clear debris
					local _debris = EntityManager:getEntitiesFilter( function(e)
						return e.entityType == ENTITIES.debris
							and e.xi == xi and e.yi == yi
					end )
					if _debris[1] and entity:canPerformJobType( "clearDebris" ) then
						local debris = _debris[1]
						local path = Pathfinding.getPath( xi, yi, worker_tx, worker_ty, "ground_coarse" )
						if not path then
							print( "Game info - no path to debris" )
							return
						end
						local job = {
							type = "clearDebris",
							entityId = debris.id,
							time = self.gametime,
							-- important: mark this job as created from player command
							--> it does not go into the global queue, it will not be listed again if the job is canceled
							manuallyAssigned = true,
						}
						entity:assignJob( job )
						return
					end
					
				end
				
				-- move
				if entity.moveToDestination
				and entity:canIssueMoveCommand() then
					entity:moveToDestination( mini_xi, mini_yi, x, y )
					return
				end
				
			end
			
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:onWheelMoved( dx, dy )
		
		if GUImgr:isMouseHoveringGuiElement() or self.isPaused then
			return
		end
		
		-- camera zoom
		if dy < 0 then
			-- moved down
			Camera:zoomOut()
		elseif dy > 0 then
			-- moved up
			Camera:zoomIn()
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- note: while this works fine for selecting units, it has one drawback:
	-- when right clicking on a vehicle while a RR is selected and standing directly next to the vehicle,
	-- the RR will be returned again by this function (so he will walk instead of getting into the vehicle)
	
	-- there might be multiple entities at this position...
	function game:getEntityAtScreenPosition( _x, _y )
		local point = { x = _x, y = _y }
		local filter = function(e)
			if not e.get2dHitbox then return false end
			local hitbox = e:get2dHitbox()
			return Tools.isPointInsideShape( point, hitbox )
		end
		local entities = EntityManager:getEntitiesFilter( filter )
		-- print( "Game:getEntityAtScreenPosition - #entities", #entities )
		if #entities == 0 then
			return false
		end
		-- always try to pick Rock Raiders first
		for i, entity in pairs( entities ) do
			if entity.entityType == ENTITIES.rockRaider then
				return entity
			end
		end
		-- otherwise, return the first hit
		return entities[1]
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:getSelectedEntities()
		return self.selectedEntities
	end
	
	-- return the first of all selected entities
	function game:getSelectedEntity()
		return self.selectedEntities[1]
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:selectTile( xi, yi )
		
		if self.terrain then
			self.terrain:selectTile( xi, yi )
		end
		
		if not xi then
			self.selectedTile = nil
			self:switchToSelectionState( "global" )
			return false
		end
		
		self.selectedTile = { xi = xi, yi = yi }
		
		self:switchToSelectionState( "tile" )
		
		-- wall
		local wallSelected = false
		local walls = EntityManager:getEntitiesInCategory( ENTITYCATEGORY.wall )
		for i, wall in pairs( walls ) do
			if wall.xi == xi and wall.yi == yi then
				if wall.entityType == ENTITIES.solidRock then	-- do not select solid rock...
					self:selectTile()
					return false
				end
				-- self:switchToSelectionState( "wall" )
				self:selectEntity( wall )
				-- print( "Game: select wall at", xi, yi )
				wallSelected = true
				break
			end
		end
		
		-- building
		if not wallSelected then
			local buildingSelected = false
			local building = self:getBuildingAtTile( xi, yi )
			if building then
				-- self:switchToSelectionState( "building" )
				self:selectEntity( building )
				buildingSelected = true
			end
			
			-- building area?
			if not buildingSelected then
				local buildingAreaSelected = false
				local buildings = EntityManager:getEntitiesInCategory( ENTITYCATEGORY.buildingArea )
				for i, building in pairs( buildings ) do
					if building.xi == xi and building.yi == yi then
					-- if building.xi == xi and building.yi == yi
					-- and building.entityType ~= ENTITIES.buildingArea then
						-- self:switchToSelectionState( "tile" )
						self:selectEntity( building )
						buildingAreaSelected = true
						break
					end
				end
			
				-- debris?
				if not buildingAreaSelected then
					local debris = EntityManager:getEntitiesInCategory( ENTITYCATEGORY.debris )
					for i, d in pairs( debris ) do
						if d.xi == xi and d.yi == yi then
							-- self:switchToSelectionState( "tile" )
							self:selectEntity( d )
							break
						end
					end
				end
			end
		end
		
	end
	
	
	function game:reselectCurrentTile()
		if not self.selectedTile then
			return
		end
		self:selectTile( self.selectedTile.xi, self.selectedTile.yi )
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:clearEntitySelection()
		if self.selectedEntity then
			if self.selectedEntity.setSelected then
				self.selectedEntity:setSelected( false )
			end
		end
		self.selectedEntity = nil
		self.selectedEntities = {}
		self:selectTile()
		self:onEntitySelectionChanged()
	end
	
	
	-- player selected a wall, building or mobile entity
	function game:selectEntity( entity )
		-- print( "game:selectEntity of type", entity.entityType )
		
		-- if the entity is currently driving a vehicle, select the vehicle instead
		if entity.getVehicle then
			local vehicle = entity:getVehicle()
			if vehicle and vehicle.setSelected then
				return self:selectEntity( vehicle )
			end
		end
		
		self:clearEntitySelection()
		
		if entity.setSelected then
			entity:setSelected( true )
		end
		self.selectedEntity = entity
		self.selectedEntities = { entity }
		-- print( "Game: selectEntity", entity.id, entity.entityType )
		-- self:switchToSelectionState( "entity" )
		
		-- switch selection state
		if entity.entityCategory == ENTITYCATEGORY.wall then
			self:switchToSelectionState( "wall" )
		elseif entity.entityCategory == ENTITYCATEGORY.building then
			self:switchToSelectionState( "building" )
		elseif entity.entityCategory == ENTITYCATEGORY.buildingArea
		or entity.entityCategory == ENTITYCATEGORY.debris then
			self:switchToSelectionState( "tile" )
		else
			self:switchToSelectionState( "entity" )
		end
		
		self:onEntitySelectionChanged()
	end
	
	
	-- called whenever a new entity was added to the selection or removed from it
	function game:onEntitySelectionChanged()
		
		-- hide all first?
		GUImgr:setVisibility( "widGameBuildingMenu", false )
		GUImgr:setVisibility( "widGameVehicleFactoryNewVehicleMenu", false )
		GUImgr:setVisibility( "widGameRobotBayNewRobotMenu", false )
		GUImgr:setVisibility( "btnGameActionTeleportRockRaiderToBuilding", false )
		GUImgr:setVisibility( "btnGameActionExpellUnit", false )
		GUImgr:setVisibility( "btnGameActionRockRaiderToggleAutomaticJobAssignment", false )
		GUImgr:setVisibility( "btnGameActionVehicleGetDriver", false )
		GUImgr:setVisibility( "btnGameActionVehicleDriverLeave", false )
		GUImgr:setVisibility( "btnGameVehicleFactoryToggleNewVehicleMenu", false )
		GUImgr:setVisibility( "btnGameRobotBayToggleNewRobotMenu", false )
		GUImgr:setVisibility( "widGameConfirmDestroyBuilding", false )
		GUImgr:setVisibility( "widGameWorkerJobPriorities", false )
		
		
		-- TODO: show widget...
		local selectedEntities = self:getSelectedEntities()
		if #selectedEntities == 0 then
			
			-- nothing selected, reset?
			
			return
		end
		
		local entity = selectedEntities[1]
		
		if entity.entityCategory == ENTITYCATEGORY.vehicle then
			GUImgr:setVisibility( "btnGameActionVehicleGetDriver", true )
			GUImgr:setVisibility( "btnGameActionVehicleDriverLeave", true )
			GUImgr:setVisibility( "btnGameActionExpellUnit", true )
		end
		
		if entity.entityType == ENTITIES.rockRaider then
			GUImgr:setVisibility( "btnGameActionRockRaiderToggleAutomaticJobAssignment", true )
		end
		
		--[[
		if entity.entityCategory == ENTITYCATEGORY.vehicle then
			
			local buttonList = {
				"btnGameActionVehicleGetDriver",
				"btnGameActionVehicleDriverLeave",
				"btnGameActionExpellUnit",
			}
			
			local x, y = Camera:getTransform():transformPoint( convertToDimetric( entity:getPosition() ) )
			local radius = 120
			local angle = 2*math.pi / #buttonList
			for i = 1, #buttonList do
				local button = GUImgr:getObjectById( buttonList[i] )
				local a = (i-1) * angle + 0.5*math.pi
				local x = x + math.cos( a ) * radius
				local y = y + math.sin( a ) * radius
				button:setPosition( x, y )
				GUImgr:setVisibility( buttonList[i], true )
			end
			
		end
		--]]
		
		if entity.entityType == ENTITIES.teleportPad then
			GUImgr:setVisibility( "btnGameActionTeleportRockRaiderToBuilding", true )
		elseif entity.entityType == ENTITIES.vehicleFactory then
			GUImgr:setVisibility( "btnGameVehicleFactoryToggleNewVehicleMenu", true )
		elseif entity.entityType == ENTITIES.robotBay then
			GUImgr:setVisibility( "btnGameRobotBayToggleNewRobotMenu", true )
		end
		
		if entity.isWorker then
			GUImgr:setVisibility( "btnGameActionExpellUnit", true )
		end
		
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- adding a building = add its blocking data
	-- removing a building = simply use blocking:singleTileUpdate
	function game:onBuildingCreated( entity )
		if not entity.getBlocking
		or not entity.xi or not entity.yi then
			print( "Game:onBuildingCreated error - invalid tile index or no blocking", entity.xi, entity.yi )
			return false
		end
		
		local blocking = entity:getBlocking()
		self.blocking:addEntityBlockingAtTile( entity.xi, entity.yi, blocking )
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- entity (actor) may be a Rock Raider
	function game:entityRemovesMassFromDebris( entity, debris )
		if debris.removeMass then
			debris:removeMass()
		else
			print( "Game:entityRemovesMassFromDebris warning - invalid debris entity" )
		end
	end
	
	-- last mass removed from debris
	--> right before the debris is gone
	--> reveal (create) ore if the debris did not originate from a landslide
	function game:onBeforeDebrisRemoved( debris )
		if debris.createdFromLandslide then
			return
		end
		
		local x, y = debris:getPosition()
		local xi, yi = getTileIndex( x, y )
		-->use a sequence instead to make it look random but have a "good" distribution?
		local random = math.random
		local amount = random( 1, 4 )	-- TODO
		local spread = 50
		for i = 1, amount do
			local ore = EntityManager:createEntity(
				ENTITIES.ore,
				{ x = x + random(-spread,spread), y = y + random(-spread,spread) }
			)
			self:onResourceDiscovered( ore, xi, yi )
		end
		
		-- script callback
		self.scriptCallbacks.onBeforeDebrisRemoved( debris, xi, yi )
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- entity may be a Rock Raider (or a robot...)
	-- who is currently drilling a wall entity
	-- works over time, so need dt!
	function game:entityDrillWall( entity, wall, dt )
		
		-- the driller has some value that resembles his efficiency
		-- the wall has a value that states how hard it is to drill it
		if not wall.drillEndurance
		or not entity.drillWallEfficiency then
			print( "Game warning: cannot drill wall (missing properties)" )
			return false
		end
		
		local newEndurance = wall.drillEndurance - entity.drillWallEfficiency * dt
		if newEndurance > 0 then
			wall:setRemainingDrillEndurance( newEndurance )
		else
			self:removeWall( wall )
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- a Rock Raider is done reinforcing a wall
	function game:entityReinforceWall( self, wall )
		
		if not wall.entityCategory == ENTITYCATEGORY.wall then
			print( "game error: try tp reinforce a non-wall entity" )
			return false
		end
		
		-- a reinforced wall will not cause a landslide anymore
		wall.isReinforced = true
		
		-- script callback: RR reinforced a wall?
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:close()
		EntityManager:destroyAllEntities()
		if self.terrain then
			self.terrain:destroy()
		end
		
		self.mobSpawner:destroy()
		
		-- remove all jobs
		for i, job in pairs( self.simpleJobs ) do
			Simplejob:remove( job )
			-- print( "game info: job removed" )
		end
		self.simpleJobs = {}
		
		self.selectionState = "global"
		
		self.selectedEntities = {}
		
		-- see trackJob
		self.simpleJobs = {}
		
		self.workerJobs = {}
		
		-- the mouse can leave the game area again
		-- love.mouse.setGrabbed( false )
		
		PostGame:open( self.gamestate )
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:debugDestroySelectedWall()
		
		-- wall selected?
		local wall = self.selectedEntity
		if not wall or wall.entityCategory ~= ENTITYCATEGORY.wall then
			return false
		end
		
		if not wall.canBeDrilled then
			print( "Game: cannot destroy solid rock!" )
			return false
		end
		
		-- remove it
		print( "Game: destroy wall at", wall.xi, wall.yi )
		self:removeWall( wall )
		-- self.selectedTile = nil
		-- self:switchToSelectionState( "tile" )
		-- self:clearEntitySelection()
		self:onEntitySelectionChanged()
		self:reselectCurrentTile()
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:debugReinforceWall( wall )
		wall.isReinforced = true
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- called from terrain
	-- e.g. when a tile was consumed by lava
	function game:onTerrainChanged( xi, yi, type )
		
		if not self:isValidTileIndex( xi, yi ) then
			return false
		end
		
		-- update blocking
		self.blocking:singleTileUpdate( xi, yi, self.terrain, self.wallData )
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:removeWall( wall )
		if not wall then
			return false
		end
		
		-- remember the position and tile index
		local x, y = wall:getPosition()
		local xi, yi = wall.xi, wall.yi
		local entityId = wall.id
		
		-- if the wall was selected, reselect the tile
		local wasSelected = wall.selected == true
		
		EntityManager:remove( wall )
		-- and be sure to update the wallData!
		self.wallData[yi][xi] = 0
		
		-- wall braking animation
		Particles:createWallBreakingParticles( x, y )
		
		-- update blocking
		self.blocking:singleTileUpdate( xi, yi, self.terrain, self.wallData )
		
		-- check up to 8 surrounding walls if they are still valid (or should collapse)
		local wallsToUpdate = {}
		for dy = -1,1 do
			for dx = -1,1 do
				local x = xi+dx
				local y = yi+dy
				if (dx ~= 0 or dy ~= 0)
				and self:isValidTileIndex( x, y ) then
					-- print( "Game: sanityCheckWall at", x, y )
					self:sanityCheckWall( x, y )
				end
			end
		end
		
		-- update wall images
		-- local walls = EntityManager:getEntitiesInCategory( ENTITYCATEGORY.wall )
		-- only need to check walls in direct vicinity
		local walls = EntityManager:getEntitiesFilter( function(e)
			return e.entityCategory == ENTITYCATEGORY.wall
					and math.abs( e.xi-xi ) < 2 and math.abs( e.yi-yi ) < 2
		end )
		self:updateWallImageCodes( walls )
		
		
		-- create debris...
		local debris = EntityManager:createEntity(
			ENTITIES.debris,
			-- also store the tile index (avoid to calculate them over and over again)
			{ x = (xi-1) * tilesize, y = (yi-1) * tilesize, xi = xi, yi = yi }
		)
		if self.fow:isTileDiscovered( xi, yi ) then
			self:addClearDebrisJob( debris )
		end
		
		-- create as many energy crystals as defined in the map data (moved into wall property)
		if wall.numberOfEnergyCrystals and wall.numberOfEnergyCrystals > 0 then
			local spread = 25
			local random = math.random
			-- for i = 1, self.energyCrystals[yi][xi] do
			for i = 1, wall.numberOfEnergyCrystals do
				local energyCrystal = EntityManager:createEntity(
					ENTITIES.energyCrystal,
					{ x = x + random(-spread,spread), y = y + random(-spread,spread) }
				)
				self:onResourceDiscovered( energyCrystal, xi, yi )
			end
		end
		
		
		-- entity selection
		-- if the wall was selected, select the tile (the debris) now
		if wasSelected then
			-- self:onEntitySelectionChanged()
			-- self:reselectCurrentTile()
			self:selectTile( xi, yi )
		end
		
		-- force exploration update
		self:updateFogOfWar()
		
		-- script callback
		self.scriptCallbacks.onWallDestroyed( entityId, xi, yi )
		
	end
	
	
	-- from x1, y1 to x2, y2 (x1 <= x2 and y1 <= y2)
	-- function game:sanityCheckWalls( x1, y1, x2, y2 )
	function game:sanityCheckWall( xi, yi )
		-- local width, height = self.mapSizeX, self.mapSizeY
		local walls = self.wallData
		
		local checkWall = function( xi, yi )
			for i, check in pairs( WallSanityChecks ) do
				local x1 = xi + check[1]
				local y1 = yi + check[2]
				local x2 = xi + check[3]
				local y2 = yi + check[4]
				local x3 = xi + check[5]
				local y3 = yi + check[6]
				if (not self:isValidTileIndex( x1, y1 ) or walls[y1][x1] > 0)
				and (not self:isValidTileIndex( x2, y2 ) or walls[y2][x2] > 0)
				and (not self:isValidTileIndex( x3, y3 ) or walls[y3][x3] > 0) then
					-- one check satisfyed = ok
					return true
				end
			end
			
			-- none of the four checks was successful
			return false
		end
		
		-- 0 = no wall, anything greater than 0 = there is a wall (type does not matter)
		if walls[yi][xi] == 0 then
			-- print( "(1) no wall at", xi, yi )
			return
		end
		
		if checkWall( xi, yi ) then
			-- wall is ok
			-- print( xi, yi, "is fine" )
			return true
		else
			-- wall must be removed
			local walls = EntityManager:getEntitiesFilter( function(e)
				return e.entityCategory == ENTITYCATEGORY.wall
					and e.xi == xi and e.yi == yi
			end )
			if walls[1] then
				-- print( xi, yi, "removed" )
				self:removeWall( walls[1] )
			else
				print( "Game:sanityCheckWall warning - no wall at", xi, yi )
			end
			-- some wall(s) must be removed
			return false
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- check if a wall has a free (no wall) neighbour tile
	function game:hasWallFreeNeighbourTile( wall )
		if not wall or wall.entityCategory ~= ENTITYCATEGORY.wall then
			return false
		end
		local xi, yi = wall.xi, wall.yi
		local hasFreeNeighbour = false
		-- x1,y1, x2,y2, ... for all 4 direct neighbours (not diagonal)
		local neighbourIndexOffsets = { 0,-1, 1,0, 0,1, -1,0 }
		for i = 1, #neighbourIndexOffsets, 2 do
			local x = xi + neighbourIndexOffsets[i]
			local y = yi + neighbourIndexOffsets[i+1]
			if self:isValidTileIndex( x, y ) then
				if self.wallData[y][x] == 0 then
					return true
				end
			end
		end
		return false
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- button callback
	function game:guiActionDrillWall()
		
		-- wall selected?
		local wall = game.selectedEntity
		if not wall or wall.entityCategory ~= ENTITYCATEGORY.wall then
			print( "Game: no wall selected" )
			return false
		end
		
		-- local xi, yi = wall.xi, wall.yi
		
		-- wall can only be destroyed if there is at least one neighbour free (no wall)
		if not self:hasWallFreeNeighbourTile( wall ) then
			print( "Game: cannot drill wall - has no empty neighbour!" )
			return false
		end
		
		-- DEBUG
		if love.keyboard.isDown( "lshift" ) then
			self:debugDestroySelectedWall()
			return false
		end
		-- /DEBUG
		
		-- if wall.entityType == ENTITIES.solidRock then
		if not wall.canBeDrilled then
			print( "Game: cannot destroy this wall!" )
			return false
		end
		
		-- add a task to a job queue
		self:addDrillWallJob( wall )
		return true
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:guiActionReinforceWall()
		
		-- wall selected?
		local wall = game.selectedEntity
		if not wall or wall.entityCategory ~= ENTITYCATEGORY.wall
		or wall.entityType == ENTITIES.solidRock then
			print( "Game: invalid or no wall selected" )
			return false
		end
		
		-- wall can only be reinforced if there is at least one neighbour free (no wall)
		if not self:hasWallFreeNeighbourTile( wall ) then
			print( "Game: cannot reinforce wall - has no empty neighbour!" )
			return false
		end
		
		-- DEBUG
		if love.keyboard.isDown( "lshift" ) then
			self:debugReinforceWall( wall )
			return false
		end
		-- /DEBUG
		
		-- add a task to a job queue
		self:addReinforceWallJob( wall )
		return true
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- button callback
	function game:guiActionTeleportRockRaiderToBuilding()
		
		-- valid building selected?
		local building = self.selectedEntity
		if not building
		or building.entityCategory ~= ENTITYCATEGORY.building then
			return false
		end
		
		if building.entityType == ENTITIES.teleportPad then
			
			if not building:isReadyToTeleportUnit() then
				return
			end
			
			-- print( "Game: teleport RR to teleport pad" )
			
			-- local x, y = building:getPosition()
			local x1, y1 = building:getTeleportUnitPosition()
			-- local x2, y2 = building:getLeaveBuildingPosition()
			
			local rockRaider = EntityManager:createEntity( ENTITIES.rockRaider, { x = x1, y = y1 } )
			-- rockRaider:setLeaveBuildingState( x2, y2 )
			rockRaider:setTeleportToBuildingState( building )
			
			building:teleportUnit()
			
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- button callback
	-- teleport currently selected entity up (effectively expelling it)
	function game:guiActionExpellUnit()
		local entity = self.selectedEntity
		if not entity then
			return false
		end
		
		-- cannot expell vehicle as long as it has a driver
		if entity.entityCategory == ENTITYCATEGORY.vehicle
		and entity:hasDriver() then
			return false
		end
		
		if entity.cancelCurrentJob then
			entity:cancelCurrentJob()
		end
		
		-- TODO: teleport animation
		EntityManager:remove( entity )
		self:clearEntitySelection()	-- TODO?
		-- self:onEntitySelectionChanged()
		return true
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:onBeforeEntityTeleportedUp( entity )
		if self:getSelectedEntity() == entity then
			self:clearEntitySelection()
		end
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- button callback
	function game:guiActionDestroyBuilding()
		local entity = self.selectedEntity
		if not entity
		or entity.entityCategory ~= ENTITYCATEGORY.building then
			return false
		end
		
		-- TODO: cancel current production or research
		-- if entity.cancelCurrentJob then
			-- entity:cancelCurrentJob()
		-- end
		
		if entity.xi and entity.yi and entity.getBlocking then
			self.blocking:removeEntityBlockingAtTile( entity.xi, entity.yi, entity:getBlocking() )
		end
		
		-- create resource entities at this position
		self:createResourcesForBuildingToDestroy( entity )
		
		-- TODO: teleport animation
		EntityManager:remove( entity )
		self:clearEntitySelection()
		return true
	end
	
	
	function game:createResourcesForBuildingToDestroy( entity )
		local x, y = entity:getPosition()
		local costs = {}
		if entity.entityType == ENTITIES.constructionSite then
			-- for construction sites: create only the resources that are already there
			costs = entity:getStoredResources()
		else
			costs = self:getBuildingCost( entity.entityType )
		end
		-- print( "game:createResourcesForBuildingToDestroy", entity.entityType )
		for resourceType, required in pairs( costs ) do
			-- print( resourceType, required )
			for i = 1, required do
				local resourceEntity = EntityManager:createEntity(
					resourceType,
					{ x = x, y = y }
				)
				self:addCollectResourceJob( resourceEntity )
			end
		end
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- button callback
	function game:guiActionVehicleGetDriver()
		
		local vehicle = self.selectedEntity
		if not vehicle
		or vehicle.entityCategory ~= ENTITYCATEGORY.vehicle then
			return false
		end
		
		-- vehicle is already manned or somehow not able to get a driver
		if not vehicle.canBeManned or not vehicle:canBeManned() then
			return false
		end
		
		self:addVehicleGetDriverJob( vehicle )
		return true
		
	end
	
	
	-- button callback
	function game:guiActionVehicleDriverLeave()
		
		local vehicle = self.selectedEntity
		if not vehicle
		or vehicle.entityCategory ~= ENTITYCATEGORY.vehicle then
			return false
		end
		
		if not vehicle:hasDriver() then
			return false
		end
		
		-- check if the driver can safely leave the vehicle there
		-- e.g. not hovering over water or lava
		local x, y = vehicle:getPosition()
		if vehicle.getLeaveVehiclePosition then
			x, y = vehicle:getLeaveVehiclePosition()
		end
		local xi, yi = getTileIndex( x, y )
		local mini_xi, mini_yi = getMiniTileIndex( x, y )
		-- if self.terrain:getTerrainTypeAtTile( xi, yi ) ~= 1 then
		if self.blocking:isMiniTileBlocked( mini_xi, mini_yi ) then
			print( "game:guiActionVehicleDriverLeave - cannot leave here, tile blocked!" )
			return false
		end
		
		self:vehicleDriverLeave( vehicle )
		return true
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:vehicleDriverLeave( vehicle )
		local driver = vehicle:getDriver()
		if driver then
			driver:leaveVehicle()
		end
		self:vehicleSetDriver( vehicle, nil )
	end
	
	
	function game:vehicleSetDriver( vehicle, driver )
		vehicle:setDriver( driver )
		if driver then
			local isDriverSelected = driver.selected
			driver:setVehicle( vehicle )
			-- automatically select the vehicle if the driver was selected
			if isDriverSelected then
				self:selectEntity( vehicle )
			end
		end
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:debugDestroySelectedDebris()
		
		-- debris (tile) selected?
		local debris = self.selectedEntity
		if not debris
		or debris.entityType ~= ENTITIES.debris then
			return false
		end
		
		-- remove it
		EntityManager:remove( debris )
		
		-- self.selectedTile = nil
		-- tile remains selected!
		self:clearEntitySelection()
		self:switchToSelectionState( "tile" )
		self:reselectCurrentTile()
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- at currently selected tile (if any)
	function game:debugCreateBuildingArea()
		
		if not self.selectedTile then
			return false
		end
		
		local xi, yi = self.selectedTile.xi, self.selectedTile.yi
		
		if self.terrain:getTerrainTypeAtTile( xi, yi ) ~= 1 then
			print( "Game: cannot create building area - invalid terrain!", xi, yi )
			return false
		end
		
		if self.selectedEntity and not EntityManager:isValidEntity( self.selectedEntity ) then
			print( "Game:debugCreateBuildingArea - clear entity selection (invalid entity)" )
			self:clearEntitySelection()
		end
		
		if self.selectedEntity then
			print( "Game: cannot create building area, some entity selected!", xi, yi )
			print( self.selectedEntity.id, self.selectedEntity.entityType )
			return false
		end
		
		
		-- create building area ...
		EntityManager:createEntity(
			ENTITIES.buildingArea,
			-- also store the tile index (avoid to calculate them over and over again)
			{ x = (xi-1) * tilesize, y = (yi-1) * tilesize, xi = xi, yi = yi }
		)
		-- print( "Game: created building area at", xi, yi )
		self:reselectCurrentTile()	-- !
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- return the building entity at the given tile or false if there is none
	function game:getBuildingAtTile( xi, yi )
		local buildings = EntityManager:getEntitiesFilter( function(e)
			return e.entityCategory == ENTITYCATEGORY.building
				and e.xi == xi and e.yi == yi
		end )
		return buildings[1] or false
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:guiActionCreateBuilding( entityType )
		
		if not self.selectedTile then
			return false
		end
		
		local xi, yi = self.selectedTile.xi, self.selectedTile.yi
		
		if self.terrain:getTerrainTypeAtTile( xi, yi ) ~= 1 then
			-- print( "Game: cannot create building site - invalid terrain!", xi, yi )
			return false
		end
		
		-- cannot build if there is blocked!
		if self.wallData[yi][xi] > 0 then
			return false
		end
		
		-- there is already a building at this tile...
		if self:getBuildingAtTile( xi, yi ) then
			return false
		end
		
		
		-- look for the closest tool store to create (4?) construction fences
		-- if no toolstore can be reached, the building cannot be placed here!
		local toolStore = self:getClosestEntityOfType( ENTITIES.toolStore, xi, yi )
		if toolStore then
			-- create a construction site + remember which building type will be created
			-- return the entity if the creation succeded
			local constructionSite = self:createBuildingSite( entityType, xi, yi )
			if constructionSite then
				local fx, fy = toolStore:getConstructionFenceSpawnPosition()
				for index = 1, 4 do
					local fence = EntityManager:createEntity(
						ENTITIES.constructionFence,
						{ x = fx, y = fy, constructionSiteId = constructionSite.id, constructionSitePositionIndex = index, }
					)
					-- and create a new job for each of these
					self:addPlaceConstructionFenceJob( fence )
				end
			end
		else
			print( "Game info: cannot create building here - no reachable toolstore." )
		end
		
		
		-- for now, directly create the building, debug only
		-- local building = EntityManager:createEntity( entityType,
			-- { x = (xi-1) * tilesize, y = (yi-1) * tilesize, xi = xi, yi = yi } )
		-- if not building then
			-- print( "Game WARNING - failed to create building!" )
			-- return false
		-- end
		
		-- moved into building code
		-- self:onBuildingCreated( building )
		
		-- success -> close new building menu
		GUImgr:setVisibility( "widGameBuildingMenu", false )
		self:selectTile()	-- also reset the selection
		return true
		
	end
	
	
	-- create the building site for the specified (building) entity type
	function game:createBuildingSite( entityType, xi, yi )
		
		local costs = self:getBuildingCost( entityType )
		local building = EntityManager:createEntity( ENTITIES.constructionSite,
			{ x = (xi-1) * tilesize, y = (yi-1) * tilesize, xi = xi, yi = yi, buildingType = entityType, costs = costs, } )
		if not building then
			print( "Game WARNING - failed to create building site!" )
			return false
		end
		-- building.isConstructionComplete = false	-- TODO: set progress to 0?
		
		-- self:onBuildingCreated( building )
		return building
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:getBuildingCost( entityType )
		if not COSTS[entityType] then
			return {}
		end
		return Tools.copyTable( COSTS[entityType] )
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- called from RR when the fence is placed
	function game:onConstructionFencePlaced( fence )
		
		local constructionSite = EntityManager:getEntityById( fence.constructionSiteId )
		if not constructionSite then
			print( "Game:onConstructionFencePlaced error - mussing construction site" )
			return false
		end
		constructionSite:onFencePlaced()
		
		local xi, yi = constructionSite.xi, constructionSite.yi
		
		-- once all 4 fences are set up, the workers can start to bring resources to the building site
		if constructionSite:areAllFencesPlaced() then
			local costs = self:getBuildingCost( constructionSite.buildingType )
			local toolStore = self:getClosestEntityOfType( ENTITIES.toolStore, xi, yi )
			if toolStore then
				-- for now the position to deposit and to grab resource entities is the same
				local rx, ry = toolStore:getResourceDepositPosition()
				for resourceType, required in pairs( costs ) do
					
					local available = self:getPlayerGlobalResource( resourceType )
					local drain = math.min( required, available )
					self:addToPlayersGlobalResource( resourceType, -drain )
					for i = 1, drain do
						local resourceEntity = EntityManager:createEntity(
							resourceType,
							{ x = rx, y = ry }
						)
						-- add a worker job for each resource
						self:addCollectResourceJob( resourceEntity )
						-- if resourceType == ENTITIES.ore then
							-- self:addCollectOreJob( resourceEntity )
						-- elseif resourceType == ENTITIES.energyCrystal then
							-- self:addCollectEnergyCrystalJob( resourceEntity )
						-- end
					end
					
					-- also create a new job to bring a resource entity of the required type to the construction site
					for i = 1, required do
						self:addBringResourceToBuildingJob( constructionSite, resourceType )
					end
					
				end
			end
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- return the construction site that is closest to the given tile position
	-- that requires at least one more resource of the passed in entity type
	function game:getClosestConstructionSiteThatNeedsResourceType( xi, yi, resourceType )
		local entities = EntityManager:getEntitiesFilter( function( e )
			if e.entityType ~= ENTITIES.constructionSite or not e.isMissingResource
			or not e.areAllFencesPlaced then
				return false
			end
			return e:areAllFencesPlaced() and e:isMissingResource( resourceType )
		end )
		
		return self:getClosestEntityFromTable( entities, xi, yi )
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- actually looks for the entity with the shortest path
	function game:getClosestEntityOfType( entityType, xi, yi )
		local entities = EntityManager:getEntitiesFilter( function( e )
			return e.entityType == entityType
		end )
		
		return self:getClosestEntityFromTable( entities, xi, yi )
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- note: coarse ground pathing only!
	-- pass in a table of entities (that have already been filtered)
	-- return the closest entity to the given tile plus the associated path
	function game:getClosestEntityFromTable( entities, xi, yi )
		local bestPath
		local bestEntity
		for _, entity in pairs( entities ) do
			local x, y = entity:getPosition()
			local target_tx, target_ty = getTileIndex( x, y )
			local path = Pathfinding.getPath( xi, yi, target_tx, target_ty, "ground_coarse" )
			if path and (not bestPath or #path < bestPath.length) then
				bestPath = {
					path = path,
					length = #path,
				}
				bestEntity = entity
			end
		end
		return bestEntity, bestPath
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- vehicle factory menu button callback
	function game:guiActionCreateVehicle( entityType )
		local building = self.selectedEntity
		if not building
		or building.entityType ~= ENTITIES.vehicleFactory then
			print( "Game:guiActionCreateVehicle ERROR - must select a vehicle factory" )
			return false
		end
		
		-- TODO: resource check + it takes some time to produce vehicles
		do
			-- local x, y = building:getPosition()
			local x1, y1 = building:getTeleportUnitPosition()
			local x2, y2 = building:getLeaveBuildingPosition()
			
			local vehicle = EntityManager:createEntity( entityType, { x = x1, y = y1 } )
			vehicle:setLeaveBuildingState( x2, y2 )
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- robot bay menu button callback
	function game:guiActionCreateRobot( entityType )
		local building = self.selectedEntity
		if not building
		or building.entityType ~= ENTITIES.robotBay then
			print( "Game:guiActionCreateRobot ERROR - must select a robot bay" )
			return false
		end
		
		-- TODO: resource check + it takes some time to produce robots
		do
			-- local x, y = building:getPosition()
			local x1, y1 = building:getTeleportUnitPosition()
			local x2, y2 = building:getLeaveBuildingPosition()
			
			local robot = EntityManager:createEntity( entityType, { x = x1, y = y1 } )
			robot:setLeaveBuildingState( x2, y2 )
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- indices of the two priorities to swap
	function game:swapWorkerJobPriorities( p1, p2 )
		
		-- get the associated job types at these two priority indices
		-- note that it is possible that multiple different job types share the same priority!
		for jobType, prio in pairs( self.workerJobPriorities ) do
			if prio == p1 then
				self.workerJobPriorities[jobType] = p2
			elseif prio == p2 then
				self.workerJobPriorities[jobType] = p1
			end
		end
		
		-- swap them
		-- self.workerJobPriorities[prio1] = p2
		-- self.workerJobPriorities[prio2] = p1
		
		-- update the buttons
		self:updatePriorityButtons()
		
	end
	
	
	-- get the job types (table of strings) for the given priority (number)
	-- note that multiple job types may share the same priority level!
	function game:getJobTypesByPriorityLevel( level )
		local jobTypes = {}
		for jobType, prio in pairs( self.workerJobPriorities ) do
			if prio == level then
				table.insert( jobTypes, jobType )
			end
		end
		return jobTypes
	end
	
	
	-- buttons for changing the priority order
	function game:updatePriorityButtons()
		
		-- print( "game:updatePriorityButtons" )
		
		-- set images for all buttons
		-- TODO: amount of buttons?
		for index = 1, 7 do
			local button = GUImgr:getObjectById( "btnGameJobPriority"..index )
			local jobTypes = self:getJobTypesByPriorityLevel( index )
			local jobType = jobTypes[1]
			-- print( index, jobType )
			local imageData = self:getImageDataByWorkerJob( jobType )
			button:setImageData( imageData )
		end
		
	end
	
	
	function game:getImageDataByWorkerJob( jobType )
		if jobType == "drillWall" then
			return Images["icon_drill_wall"]
		elseif jobType == "clearDebris" then
			return Images["icon_clear_debris"]
		elseif jobType == "collectEnergyCrystal" then
			return Images["icon_collect_energy_crystal"]
		elseif jobType == "collectOre" then
			return Images["icon_collect_ore"]
		elseif jobType == "placeConstructionFence" or jobType == "bringResourceToBuilding" then
			return Images["icon_construct_building"]
		elseif jobType == "vehicleGetDriver" then
			return Images["icon_vehicle_get_driver"]
		elseif jobType == "reinforceWall" then
			return Images["icon_reinforce_wall"]
		end
		return nil
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- update the values for resource overview ui elements
	function game:updateResourceOverview()
		
		local energyCrystals = self:getPlayerGlobalResource( ENTITIES.energyCrystal )
		GUImgr:getObjectById( "btnGameResourceAmountEnergyCrystal" ):setText( energyCrystals )
		
		local ore = self:getPlayerGlobalResource( ENTITIES.ore )
		GUImgr:getObjectById( "btnGameResourceAmountOre" ):setText( ore )
		
		local rockSamples = self:getPlayerGlobalResource( ENTITIES.rockSample )
		GUImgr:getObjectById( "btnGameResourceAmountRockSample" ):setText( rockSamples )
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- jump to an arbitrary Rock Raider
	function game:guiActionJumpToRockRaider()
		local rr = EntityManager:getEntitiesFilter( function(e)
			return e.entityType == ENTITIES.rockRaider
		end )
		if #rr == 0 then
			return false
		end
		
		-- for now, select a random unit
		-- TODO: cycle through all units equally
		local entity = rr[math.random(#rr)]
		self:clearEntitySelection()
		self:selectEntity( entity )
		local x, y = entity:getPosition()
		Camera:setPosition( x, y )
		return entity
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- message can be a simple string or a table with colors and strings
	-- optional: maximum display duration
	function game:newNotificationMessage( message, duration )
		local uiElement = GUImgr:getObjectById( "novGameShortMessages" )
		if not uiElement then return end
		uiElement:newMessage( message, duration )
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- boolean: game paused state
	function game:setPaused( paused )
		self.isPaused = paused
		
		-- hide some widgets
		if self.isPaused then
			self:clearEntitySelection()
			GUImgr:setVisibility( "widGameBuildingMenu", false )
			GUImgr:setVisibility( "gameMinimap", false )
			GUImgr:setVisibility( "widGameWorkerJobPriorities", false )
			-- GUImgr:setVisibility( "widGameConfirmDestroyBuilding", false )
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- selection (GUI) state -> moved to the game screen
	-- global = player can select any entity (nothing is currently selected)
	-- wall = a wall is selected
	-- building = a building is selected
	-- entity = at least one entity is selected
	-- tile = an (empty or with debris) tile is selected
	function game:switchToSelectionState( newState )
		
		local oldState = self.selectionState
		
		GameScreen:switchToSelectionState( newState )
		
		-- switch to new state
		self.selectionState = newState
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	Mainloop:add( game )
	-- GUImgr:addScreen( game )
	
	Mouse:addListener( "mousePressed", game )	-- onMousePressed
	Mouse:addListener( "wheelMoved", game )	-- make sure to have onWheelMoved function
	
	UserInput:addListener( "keyPressed", game )	-- onKeyPressed
	
	-- game:switchToSelectionState( "global" )
	
	return game
end


return Game
