local Keybinds = {}

-- on start: check if the keybinds file exists in the user (appdata) directory
-- if not, create a file with the default values
-- if yes, and all entries valid, load them; done
-- if the file exists but not all entries, display a message to inform the user
--> OK: the file will be overwritten with the default values
--> CANCEL: quit the app without overwriting the file (to allow the user to check his keybinds)

-- the file name for the file that the user can edit
-- must be different from the internal file name...
local fileName = "user_keybinds"


local defaults = {}
defaults["toggle_entity_overhead"] = "tab"
defaults["toggle_fps_count"] = "f"
defaults["transparent_walls"] = "t"
defaults["camera_pan_up"] = "up"
defaults["camera_pan_down"] = "down"
defaults["camera_pan_left"] = "left"
defaults["camera_pan_right"] = "right"
defaults["debug_pause"] = "p"


function Keybinds:init()
	local kb = {}
	
	-- list of [key] = action
	kb.actionKey = Tools.copyTable( defaults )
	-- list of [action] = key
	kb.keyAction = {}
	
	
	-- return the contents of the user keybinds file, if present
	function kb:getUserKeybindFile()
		local fileInfo = love.filesystem.getInfo( fileName .. ".lua", "file" )
		if fileInfo then
			local fileData = require( fileName )
			if type(fileData) == "table" then
				return fileData
			end
		end
		return false
	end
	
	
	function kb:validateKeybindFile()
		local keybinds = self:getUserKeybindFile()
		if not keybinds then
			self:writeDefaultKeybindsToFile()
			self:updateKeyActionTable()
			return
		end
		
		-- are all keybinds present?
		local isKeyMissing = false
		for action, keybind in pairs( defaults ) do
			if not keybinds[action] then
				isKeyMissing = true
				print( "keybinds - key missing:", action )
				break
			end
		end
		if isKeyMissing then
			-- inform the user so he can avoid that his keybinds file will be reset
			local btnIndex = love.window.showMessageBox(
				getString("keybinds_overwrite_file_title"),
				getString("keybinds_overwrite_file_text"),
				-- this button list seems kinda confusing...
				{ getString("keybinds_overwrite_file_cancel"), "OK", enterbutton = 2, escapebutton = 1, },
				-- { "OK", getString("keybinds_overwrite_file_cancel") },
				"warning",
				true	-- attach to window
			)
			if btnIndex == 1 then
				-- go on, reset
				self:writeDefaultKeybindsToFile()
				self:updateKeyActionTable()
			else
				-- cancel
				love.event.quit()
			end
			return
		end
		
		-- all keys are present, apply them and check that there are no duplicates
		for action, keybind in pairs( self.actionKey ) do
			self.actionKey[action] = keybinds[action]
		end
		self:updateKeyActionTable()
		
	end
	
	
	function kb:writeDefaultKeybindsToFile()
		local s = "local kb = {}\r\n\r\n"
		for action, keybind in pairs( defaults ) do
			s = string.format( "%s%s%s%s%s%s", s, "kb[\"", action, "\"] = \"", keybind, "\"\r\n" )
		end
		s = s .. "\r\n" .. "return kb\r\n"
		
		-- print( "Writing keybinds file ..." )
		local file = love.filesystem.newFile( fileName .. ".lua" )
		local open_ok, open_error = file:open( "w" )
		if not open_ok then
			print( open_error )
			return false
		end
		local write_ok, write_error = file:write( s )
		if write_ok then
			-- print( "... done." )
		else
			print( write_error )
		end
		file:close()
	end
	
	
	-- get the default keybind for an action
	function kb:getDefaultKey( action )
		if not action then return false end
		return defaults[action]
	end
	
	
	-- get the active keybind for an action
	function kb:getKey( action )
		if not action then return false end
		return self.actionKey[action]
	end
	
	
	-- e.g. a key was pressed, get the associated action (if any)
	function kb:getAction( key )
		if not key then return false end
		return self.keyAction[key]
	end
	
	
	-- not used?
	-- function kb:setKey( action, key )
		-- if not action or not self:getDefaultKey( action ) then
			-- print( "keybinds:setKey error - unknown action:", action )
			-- return false
		-- elseif not key then
			-- print( "keybinds:setKey error - unknown key:", key )
			-- return false
		-- end
		-- self.actionKey[action] = key
		-- self:updateKeyActionTable()
	-- end
	
	
	-- store <keybind><action> pairs
	function kb:updateKeyActionTable()
		self.keyAction = {}
		
		for action, key in pairs( self.actionKey ) do
			if self.keyAction[key] then
				print( "keybinds:updateKeyActionTable - key is already in use!", key )
				love.window.showMessageBox(
					getString("keybinds_no_duplicate_key_title"),
					getString("keybinds_no_duplicate_key_text"),
					"warning"
				)
				love.event.quit()
				return false
			end
			self.keyAction[key] = action
		end
		return true
	end
	
	
	kb:validateKeybindFile()
	
	return kb
end


return Keybinds
