local MapEditor = {}
local lg = love.graphics

local tileSize = 32	-- in pixel

local TOOL_NONE = -1
local TOOL_TERRAIN = 0
local TOOL_WALL_PLACEMENT = 1
local TOOL_WALL_CONFIG = 2
local TOOL_RESOURCE_AMOUNT = 3
-- ...


local WallSanityChecks = {
	{-1,-1, 0,-1, -1,0},	-- top left
	{0,-1, 1,-1, 1,0},	-- top right
	{-1,0, -1,1, 0, 1},	-- bottom left
	{0,1, 1,1, 1,0}	-- bottom right
}


function MapEditor:create()
	local me = {}
	me.id = "map_editor"
	me.z = 0
	me.screenId = "map_editor"
	me.mapData = {
		sizeX = 0,
		sizeY = 0,
		terrain = {},
		walls = {},
		energyCrystals = {},	-- can be inside destructable walls or (rarely) laying on the ground
	}
	me.cameraOffsetX = 0
	me.cameraOffsetY = 0
	me.currentTool = TOOL_NONE
	
	----------------------------------------------------------------------------------------------------
	
	function me:getId()
		return self.id
	end
	
	
	function me:open()
		GUImgr:setScreen( self.id )
		
		self.cameraOffsetX = 0
		self.cameraOffsetY = 0
		
		self:readMapsToLoad()
		
		-- self:prepareMap()
	end
	
	
	function me:close()
		self:cancelTool()
		self:hideAllToolWidgets()
		-- dialogs
		GUImgr:setVisibility( "widMapEditorLoadMap", false )
		GUImgr:setVisibility( "widMapEditorNewMap", false )
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- local tileSize is just how large one fill tile (TILESIZE in world coordinates) is displayed here
	function me:mousePositionToWorldPosition( mx, my )
		local x = mx - self.cameraOffsetX - 0.5 * tileSize
		local y = my - self.cameraOffsetY - 0.5 * tileSize
		local scale = TILESIZE / tileSize
		return x * scale, y * scale
	end
	
	
	-- convert mouse x and y to tile x and y index
	function me:mousePositionToTileIndex( mx, my )
		local x = mx - self.cameraOffsetX
		local y = my - self.cameraOffsetY
		return math.ceil( x / tileSize ), math.ceil( y / tileSize )
	end
	
	
	function me:isValidTileIndex( xi, yi )
		if xi < 1 or xi > self.mapData.sizeX
		or yi < 1 or yi > self.mapData.sizeY then
			return false
		end
		return true
	end
	
	----------------------------------------------------------------------------------------------------
	
	function me:hasMap()
		if not self.mapData then return false end
		if not self.mapData.sizeX then return false end
		return self.mapData.sizeX > 0
	end
	
	----------------------------------------------------------------------------------------------------
	
	function me:tick( dt )
		
		if GUImgr:getCurrentScreenId() ~= self.id then
			return
		end
		
		if love.mouse.isDown( 1 )
		and not GUImgr:isMouseHoveringGuiElement() then
			
			local mx, my = love.mouse.getPosition()
			local xi, yi = self:mousePositionToTileIndex( mx, my )
			
			-- no map created yet; TODO this is not enough
			if not self:hasMap() then return end
			
			-- hold down left mouse button to apply a terrain type
			if self.currentTool == TOOL_TERRAIN
			and self.applyTerrainType then
				
				local terrainType = self.applyTerrainType
				
				-- do not allow to place water or lava underneath walls
				if terrainType == 1 or self.mapData.walls[yi][xi] == 0 then
					self:updateTerrainAtTile( xi, yi, terrainType )
				end
				
				-- water or lava -> remove all energy crystals on this tile
				if terrainType > 1 then
					self.mapData.energyCrystals[yi][xi] = 0
				end
				
			-- hold down left mouse button to place walls
			elseif self.currentTool == TOOL_WALL_PLACEMENT then
				
				local wallType = self.placeWallType
				
				-- can only place walls on normal ground (no water or lava)
				local terrainType = self:getTerrainAtTile( xi, yi )
				if terrainType and terrainType == 1 then
					self:placeWallAtTile( xi, yi, self.placeWallType )
				-- else
					-- print( "MapEditor: cannot place wall here!", xi, yi )
				end
				
				-- no energy crystals inside solid rock
				if wallType == ENTITIES.solidRock then
					self.mapData.energyCrystals[yi][xi] = 0
				end
				
			end
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- fetch the terrain type at a given tile index
	function me:getTerrainAtTile( xi, yi )
		if not self:isValidTileIndex( xi, yi ) then
			return false
		end
		return self.mapData.terrain[yi][xi]
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- set the terrain type at a certain tile (index position)
	function me:updateTerrainAtTile( xi, yi, terrainType )
		if not self:isValidTileIndex( xi, yi ) then
			return false
		end
		
		self.mapData.terrain[yi][xi] = terrainType
		return terrainType
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- set the wall type at a certain tile (index position)
	function me:placeWallAtTile( xi, yi, wallType )
		if not self:isValidTileIndex( xi, yi ) then
			return false
		end
		
		-- instead of clearing walls, do not change anything
		if not wallType then
			return false
		end
		self.mapData.walls[yi][xi] = wallType
		return wallType
	end
	
	----------------------------------------------------------------------------------------------------
	
	function me:prepareMap()
		-- fetch the map size from the value picker elements
		local sizeX = self.vapMapEditorMapSizeX:getValue()
		local sizeY = self.vapMapEditorMapSizeY:getValue()
		
		self.mapData.sizeX = sizeX
		self.mapData.sizeY = sizeY
		
		print( "MapEditor: prepare new map with size", sizeX, sizeY )
		
		for yi = 1, sizeY do
			self.mapData.terrain[yi] = {}
			self.mapData.walls[yi] = {}
			self.mapData.energyCrystals[yi] = {}
			for xi = 1, sizeX do
				self.mapData.terrain[yi][xi] = 1	-- normal ground
				self.mapData.walls[yi][xi] = 0		-- no wall
				self.mapData.energyCrystals[yi][xi] = 0		-- no crystals initially
			end
		end
		
		-- also: place solid walls around the entire map (border)
		for yi = 1, sizeY do
			self.mapData.walls[yi][1] = ENTITIES.solidRock
			self.mapData.walls[yi][sizeX] = ENTITIES.solidRock
		end
		for xi = 1, sizeX do
			self.mapData.walls[1][xi] = ENTITIES.solidRock
			self.mapData.walls[sizeY][xi] = ENTITIES.solidRock
		end
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- button callback
	function me:selectTool( toolType )
		
		-- hide all tool widgets
		self:hideAllToolWidgets()
		
		if toolType == self.currentTool then
			self:cancelTool( self.currentTool )
			return false
		end
		
		-- switch to a different tool (not changed until now)
		self:switchToTool( toolType )
		
	end
	
	-- actually switch to a different tool now
	function me:switchToTool( toolType )
		
		-- update logic (cancel state from previous tool)
		self:cancelTool( self.currentTool )
		
		-- set new tool
		self.currentTool = toolType
		
		-- update widgets
		if toolType == TOOL_TERRAIN then
			GUImgr:setVisibility( "widMapEditorTerrainTool", true )
		elseif toolType == TOOL_WALL_PLACEMENT then
			GUImgr:setVisibility( "widMapEditorPlaceWallTool", true )
		elseif toolType == TOOL_RESOURCE_AMOUNT then
			GUImgr:setVisibility( "widMapEditorResourceAmountTool", true )
		end
		
	end
	
	-- cancel everything that is related to the currently selected tool
	function me:cancelTool( toolType )
		
		self.currentTool = TOOL_NONE
		
		self.applyTerrainType = nil
		self.placeWallType = nil
		self.editResourceType = nil
		
	end
	
	
	function me:hideAllToolWidgets()
		
		GUImgr:setVisibility( "widMapEditorTerrainTool", false )
		GUImgr:setVisibility( "widMapEditorPlaceWallTool", false )
		GUImgr:setVisibility( "widMapEditorResourceAmountTool", false )
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- 1 = normal ground
	-- 2 = water
	-- 3 = lava
	function me:selectTerrainType( terrainType )
		-- print( "Map Editor: apply terrain type", terrainType )
		self.applyTerrainType = terrainType
	end
	
	----------------------------------------------------------------------------------------------------
	
	function me:selectWallType( wallType )
		-- print( "Map Editor: place wall type", wallType )
		self.placeWallType = wallType
	end
	
	----------------------------------------------------------------------------------------------------
	
	function me:selectEditResourceType( resType )
		-- print( "Map Editor: edit resource type", resType )
		self.editResourceType = resType
	end
	
	----------------------------------------------------------------------------------------------------
	
	function me:draw()
		if GUImgr:getCurrentScreenId() ~= self.id then
			return
		end
		
		-- draw the "map"
		local width, height = self.mapData.sizeX, self.mapData.sizeY
		lg.push()
		lg.translate( self.cameraOffsetX, self.cameraOffsetY )
		
		for r = 1, height do
			for c = 1, width do
				
				-- draw terrain
				local tileType = self.mapData.terrain[r][c]
				if tileType == 1 then			-- normal ground
					lg.setColor( COLORS.gray )
				elseif tileType == 2 then	-- water
					lg.setColor( COLORS.blue )
				elseif tileType == 3 then	-- lava
					lg.setColor( COLORS.orange )
				end
				lg.rectangle( "fill", (c-1)*tileSize, (r-1)*tileSize, tileSize, tileSize )
				--lg.setColor( COLORS.white )
				
				-- draw walls
				local wallType = self.mapData.walls[r][c]
				if wallType == ENTITIES.solidRock then
					lg.setColor( COLORS.lightgray )
				elseif wallType == ENTITIES.hardRock then
					lg.setColor( COLORS.darkgray )
				elseif wallType == ENTITIES.looseRock then
					lg.setColor( COLORS.brown )
				end
				-- make them smaller...
				lg.rectangle( "fill", (c-1)*tileSize+2, (r-1)*tileSize+2, tileSize-4, tileSize-4 )
				--lg.setColor( COLORS.white )
				
				-- energy crystal count per tile
				lg.setColor( COLORS.green )
				local energyAmount = self.mapData.energyCrystals[r][c]
				if energyAmount > 0 then
					-- leave a small border for better readability
					lg.print( energyAmount, (c-1)*tileSize + 2, (r-1)*tileSize + 2 )
				end
				
			end
		end
		
		lg.pop()
		
		local mouse_x, mouse_y = love.mouse.getPosition()
		local x, y = self:mousePositionToWorldPosition( mouse_x, mouse_y )
		local xi, yi = self:mousePositionToTileIndex( mouse_x, mouse_y )
		lg.setColor( 0.2, 0.8, 0.4, 0.5 )
		lg.line( 0, mouse_y, SCREEN.width, mouse_y )
		lg.line( mouse_x, 0, mouse_x, SCREEN.height )
		if self:isValidTileIndex( xi, yi ) then
			-- offset and "shadow" for better readability
			local text = string.format( "X:%d Y:%d [%d;%d]", x, y, xi, yi )
			lg.setColor( COLORS.black )
			lg.print( text, mouse_x + 21, mouse_y + 21 )
			lg.setColor( COLORS.white )
			lg.print( text, mouse_x + 20, mouse_y + 20 )
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function me:onMousePressed( x, y, button )
		local xi, yi = self:mousePositionToTileIndex( x, y )
		if not self:isValidTileIndex( xi, yi ) then
			return false
		end
		
		if self.currentTool == TOOL_RESOURCE_AMOUNT then
			
			if self.editResourceType == "energyCrystal" then
				if button == 1 then	-- left click to increase
					self:addEnergyCrystalAtTile( xi, yi, 1 )
				elseif button == 2 then	-- right click to decrease
					self:addEnergyCrystalAtTile( xi, yi, -1 )
				end
			end
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function me:onMouseMoved( dx, dy )
		
		if GUImgr:getCurrentScreenId() ~= self.id then
			return
		end
		
		-- press middle mouse button to move the map
		if love.mouse.isDown( 3 ) then
			self.cameraOffsetX = self.cameraOffsetX + dx
			self.cameraOffsetY = self.cameraOffsetY + dy
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- amount can also be negative
	function me:addEnergyCrystalAtTile( xi, yi, amount )
		if not self:isValidTileIndex( xi, yi ) then
			return false
		end
		
		-- make sure the amount does not exceed a limit, or is below 0
		local newAmount = Tools.clamp( self.mapData.energyCrystals[yi][xi] + amount, 0, 10 )
		
		-- no crystals if there is water or lava or solid rock (indestructible)
		if self:getTerrainAtTile( xi, yi ) ~= 1 then
			newAmount = 0
		end
		
		if self.mapData.walls[yi][xi] == ENTITIES.solidRock then
			newAmount = 0
		end
		
		self.mapData.energyCrystals[yi][xi] = newAmount
		return newAmount
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- for all walls: check that it has enough neighbouring walls
	-- that is, at least one in x and one in y direction
	function me:sanityCheckWalls()
		if not self:hasMap() then return false end
		local width, height = self.mapData.sizeX, self.mapData.sizeY
		local walls = self.mapData.walls
		local bad_walls = 0
		
		local checkWall = function( xi, yi )
			for i, check in pairs( WallSanityChecks ) do
				local x1 = xi + check[1]
				local y1 = yi + check[2]
				local x2 = xi + check[3]
				local y2 = yi + check[4]
				local x3 = xi + check[5]
				local y3 = yi + check[6]
				if (not self:isValidTileIndex( x1, y1 ) or walls[y1][x1] > 0)
				and (not self:isValidTileIndex( x2, y2 ) or walls[y2][x2] > 0)
				and (not self:isValidTileIndex( x3, y3 ) or walls[y3][x3] > 0) then
					-- one check satisfyed = ok
					return true
				end
			end
			
			-- none of the four checks was successful
			return false
		end
		
		local wallsToRemove = {}
		
		for yi = 1, height do
			for xi = 1, width do
				-- 0 = no wall, anything greater than 0 = there is a wall (type does not matter)
				if walls[yi][xi] > 0 then
					if not checkWall( xi, yi ) then
						bad_walls = bad_walls + 1
						print( "MapEditor: bad wall at", xi, yi )
						table.insert( wallsToRemove, {xi,yi} )
					end
				end
			end
		end
		
		-- remove the bad walls
		for k, v in pairs( wallsToRemove ) do
			local xi = v[1]
			local yi = v[2]
			walls[yi][xi] = 0
		end
		
		if bad_walls > 0 then
			print( "MapEditor: removed bad walls:", bad_walls )
		else
			print( "MapEditor: no bad walls to remove." )
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- store map data in a lua file
	-- that can be loaded again using require instead of reading and evaluating strings
	function me:saveMap()
		
		local map = self.mapData
		
		if not map then return false end
		
		local file = love.filesystem.newFile( "maps/rr_map_" .. os.time() .. ".lua" )
		
		local open_ok, open_error = file:open( "w" )
		if not open_ok then
			print( open_error )
			return false
		end
		
		-- generic map info
		local saveDataString = string.format(
			"return {\n sizeX=%d,\n sizeY=%d,\n ",
			map.sizeX,
			map.sizeY
		)
		
		-- terrain
		saveDataString = saveDataString .. "\n terrain={"
		for y = 1, map.sizeY do
			saveDataString = saveDataString ..
			"\n  {" .. table.concat( map.terrain[y], ", " ) .. "},"
		end
		saveDataString = saveDataString .. "\n },\n"
		
		-- walls
		saveDataString = saveDataString .. "\n walls={"
		for y = 1, map.sizeY do
			saveDataString = saveDataString ..
			"\n  {" .. table.concat( map.walls[y], ", " ) .. "},"
		end
		saveDataString = saveDataString .. "\n },\n"
		
		-- energy crystals
		saveDataString = saveDataString .. "\n energyCrystals={"
		for y = 1, map.sizeY do
			saveDataString = saveDataString ..
			"\n  {" .. table.concat( map.energyCrystals[y], ", " ) .. "},"
		end
		saveDataString = saveDataString .. "\n },\n"
		
		-- buildings?
		
		
		-- conclude file...
		saveDataString = saveDataString .. "\n}"
		
		local write_ok, write_error = file:write( saveDataString )
		local path = file:getFilename()
		
		if write_ok then
			print( "map editor - saved map to", path )
			file:close()
			return true
		else
			print( write_error )
			file:close()
			return false
		end
		
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- check what maps are available
	function me:readMapsToLoad()
		MapLoader:loadMapsFromMapFolder()
		local names, n = MapLoader:getMapFileNames()
		self.lslMapEditorLoadMap:setItems( names )
	end
	
	
	-- note: fileName without extension
	function me:loadMap( fileName )
		
		local map = Tools.copyTable( MapLoader:loadMap( fileName ) )
		if not map then
			print( "map editor: failed to load map data for", fileName )
			return false
		end
		
		-- load terrain and walls etc.
		self.mapData = {
			sizeX = map.sizeX,
			sizeY = map.sizeY,
			terrain = map.terrain,
			walls = map.walls,
			energyCrystals = map.energyCrystals,
		}
		
		-- success - close widget
		GUImgr:setVisibility( "widMapEditorLoadMap", false )
		return true
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- CREATE A NEW MAP DIALOG
	Button:create{
		id = "btnMapEditorCreateNewMapDialog",
		x = SCREEN.width * 0.9,
		y = SCREEN.height * 0.15,
		sizeX = SCREEN.width * 0.19,
		sizeY = SCREEN.height * 0.09,
		text = getString( "map_editor_new_map" ),
		fSize = 22,
		callback = function()
			local vis = GUImgr:isVisible( "widMapEditorNewMap" )
			GUImgr:setVisibility( "widMapEditorNewMap", not vis )
			
			GUImgr:setVisibility( "widMapEditorLoadMap", false )
		end,
		screenId = "map_editor",
	}
	
	-- SAVE CURRENT MAP TO FILE
	Button:create{
		id = "btnMapEditorSaveMap",
		x = SCREEN.width * 0.9,
		y = SCREEN.height * 0.25,
		sizeX = SCREEN.width * 0.19,
		sizeY = SCREEN.height * 0.09,
		text = getString( "map_editor_save_map" ),
		fSize = 22,
		callback = function()
			me:sanityCheckWalls()
			me:saveMap()
		end,
		screenId = "map_editor",
	}
	
	-- LOAD MAP FROM FILE
	Button:create{
		id = "btnMapEditorToggleLoadMapWidget",
		x = SCREEN.width * 0.9,
		y = SCREEN.height * 0.35,
		sizeX = SCREEN.width * 0.19,
		sizeY = SCREEN.height * 0.09,
		text = getString( "map_editor_load_map" ),
		fSize = 22,
		callback = function()
			local vis = GUImgr:isVisible( "widMapEditorLoadMap" )
			GUImgr:setVisibility( "widMapEditorLoadMap", not vis )
			
			GUImgr:setVisibility( "widMapEditorNewMap", false )
		end,
		screenId = "map_editor",
	}
	
	-- QUIT TO MAIN MENU
	Button:create{
		id = "btnMapEditorQuit",
		x = SCREEN.width * 0.9,
		y = SCREEN.height * 0.95,
		sizeX = SCREEN.width * 0.19,
		sizeY = SCREEN.height * 0.09,
		text = getString( "map_editor_quit" ),
		fSize = 22,
		callback = function()
			me:close()
			MainMenu:open()
		end,
		screenId = "map_editor",
	}
	
	----------------------------------------------------------------------------------------------------
	
	-- CREATE NEW MAP - SELECT SIZE...
	me.widMapEditorNewMap = Widget:create{
		id = "widMapEditorNewMap",
		x = SCREEN.width * 0.5,
		y = SCREEN.height * 0.5,
		sizeX = SCREEN.width * 0.4,
		sizeY = SCREEN.height * 0.3,
		screenId = "map_editor",
		visible = false,
	}
	
	-- SELECT MAP WIDTH
	me.vapMapEditorMapSizeX = ValuePicker:create{
		id = "vapMapEditorMapSizeX",
		x = SCREEN.width * 0.4,
		y = SCREEN.height * 0.45,
		sizeX = SCREEN.width * 0.1,
		sizeY = SCREEN.height * 0.05,
		caption = "X",
		minValue = 16,
		maxValue = 64,
		increment = 2,
		screenId = "map_editor",
		parentId = "widMapEditorNewMap",
	}
	
	-- SELECT MAP HEIGHT
	me.vapMapEditorMapSizeY = ValuePicker:create{
		id = "vapMapEditorMapSizeY",
		x = SCREEN.width * 0.6,
		y = SCREEN.height * 0.45,
		sizeX = SCREEN.width * 0.1,
		sizeY = SCREEN.height * 0.05,
		caption = "Y",
		minValue = 16,
		maxValue = 64,
		increment = 2,
		screenId = "map_editor",
		parentId = "widMapEditorNewMap",
	}
	
	-- CREATE A NEW MAP
	Button:create{
		id = "btnMapEditorCreateNewMap",
		x = SCREEN.width * 0.5,
		y = SCREEN.height * 0.55,
		sizeX = SCREEN.width * 0.19,
		sizeY = SCREEN.height * 0.09,
		text = getString( "map_editor_create_map" ),
		fSize = 22,
		callback = function()
			me:prepareMap()
			GUImgr:setVisibility( "widMapEditorNewMap", false )
		end,
		screenId = "map_editor",
		parentId = "widMapEditorNewMap",
	}
	
	----------------------------------------------------------------------------------------------------
	
	-- LOAD MAP
	me.widMapEditorLoadMap = Widget:create{
		id = "widMapEditorLoadMap",
		x = SCREEN.width * 0.5,
		y = SCREEN.height * 0.5,
		sizeX = SCREEN.width * 0.4,
		sizeY = SCREEN.height * 0.6,
		screenId = "map_editor",
		visible = false,
	}
	
	-- MAP FILES LIST
	me.lslMapEditorLoadMap = Listselect:create{
		id = "lslMapEditorLoadMap",
		screenId = "map_editor",
		parentId = "widMapEditorLoadMap",
		x = 0.5 * SCREEN.width,
		y = 0.45 * SCREEN.height,
		maxItems = 8,
		sizeX = 0.3 * SCREEN.width,
		sizeY = 0.4 * SCREEN.height,
		onItemSelected = function( item )
			
			me.mapFileToLoad = item
			if not item then
				return false
			end
			
			--print( item )
			GUImgr:setEnabledFlag( "btnMapEditorLoadMap", true )
		end,
	}
	
	-- LOAD A SELECTED MAP
	Button:create{
		id = "btnMapEditorLoadMap",
		x = SCREEN.width * 0.5,
		y = SCREEN.height * 0.7,
		sizeX = SCREEN.width * 0.19,
		sizeY = SCREEN.height * 0.09,
		text = getString( "map_editor_open_map" ),
		fSize = 22,
		callback = function()
			if me.mapFileToLoad then
				me:loadMap( me.mapFileToLoad )
			end
		end,
		screenId = "map_editor",
		parentId = "widMapEditorLoadMap",
	}
	
	----------------------------------------------------------------------------------------------------
	
	-- SELECT TERRAIN TOOL
	Button:create{
		id = "btnMapEditorTerrainTool",
		x = SCREEN.width * 0.05,
		y = SCREEN.height * 0.05,
		sizeX = SCREEN.width * 0.09,
		sizeY = SCREEN.height * 0.09,
		text = getString( "map_editor_terrain_tool" ),
		fSize = 18,
		callback = function()
			me:selectTool( TOOL_TERRAIN )
		end,
		screenId = "map_editor",
	}
	
	-- SELECT WALL PLACEMENT TOOL
	Button:create{
		id = "btnMapEditorPlaceWallTool",
		x = SCREEN.width * 0.05,
		y = SCREEN.height * 0.15,
		sizeX = SCREEN.width * 0.09,
		sizeY = SCREEN.height * 0.09,
		text = getString( "map_editor_wall_tool" ),
		fSize = 18,
		callback = function()
			me:selectTool( TOOL_WALL_PLACEMENT )
		end,
		screenId = "map_editor",
	}
	
	-- SELECT RESOURCE AMOUNT TOOL
	Button:create{
		id = "btnMapEditorResourceAmountTool",
		x = SCREEN.width * 0.05,
		y = SCREEN.height * 0.25,
		sizeX = SCREEN.width * 0.09,
		sizeY = SCREEN.height * 0.09,
		text = getString( "map_editor_resource_tool" ),
		fSize = 18,
		callback = function()
			me:selectTool( TOOL_RESOURCE_AMOUNT )
		end,
		screenId = "map_editor",
	}
	
	----------------------------------------------------------------------------------------------------
	
	-- TOOLS: TERRAIN
	me.widMapEditorTerrainTool = Widget:create{
		id = "widMapEditorTerrainTool",
		x = SCREEN.width * 0.15,
		y = SCREEN.height * 0.15,
		sizeX = SCREEN.width * 0.1,
		sizeY = SCREEN.height * 0.3,
		screenId = "map_editor",
		visible = false,
	}
	
	-- SELECT TERRAIN: NORMAL GROUND
	me.btnMapEditorTerrainGround = Button:create{
		id = "btnMapEditorTerrainGround",
		x = SCREEN.width * 0.15,
		y = SCREEN.height * 0.05,
		sizeX = SCREEN.width * 0.09,
		sizeY = SCREEN.height * 0.09,
		autoSizeSquare = true,
		-- text = "Boden",
		-- fSize = 22,
		imageData = Images["icon_terrain_dirt"],
		callback = function()
			me:selectTerrainType( 1 )
		end,
		screenId = "map_editor",
		parentId = "widMapEditorTerrainTool",
	}
	
	-- SELECT TERRAIN: WATER
	me.btnMapEditorTerrainWater = Button:create{
		id = "btnMapEditorTerrainWater",
		x = SCREEN.width * 0.15,
		y = SCREEN.height * 0.15,
		sizeX = SCREEN.width * 0.09,
		sizeY = SCREEN.height * 0.09,
		autoSizeSquare = true,
		-- text = "Wasser",
		-- fSize = 22,
		imageData = Images["icon_terrain_water"],
		callback = function()
			me:selectTerrainType( 2 )
		end,
		screenId = "map_editor",
		parentId = "widMapEditorTerrainTool",
	}
	
	-- SELECT TERRAIN: LAVA
	me.btnMapEditorTerrainLava = Button:create{
		id = "btnMapEditorTerrainLava",
		x = SCREEN.width * 0.15,
		y = SCREEN.height * 0.25,
		sizeX = SCREEN.width * 0.09,
		sizeY = SCREEN.height * 0.09,
		autoSizeSquare = true,
		-- text = "Lava",
		-- fSize = 22,
		imageData = Images["icon_terrain_lava"],
		callback = function()
			me:selectTerrainType( 3 )
		end,
		screenId = "map_editor",
		parentId = "widMapEditorTerrainTool",
	}
	
	----------------------------------------------------------------------------------------------------
	
	-- TOOLS: PLACE WALLS
	me.widMapEditorPlaceWallTool = Widget:create{
		id = "widMapEditorPlaceWallTool",
		x = SCREEN.width * 0.15,
		y = SCREEN.height * 0.3,
		sizeX = SCREEN.width * 0.1,
		sizeY = SCREEN.height * 0.4,
		screenId = "map_editor",
		visible = false,
	}
	
	-- PLACE WALL: NONE
	me.btnMapEditorWallNone = Button:create{
		id = "btnMapEditorWallNone",
		x = SCREEN.width * 0.15,
		y = SCREEN.height * 0.15,
		sizeX = SCREEN.width * 0.09,
		sizeY = SCREEN.height * 0.09,
		autoSizeSquare = true,
		-- text = "(frei)",
		-- fSize = 22,
		imageData = Images["icon_no"],
		callback = function()
			me:selectWallType( 0 )
		end,
		screenId = "map_editor",
		parentId = "widMapEditorPlaceWallTool",
	}
	
	-- PLACE WALL: SOLID ROCK
	me.btnMapEditorWallSolidRock = Button:create{
		id = "btnMapEditorWallSolidRock",
		x = SCREEN.width * 0.15,
		y = SCREEN.height * 0.25,
		sizeX = SCREEN.width * 0.09,
		sizeY = SCREEN.height * 0.09,
		autoSizeSquare = true,
		-- text = "Solid Rock",
		-- fSize = 22,
		imageData = Images["icon_solid_rock"],
		callback = function()
			me:selectWallType( ENTITIES.solidRock )
		end,
		screenId = "map_editor",
		parentId = "widMapEditorPlaceWallTool",
	}
	
	-- PLACE WALL: HARD ROCK
	me.btnMapEditorWallHardRock = Button:create{
		id = "btnMapEditorWallHardRock",
		x = SCREEN.width * 0.15,
		y = SCREEN.height * 0.35,
		sizeX = SCREEN.width * 0.09,
		sizeY = SCREEN.height * 0.09,
		autoSizeSquare = true,
		-- text = "Hard Rock",
		-- fSize = 22,
		imageData = Images["icon_hard_rock"],
		callback = function()
			me:selectWallType( ENTITIES.hardRock )
		end,
		screenId = "map_editor",
		parentId = "widMapEditorPlaceWallTool",
	}
	
	-- PLACE WALL: LOOSE ROCK
	me.btnMapEditorWallLooseRock = Button:create{
		id = "btnMapEditorWallLooseRock",
		x = SCREEN.width * 0.15,
		y = SCREEN.height * 0.45,
		sizeX = SCREEN.width * 0.09,
		sizeY = SCREEN.height * 0.09,
		autoSizeSquare = true,
		-- text = "Loose Rock",
		-- fSize = 22,
		imageData = Images["icon_loose_rock"],
		callback = function()
			me:selectWallType( ENTITIES.looseRock )
		end,
		screenId = "map_editor",
		parentId = "widMapEditorPlaceWallTool",
	}
	
	----------------------------------------------------------------------------------------------------
	
	-- TOOLS: RESOURCE AMOUNT
	me.widMapEditorResourceAmountTool = Widget:create{
		id = "widMapEditorResourceAmountTool",
		x = SCREEN.width * 0.15,
		y = SCREEN.height * 0.25,
		sizeX = SCREEN.width * 0.1,
		sizeY = SCREEN.height * 0.1,
		screenId = "map_editor",
		visible = false,
	}
	
	-- ENERGY CRYSTALS
	me.btnMapEditorEnergyCrystalAmount = Button:create{
		id = "btnMapEditorEnergyCrystalAmount",
		x = SCREEN.width * 0.15,
		y = SCREEN.height * 0.25,
		sizeX = SCREEN.width * 0.09,
		sizeY = SCREEN.height * 0.09,
		autoSizeSquare = true,
		-- text = "Energiekristalle",
		-- fSize = 22,
		imageData = Images["icon_collect_energy_crystal"],
		callback = function()
			me:selectEditResourceType( "energyCrystal" )
		end,
		screenId = "map_editor",
		parentId = "widMapEditorResourceAmountTool",
	}
	
	----------------------------------------------------------------------------------------------------
	
	Button:create{
		id = "btnMapEditorSanityCheckWalls",
		x = SCREEN.width * 0.05,
		y = SCREEN.height * 0.45,
		sizeX = SCREEN.width * 0.09,
		sizeY = SCREEN.height * 0.09,
		text = getString( "map_editor_sanity_check_walls" ),
		fSize = 22,
		callback = function()
			me:sanityCheckWalls()
		end,
		screenId = "map_editor",
		tooltip = {
			text = getString( "map_editor_sanity_check_walls_tooltip" ),
		},
	}
	
	----------------------------------------------------------------------------------------------------
	
	Mainloop:add( me )
	GUImgr:addScreen( me )
	Renderer:add( me, "mapeditor" )
	
	Mouse:addListener( "mousePressed", me )	-- onMousePressed
	-- Mouse:addListener( "mouseReleased", me )	-- onMouseReleased
	Mouse:addListener( "mouseMoved", me )	-- onMouseMoved
	
	return me
end

return MapEditor
