local Renderer = {}
local lg = love.graphics

-- make sure the sort function is valid
local sortFunction = function( a, b )
	local ax = getRelativeX(a)
	local ay = getRelativeY(a)
	local bx = getRelativeX(b)
	local by = getRelativeY(b)
	-- when multiple entities are on the exact same location, sort by their entity ID
	if ax + ay == bx + by then return a.id < b.id end
	return ax + ay < bx + by
end

function getRelativeX( entity )
	local z = entity.body.z or 0
	return entity.body.x + z
end

function getRelativeY( entity )
	local z = entity.body.z or 0
	return entity.body.y + z
end


function Renderer:create()
	local renderer = {
		objects = {
			terrain = {},
			ground_entities = {},
			entities = {},
			projectiles = {},
			mapeditor = {},
			particles = {},
			overhead = {},
		},
		lightCanvas = lg.newCanvas( lg.getWidth(), lg.getHeight() ),
		ambienteLight = { 0.3, 0.2, 0.2 },
	}
	
	
	-------------------------------------------------------------------------------------------------------------------
	
	function renderer:add( obj, layerName )
		assert( obj.draw )
		assert( self.objects[layerName] )
		if obj.visible == nil then obj.visible = true end
		table.insert( self.objects[layerName], obj )
	end
	
	
	function renderer:remove( _obj, layerName )
		assert( self.objects[layerName] )
		for i, obj in pairs( self.objects[layerName] ) do
			if obj == _obj then
				table.remove( self.objects[layerName], i )
			end
		end
	end
	
	-------------------------------------------------------------------------------------------------------------------
	
	function renderer:draw()
		
		local screen = GUImgr:getCurrentScreenId()
		
		if screen == "game" then
			
			-- set camera
			Camera:set()
			
			-- light canvas to make everything dark
			local drawLights = true
			if drawLights then
				lg.setCanvas( self.lightCanvas )
				local light = self.ambienteLight
				lg.clear( light )
				lg.setCanvas()
			end
			
			-- draw terrain first
			for i, obj in pairs( self.objects.terrain ) do
				lg.setColor( 1, 1, 1 )
				obj:draw()
			end
			
			-- draw entities that are directly on top of the terrain and always below other entities
			-- such as debris and building areas
			-- before sorting the table, check if the entity is within the area that is visible on screen
			local groundEntitiesToDraw = {}
			for i, obj in pairs( self.objects.ground_entities ) do
				local x, y = obj:getPosition()
				local xi, yi = getTileIndex( x, y )
				if Game.fow:isTileExplored( xi, yi )
				or obj.drawInFogOfWar and Game.fow:isTileDiscovered( xi, yi ) then
					if Camera:isWorldPositionOnScreen( obj:getPosition() ) then
						table.insert( groundEntitiesToDraw, obj )
					end
				end
			end
			table.sort( groundEntitiesToDraw, sortFunction )
			for i, obj in pairs( groundEntitiesToDraw ) do
				lg.setColor( 1, 1, 1 )
				obj:draw()
			end
			
			-- all other entities (projectiles too?)
			-- before sorting the table, check if the entity is within the area that is visible on screen
			-- draw entities only if the tile is currently explored
			-- some entities (walls) can also be drawn when the tile has been discovered (but is not explored now)
			local entitiesToDraw = {}
			for i, obj in pairs( self.objects.entities ) do
				local x, y = obj:getPosition()
				local xi, yi = getTileIndex( x, y )
				if Game.fow:isTileExplored( xi, yi )
				or obj.drawInFogOfWar and Game.fow:isTileDiscovered( xi, yi ) then
					if Camera:isWorldPositionOnScreen( obj:getPosition() ) then
						table.insert( entitiesToDraw, obj )
					end
				end
			end
			table.sort( entitiesToDraw, sortFunction )
			for i, obj in pairs( entitiesToDraw ) do
				lg.setColor( 1, 1, 1 )
				obj:draw()
			end
			
			-- projectiles
			for i, obj in pairs( self.objects.projectiles ) do
				local x, y = obj:getPosition()
				local xi, yi = getTileIndex( x, y )
				if obj.drawInFogOfWar or Game.fow:isTileDiscovered( xi, yi ) then
					lg.setColor( 1, 1, 1 )
					obj:draw()
				end
			end
			
			-- particles
			for i, obj in pairs( self.objects.particles ) do
				lg.setColor( 1, 1, 1 )
				obj:draw()
			end
			
			-- overhead widgets
			for i, obj in pairs( self.objects.overhead ) do
				lg.setColor( 1, 1, 1 )
				obj:draw()
			end
			
			
			-- light around mouse
			if not Game.isPaused then
				local mouseX, mouseY = love.mouse.getPosition()
				local x, y = Camera:getTransform():inverseTransformPoint( mouseX, mouseY )
				-- do not need convertTo2D here
				local liSize = 900
				local alpha = 0.7
				local color = { 0.8, 0.7, 0.7, alpha }
				self:drawCircularLight( x, y, liSize, color )
			end
			
			-- unset camera
			Camera:unset()
			
			if drawLights then
				lg.setColor( 1, 1, 1 )
				lg.setBlendMode( "multiply", "premultiplied" )
				lg.draw( self.lightCanvas )
				lg.setBlendMode( "alpha" )
			end
			
		elseif screen == "map_editor" then
			
			for i, obj in pairs( self.objects.mapeditor ) do
				lg.setColor( 1, 1, 1 )
				obj:draw()
			end
			
		end
		
	end
	
	-------------------------------------------------------------------------------------------------------------------
	
	-- can be called from within entity draw function
	function renderer:drawCircularLight( x, y, radius, color )
		local color = color or { 1, 0.7, 0.7, 0.8 }
		local liSize = radius or 600
		local angle = angle or 0
		local lightImage = Images["light_round"]
		local scaleX, scaleY = liSize/lightImage.width, liSize/lightImage.height
		lg.setCanvas( self.lightCanvas )
		lg.setColor( color )
		lg.draw( lightImage.image, x, y, angle, scaleX, scaleY, liSize * 0.5 / scaleX, liSize * 0.5 / scaleY )
		lg.setCanvas()
		lg.setColor( 1, 1, 1 )
	end
	
	function renderer:drawConeLight( x, y, radius, angle, color )
		local color = color or { 1, 0.7, 0.7, 0.8 }
		local liSize = radius or 600
		local angle = angle or 0
		local lightImage = Images["light_cone"]
		local scaleX, scaleY = liSize/lightImage.width, liSize/lightImage.height
		lg.setCanvas( self.lightCanvas )
		lg.setColor( color )
		lg.draw( lightImage.image, x, y, angle, scaleX, scaleY, liSize * 0.5 / scaleX, liSize * 0.5 / scaleY )
		lg.setCanvas()
		lg.setColor( 1, 1, 1 )
	end
	
	
	return renderer
end

return Renderer
