local Terrain = {}
local lg = love.graphics
local random = math.random

local tilesize = TILESIZE

-- TODO: add different themes such as ice (for both walls and ground)

-- note: images are rendered in studio with size 256x256 and scale 2.0 (final size is 250x circa 128)

function Terrain:create( mapData )
	local terrain = {}
	
	assert( mapData.sizeX )
	assert( mapData.sizeY )
	assert( mapData.terrain )
	
	terrain.sizeX = mapData.sizeX
	terrain.sizeY = mapData.sizeY
	-- terrain.tiles = Tools.copyTable( mapData.terrain )
	terrain.tiles = {}
	-- for yi = 1, terrain.sizeY do
		-- terrain.tiles[yi] = {}
		-- for xi = 1, terrain.sizeX do
			-- terrain.tiles[yi][xi] = {}
		-- end
	-- end
	terrain.selectedTile = { xi = 0, yi = 0 }
	terrain.animationTick = 0
	
	----------------------------------------------------------------------------------------------------
	
	function terrain:init( terrainData )
		for yi = 1, terrain.sizeY do
			self.tiles[yi] = {}
			for xi = 1, terrain.sizeX do
				self.tiles[yi][xi] = self:createTile( xi, yi, terrainData[yi][xi] )
			end
		end
	end
	
	
	function terrain:createTile( xi, yi, type )
		local tile = {
			xi = xi,
			yi = yi,
			type = type,
			tick = 0,
			lavaResistance = 100,	-- TODO
		}
		tile.animation = self:getAnimationByTerrainType( type )
		if type == 1 then
			tile.animated = false	-- no dynamic (real) animation for dirt, just a static image
		elseif type == 2 then
			tile.animated = true
		elseif type == 3 then
			tile.animated = true
		end
		local x = (xi-1) * tilesize
		local y = (yi-1) * tilesize
		x, y = convertToDimetric( x, y )
		tile.animation:setPosition( x, y )
		tile.animation:setRelativeOrigin( 0.5, 0.5 )
		return tile
	end
	
	
	-- TODO: terrain theme
	function terrain:getAnimationByTerrainType( type )
		if type == 1 then
			return Animation.new( "animations/terrain/dirt.lua" )
		elseif type == 2 then
			return Animation.new( "animations/terrain/water.lua" )
		elseif type == 3 then
			return Animation.new( "animations/terrain/lava.lua" )
		end
		print( "terrain:getAnimationByTerrainType - warning: no animation for terrain type", type )
		return false
	end
	
	----------------------------------------------------------------------------------------------------
	
	function terrain:getTile( xi, yi )
		if not self:isValidTileIndex( xi, yi ) then return false end
		return self.tiles[yi][xi]
	end
	
	----------------------------------------------------------------------------------------------------
	
	function terrain:isValidTileIndex( xi, yi )
		if xi < 1 or xi > self.sizeX
		or yi < 1 or yi > self.sizeY then
			return false
		end
		return true
	end
	
	----------------------------------------------------------------------------------------------------
	
	function terrain:getTerrainTypeAtTile( xi, yi )
		if xi < 1 or xi > self.sizeX
		or yi < 1 or yi > self.sizeY then
			return 0
		end
		return self.tiles[yi][xi].type
	end
	
	----------------------------------------------------------------------------------------------------
	
	function terrain:setTerrainTypeAtTile( xi, yi, type )
		if not self:isValidTileIndex( xi, yi ) then return false end
		
		local tile = self.tiles[yi][xi]
		
		-- nothing changes
		if tile.type == type then return end
		
		self.tiles[yi][xi] = self:createTile( xi, yi, type )
		Game:onTerrainChanged( xi, yi, type )
		return true
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- e.g. check if a tile has water next to it
	function terrain:hasNeighbourOfType( xi, yi, type )
		if not self:isValidTileIndex( xi, yi ) then return false end
		
		for dx = -1, 1 do
			for dy = -1, 1 do
				if dx ~= 0 or dy ~= 0 then
					local x = xi + dx
					local y = yi + dy
					local t = self:getTerrainTypeAtTile( x, y )
					if t > 0 and t == type then
						return true
					end
				end
			end
		end
		return false
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- can be nil = reset
	function terrain:selectTile( xi, yi )
		if xi then
			self.selectedTile = { xi = xi, yi = yi }
		else
			self.selectedTile = { xi = 0, yi = 0 }
		end
		self.animationTick = 0
	end
	
	----------------------------------------------------------------------------------------------------
	
	function terrain:tick( dt )
		
		self.animationTick = self.animationTick + dt
		
		local fow = Game.fow
		for yi = 1, self.sizeY do
			for xi = 1, self.sizeX do
				local tile = self.tiles[yi][xi]
				
				-- update animation for tile that is not currently explored or not?
				if tile.animated
				and fow:isTileDiscovered( xi, yi )
				and Camera:isWorldPositionOnScreen( (xi-1) * tilesize, (yi-1) * tilesize ) then
					tile.animation:update( dt )
				end
				
				if tile.type == 1 then
					if tile.lavaResistance < 33 then
						tile.animation:setState( "very_erosive" )
					elseif tile.lavaResistance < 66 then
						tile.animation:setState( "erosive" )
					end
				end
			end
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function terrain:lavaSpread( tile )
		if tile.type ~= 3 then return false end
		-- attempts will fail 99% of the time
		-- if random() < 0.99 then return false end
		
		local neighbours = {}
		for dx = -1, 1 do
			for dy = -1, 1 do
				if dx ~= 0 or dy ~= 0 then
					local x = tile.xi + dx
					local y = tile.yi + dy
					if self:isValidTileIndex( x, y ) then
						table.insert( neighbours, self.tiles[y][x] )
					end
				end
			end
		end
		
		for i, tile in pairs( neighbours ) do
			-- per tile chance
			if random() < 0.1 and tile.lavaResistance and tile.lavaResistance > 0 then
				tile.lavaResistance = tile.lavaResistance - 1
				if tile.lavaResistance == 0 then
					self:setTerrainTypeAtTile( tile.xi, tile.yi, 3 )
					print( "terrain: set lava at tile", tile.xi, tile.yi )
				end
			end
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function terrain:draw()
		
		for yi = 1, self.sizeY do
			for xi = 1, self.sizeX do
				local x = (xi-1) * tilesize
				local y = (yi-1) * tilesize
				
				local tile = self.tiles[yi][xi]
				local fow = Game.fow
				-- do not draw at all when not discovered yet
				-- only draw what is visible on screen
				if fow:isTileDiscovered( xi, yi )
				and Camera:isWorldPositionOnScreen( x, y ) then
					if self.selectedTile.xi == xi
					and self.selectedTile.yi == yi then
						-- a simple but decent looking "animation"; value goes from 0.6 to 1.0
						local v = 0.2 * math.sin( 3 * self.animationTick ) + 0.8
						lg.setColor( v, v, 1 )
					else
						if fow:isTileExplored( xi, yi ) then
							lg.setColor( 1, 1, 1 )
						else
							lg.setColor( 0.7, 0.7, 0.7 )
						end
					end
					tile.animation:draw()
					-- lg.setColor( 1, 1, 1 )
					
					-- x, y = convertToDimetric( x, y )
					-- lg.circle( "line", x, y, 2 )	-- debug
				end
			end
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function terrain:destroy()
		Gameloop:remove( self )
		Renderer:remove( self, "terrain" )
		self = nil
	end
	
	----------------------------------------------------------------------------------------------------
	
	terrain:init( mapData.terrain )
	
	Gameloop:add( terrain )
	Renderer:add( terrain, "terrain" )
	
	return terrain
end

return Terrain
