local Button = {}
local lg = love.graphics

-- see: https://love2d.org/forums/viewtopic.php?p=235000#p235000
--> for drawing the text use the same width as in text:setf

-- button state:
-- 0 = disabled, not visible (not used?)
-- 1 = enabled, normal
-- 2 = enabled, mouseover
-- 3 = enabled, mousedown
-- 4 = disabled (gray)

function Button:create( data )
	local button = {}
	button.id = data.id
	button.z = 2
	
	button.screenId = data.screenId	-- parent screen (must be!)
	button.parentId = data.parentId	-- parent widget (if any)
	button.canHover = true
	button.canClick = true
	
	button.callback = data.callback
	button.state = 1
	button.hover = false
	button.active = false
	button.hoverTick = 0
	
	if data.enabled == nil then
		button.enabled = true
	else
		button.enabled = data.enabled
	end
	if data.visible == nil then
		button.visible = true
	else
		button.visible = data.visible
	end
	
	button.bg_colors	= { [0] = {1, 1, 1, 0}, [1] = {0.28, 0.27, 0.70, 1}, [2] = {0.36, 0.35, 0.78, 1}, [3] = {0.32, 0.31, 0.74, 1}, [4] = {0.2, 0.2, 0.2, 1} }
	button.text_colors	= { [0] = {1, 1, 1, 0}, [1] = {1, 1, 1, 1}, [2] = {1, 1, 1, 1}, [3] = {0.15, 0.9, 0, 1}, [4] = {1, 1, 1, 1} }
	button.frame_colors	= { [0] = {1, 1, 1, 0}, [1] = {1, 1, 1, 1}, [2] = {0.15, 0.9, 0, 1}, [3] = {0.15, 0.9, 0, 1}, [4] = {1, 1, 1, 1} }
	
	-- for user interaction (mouse)
	button.body = {}
	button.body.shape = data.shape or "rect"
	button.body.sizeX = data.sizeX or 100
	button.body.sizeY = data.sizeY or 100
	button.body.x = data.x or 0 -- (button.body.sizeX/2)
	button.body.y = data.y or 0 -- (button.body.sizeY/2)
	
	-- optional: auto adjust the button size so that it is quadratic
	-- in that case, the sizeX and sizeY values are upper limits
	if data.autoSizeSquare then
		if button.body.sizeX < button.body.sizeY then
			button.body.sizeY = button.body.sizeX
		elseif button.body.sizeY < button.body.sizeX then
			button.body.sizeX = button.body.sizeY
		end
	end
	
	-- text...
	button.maxFontSize = data.fSize or 40
	button.maxTextWidth = 0.9 * button.body.sizeX
	button.maxTextHeight = 0.9 * button.body.sizeY
	
	----------------------------------------------------------------------------------------------------
	
	-- basically just returns the body (for mouse interaction)
	function button:getShape()
		return self.body
	end
	
	
	function button:setPosition( x, y )
		self.body.x = x or self.body.x
		self.body.y = y or self.body.y
	end
	
	----------------------------------------------------------------------------------------------------
	
	function button:onVisibilityChanged()
		if not self.visible then
			self.hoverTick = 0
			self.tooltipTick = 0
			self.hover = false
			self.active = false
		end
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- click begins (mouse is down; set active in terms of appearance)
	function button:onClickStart()
		self.active = true
		--print( "button start click", self.id )
	end
	
	-- on mouse (left?) click event
	function button:onClickEnd()
		if self.callback ~= nil
		and self.enabled then
			self.callback()
		end
		self.active = false
	end
	
	-- mouse button was released but not while hovering the button ... set inactive and be done
	function button:onClickCancel()
		self.active = false
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- manage mouseover
	function button:onMouseHoverStart()
		self.hover = true
	end
	
	function button:onMouseHoverEnd()
		self.hover = false
	end
	
	----------------------------------------------------------------------------------------------------
	
	function button:tick(dt)
		if not self.visible then
			return
		end
		
		if not self.enabled then
			self.state = 4
			return
		end
		
		-- from here: execute only if the button is visible and enabled
		
		self.drawTooltip = false
		
		if self.hover then

			self.hoverTick = self.hoverTick + dt
			
			if self.tooltipEnabled and self.hoverTick >= self.tooltipDelay then
				if self.tooltipDuration == -1 or self.tooltipTick < self.tooltipDuration then
					self.drawTooltip = true
					self.tooltipTick = self.tooltipTick + dt
				end
			end
			
		else
		
			self.hoverTick = 0
			self.tooltipTick = 0
			--self.hover = false
			--self.active = false

		end
		
		if self.active then
			self.state = 3
		else
			if self.hover then
				self.state = 2
			else
				self.state = 1
			end
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function button:draw()
		--print("draw btn")
		if not self.visible then
			return
		end
		
		if self.parentId ~= nil
		and not (GUImgr:isEnabled(self.parentId) and GUImgr:isVisible(self.parentId)) then
			return
		end
		
		--if self.state == 4 then
			--love.graphics.setColor(self.bg_colors[self.state])
		--end
		
		local body = self.body
		
		if self.imageData then
			-- print( "draw img")
			local img = self.imageData
			local scaleX = img.scaleX
			local scaleY = img.scaleY
			-- scale the image a bit if the mouse is held down while hovering the button
			if self.state == 3 then
				scaleX = scaleX * 0.9
				scaleY = scaleY * 0.9
			end
			-- lg.setScissor( body.x - 0.5 * body.sizeX, body.y - 0.5 * body.sizeY, body.sizeX, body.sizeY )
			lg.draw( img.image, body.x, body.y, 0, scaleX, scaleY, img.offsetX, img.offsetY )
			-- lg.setScissor()
		else
			love.graphics.setColor( self.bg_colors[self.state] )
			love.graphics.rectangle( "fill", body.x - 0.5 * body.sizeX, body.y - 0.5 * body.sizeY, body.sizeX, body.sizeY, 4, 4 )
		end
		
		love.graphics.setColor( self.frame_colors[self.state] )
		love.graphics.rectangle( "line", body.x - 0.5 * body.sizeX, body.y - 0.5 * body.sizeY, body.sizeX, body.sizeY, 4, 4 )
		
		love.graphics.setColor( self.text_colors[self.state] )
		
		-- use the defined text width, not the current width of the text object for the horizontal axis
		-- but use the current text height for the y coordinate
		lg.draw( self.text, body.x, body.y, 0, 1, 1, self.maxTextWidth * 0.5, self.text:getHeight() * 0.5 )
		
		
		if self.tooltip then
			self.tooltip.visible = self.drawTooltip
		end
		
		-- love.graphics.setColor(1, 1, 1)
	end
	
	----------------------------------------------------------------------------------------------------
	
	function button:setText( _text, alignMode )
		self.text = self.text or lg.newText( lg.newFont( self.maxFontSize ), "" )
		self.text:setf( _text or "", self.maxTextWidth, alignMode or self.alignMode or "center" )
		self:autoAdjustFontSize( self.text, self.maxTextWidth, self.maxTextHeight )
	end
	
	
	function button:autoAdjustFontSize( text, sizeX, sizeY )
		local fSize = self.maxFontSize or 40
		
		while text:getWidth() >= sizeX
		or text:getHeight() >= sizeY do
			fSize = fSize - 2
			if fSize < 8 then
				print( "button warning: UI element too small for text!", self.id )
				break
			end
			text:setFont( lg.newFont( fSize ) )
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function button:setImageData( imageData )
		
		if not imageData then
			self.imageData = nil
			return
		end
		
		local body = self.body
		
		self.imageData = {
			image = imageData.image,
			scaleX = body.sizeX / imageData.image:getWidth(),
			scaleY = body.sizeY / imageData.image:getHeight(),
		}
		self.imageData.offsetX = 0.5 * body.sizeX / self.imageData.scaleX
		self.imageData.offsetY = 0.5 * body.sizeY / self.imageData.scaleY
		
		-- print( "btn:setImageData")
		-- for k,v in pairs(self.imageData) do
			-- print( k,v)
		-- end
	end
	
	----------------------------------------------------------------------------------------------------
	
	function button:initTooltip( tt )
		self.tooltip = Tooltip:create{
			id = self.id .. "_tooltip",
			element = self,
			text = tt.text,
			maxFontSize = tt.maxFontSize,
			color = tt.tooltipColor,
			width = tt.tooltipWidth,
			height = tt.tooltipHeight,
			alignMode = tt.alignMode,
		}
		self.tooltipDelay = tt.delay or 1.0			-- show tt after X seconds
		self.tooltipDuration = tt.duration or 6	-- then show it for up to X seconds (-1 = show forever)
		self.tooltipTick = 0
	end
	
	function button:setTooltipText( text, alignMode )
		if not self.tooltip then
			self:initTooltip{ text = text }
		end
		self.tooltip:setText( text, alignMode )
	end
	
	----------------------------------------------------------------------------------------------------
	
	button:setText( data.text )
	button:setImageData( data.imageData )
	
	-- init tooltip (either exists from the start, or not)
	-- can be turned on and off with tooltipEnabled
	if data.tooltip then
		button.tooltipEnabled = data.tooltip.enabled ~= false
		button:initTooltip( data.tooltip )
	end
	
	Mainloop:add( button )
	GUImgr:addObject( button )
	UIRenderer:add( button )
	
	return button
end


return Button
