local lg = love.graphics
local Guimgr = {}

local sortFunction = function( a, b )
	return a.z < b.z
end


function Guimgr:create()
	local guimgr = {
		screenId = nil,
		screens = {},
		objects = {},
		objectsOnCurrentScreen = {},
		mouseHoveredElementId = nil,
		mouseClickElementId = nil,
		buttonClickCooldown = 0.1,	-- to prevent callbacks (including mouse hover) through different uiStates
	}
	
	-------------------------------------------------------------------------------------------------------------------
	
	function guimgr:addObject(obj)
		if not obj.id then
			print("guimgr error: invalid object ID")
			return false
		end
		
		table.insert(self.objects, obj)
		--print("guimgr: add object with ID", obj.id)
		
		local parent = self:getObjectById( obj.parentId )
		if parent then
			--print( "guimgr: set visibility of child", obj.id, "parent:", parent.id )
			self:setVisibility( obj.parentId, parent.visible )
		end
		return true
	end
	
	function guimgr:getObjectById(id)
		if not id then return nil end
		
		for k, element in pairs(self.objects) do
			if element.id == id then
				return element
			end
		end
		
		print( "guimgr warning: cannot find element", id )
		return nil
	end
	
	-------------------------------------------------------------------------------------------------------------------
	
	function guimgr:addScreen(obj)
		if not obj.id then
			print("guimgr error: invalid screen ID")
			return false
		end
		
		table.insert(self.screens, obj)
		--print("guimgr: add screen with ID", obj.id)
		return true
	end
	
	
	function guimgr:setScreen(screenId)
		for i = 1, #self.screens do
			screen = self.screens[i]
			if screen.id == screenId then
				self.screenId = screenId
				
				-- set background color
				if screen.bgColor then
					lg.setBackgroundColor(screen.bgColor)
				else
					lg.setBackgroundColor(0, 0, 0, 1)
				end
				
				-- store all UI objects in a table that belong to the currently active screen
				self.objectsOnCurrentScreen = {}
				for _, obj in pairs(self.objects) do
					if obj.screenId == screenId then
						table.insert(self.objectsOnCurrentScreen, obj)
						--print( "guimgr: set active:", obj.id )
					end
				end
				-- sort by z layer
				table.sort( self.objectsOnCurrentScreen, sortFunction )
				
				-- if there is an element, that has been hovered, stop...
				local previousHoverElement = self:getObjectById( self.mouseHoveredElementId )
				if previousHoverElement and previousHoverElement.onMouseHoverEnd then
					previousHoverElement:onMouseHoverEnd()
				end
				self.mouseHoveredElementId = nil
				
				return true
			end
		end
		return false
	end
	
	
	function guimgr:getCurrentScreenId()
		return self.screenId
	end
	
	-------------------------------------------------------------------------------------------------------------------
	
	-- deprecated, no?
	function guimgr:draw()
		local currentScreenId = self:getCurrentScreenId()
		for i, screen in pairs( self.screens ) do
			if screen.id == currentScreenId then
				if screen.draw then
					screen:draw()
				end
				return true
			end
		end
	end
	
	-------------------------------------------------------------------------------------------------------------------
	
	-- check if the mouse is currently hovering over a specific or any element that is visible
	function guimgr:isMouseHoveringGuiElement( elementId )
		
		if not self.mouseHoveredElementId then
			return false
		end
		
		if self.mouseHoveredElementId and elementId then
			if self.mouseHoveredElementId ~= elementId then
				return false
			end
		end
		
		return self:isVisible( self.mouseHoveredElementId )
		--return true
	end
	
	-------------------------------------------------------------------------------------------------------------------
	
	function guimgr:tick(dt)
		self.buttonClickCooldown = Tools.max(self.buttonClickCooldown - dt, 0)
		
		local mousePos = Mouse:getPosition()
		
		for _, element in pairs(self.objectsOnCurrentScreen) do
			if element.visible and element.canHover and element.getShape and element.id ~= self.mouseHoveredElementId then
				local point = { shape = "point", x = mousePos.x, y = mousePos.y }
				local shape = element:getShape()
				if Tools.isPointInsideShape(point, shape) then
					
					-- if there is an element, that has been hovered, stop...
					local previousElement = self:getObjectById(self.mouseHoveredElementId)
					
					if previousElement and previousElement.onMouseHoverEnd then
						previousElement:onMouseHoverEnd()
						-- print("guimgr: stop mouse hover for", self.mouseHoveredElementId)
					end
					
					-- store this information locally
					self.mouseHoveredElementId = element.id
					-- print("guimgr: begin mouse hover for", self.mouseHoveredElementId)
					
					--[[
					if previousElement and element.parentId then
						if previousElement.id == element.parentId then
							print( element.parentId )
							break
						end
					end
					--]]
					
					-- and inform the gui object that the mouse currently hovers over it...
					if element.onMouseHoverStart then
						element:onMouseHoverStart()
					end
					
					break	-- only allow one element at a time/position
				end
			end
		end
		
		if self.mouseHoveredElementId then
			local point = { shape = "point", x = mousePos.x, y = mousePos.y }
			local element = self:getObjectById(self.mouseHoveredElementId)
			local shape = element:getShape()
			if not Tools.isPointInsideShape(point, shape) then
				--print("guimgr: stop mouse hover for", self.mouseHoveredElementId)
				
				-- inform the gui object that the mouse does NOT longer hover over it...
				if element.onMouseHoverEnd then
					element:onMouseHoverEnd()
				end
				
				-- and remove the value locally
				self.mouseHoveredElementId = nil
			end
		end
		
	end
	
	-------------------------------------------------------------------------------------------------------------------
	
	-- invoked from mouse on click event
	function guimgr:onMouseClick( x, y, mouseButton )
		
		-- left click --> highlight (activate) the element
		-- note: on click callbacks are actually fired when the mouse button is released again
		local clickCatcher = false
		if mouseButton == 1 then
			for _, element in pairs( self.objectsOnCurrentScreen ) do
				if element.visible and element.canClick and element.getShape then
					local point = { shape = "point", x = x, y = y }
					local shape = element:getShape()
					if Tools.isPointInsideShape(point, shape) then
						
						-- store this information locally
						self.mouseClickElementId = element.id
						--print("guimgr: begin mouse click for", self.mouseClickElementId)
						
						if element.onClickStart then
							element:onClickStart()
						end
						
						-- start drag and drop if possible and if allowed right now (for this particular element)
						if element.onDragStart then
							if element.canStartDrag and element:canStartDrag() then
								self.mouseDragElementId = element.id
								element:onDragStart()
							else
								print( "guimgr: cannot start to drag on element", element.id )
							end
						end
						
						if element.onClickStart or element.onDragStart then
							return true	-- only allow one element at a time/position; this click event has been dealth with (no further callbacks)
						end
						
						-- click was caught by a GUI element, do not let it pass through to the game
						clickCatcher = true
					end
				end
			end
			
		end
		
		return clickCatcher == true
	end
	
	-- invoked from mouse on release event
	function guimgr:onMouseReleased(x, y, mouseButton)
		-- mouse click is done
		--> check if the cursor is still inside the element (if one at all) where the click event began
		if mouseButton == 1 then
			if self.mouseClickElementId then
				--print( "guimgr: mouseClickElementId:", self.mouseClickElementId )
				local point = { shape = "point", x = x, y = y }
				local element = self:getObjectById( self.mouseClickElementId )
				local shape = element:getShape()
				-- it might happen that the element is not visible anymore -> do not trigger any callback
				if element.visible and Tools.isPointInsideShape( point, shape ) then
					if element.onClickEnd then
						--print( "guimgr: on element clicked", element.id )
						element:onClickEnd()
						self:onButtonClicked()	-- set a short cooldown
					end
				else
					--> mouse release was not inside the element, but still deactivate it again
					if element.onClickCancel then
						element:onClickCancel()
					end
					
					-- drag and drop ended?
					local dragElement = self:getObjectById( self.mouseDragElementId )
					if dragElement then
						for _, e in pairs( self.objectsOnCurrentScreen ) do
							if e.visible and e.onDragEnd and e.getShape then
								if Tools.isPointInsideShape( point, e:getShape() ) then
									e:onDragEnd()
									self.mouseDragElementId = nil
									break
								end
							end
						end
					end
				end
				
				-- either way: forget the last active element
				self.mouseClickElementId = nil
			end
		end
		
	end
	
	-------------------------------------------------------------------------------------------------------------------
	
	-- short cooldown to prevent a ghost button click when the ui state changed
	function guimgr:onButtonClicked()
		self.buttonClickCooldown = 0.1
	end
	
	
	function guimgr:canClickButton()
		return self.buttonClickCooldown == 0
	end
	
	-------------------------------------------------------------------------------------------------------------------
	
	function guimgr:isEnabled(objectId)
		local obj = self:getObjectById(objectId)
		if obj then
			return obj.enabled
		end
		return false
	end
	
	
	function guimgr:setEnabledFlag(objectId, flag)
		local obj = self:getObjectById(objectId)
		if obj then
			obj.enabled = flag
			
			-- eg. widget
			local childs = self:getChildElements(obj)
			if childs then
				--> set enabled for child elements
				for i, child in pairs(childs) do
					child.enabled = flag
				end
			end
			
			return true
		end
		return false
	end
	
	-------------------------------------------------------------------------------------------------------------------
	
	function guimgr:isVisible( objectId )
		local obj = self:getObjectById( objectId )
		if obj then
			if not obj.parentId or self:isVisible( obj.parentId ) then
				return obj.visible == true
			end
		end
		return false
	end
	
	
	function guimgr:setVisibility( objectId, visibility )
		local obj = self:getObjectById( objectId )
		if obj then
			obj.visible = visibility
			--print( "guimgr: set visibility for", objectId, visibility )
			
			if obj.onVisibilityChanged then
				obj:onVisibilityChanged()
			end
			
			-- eg. widget
			local childs = self:getChildElements( obj )
			if childs then
				--> set visibility for child elements
				for i, child in pairs(childs) do
					child.visible = visibility
					if child.onVisibilityChanged then
						child:onVisibilityChanged()
					end
				end
			end
			
			return true
		end
		
		print( "guimgr invalid element (setVisibility)", objectId )
		return false
	end
	
	
	-- shortcut for setVisibility( objectId, not isVisible )
	function guimgr:toggleVisibility( objectId )
		return self:setVisibility( objectId, not self:isVisible( objectId ) )
	end
	
	
	-- this should be a widget or something that can have childs
	function guimgr:getChildElements(object)
		if not object.canHaveChilds then
			return false
		end
		
		local childs = {}
		for _, obj in pairs(self.objects) do
			if obj.parentId and obj.parentId == object.id then
				table.insert(childs, obj)
			end
		end
		
		return childs
	end
	
	-------------------------------------------------------------------------------------------------------------------
	
	function guimgr:enable(_id)
		local element = guimgr:getObjectById(_id)
		if not element then return false end
		element.enabled = true
	end
	
	
	function guimgr:disable(_id)
		local element = guimgr:getObjectById(_id)
		if not element then return false end
		element.enabled = false
	end
	
	-------------------------------------------------------------------------------------------------------------------
	
	function guimgr:setText(_id, _text)
		local element = guimgr:getObjectById(_id)
		if not element or not element.setText then return false end
		element:setText(_text)
		return true
	end
	
	
	Mainloop:add( guimgr )
	-- UIRenderer:add( guimgr )
	
	return guimgr
end


return Guimgr

