-- listselect; items can be clicked on; use mouse wheel to scroll through

local List = {}
local lg = love.graphics

function List:create( data )
	local list = {}
	list.id = data.id
	list.z = 2
	
	list.screenId = data.screenId	-- parent screen (must be!)
	list.parentId = data.parentId	-- parent widget (if any)
	list.canHover = true
	list.canClick = true
	
	list.body = {}
	list.body.shape = "rect"
	list.body.x = data.x
	list.body.y = data.y
	list.body.sizeX = data.sizeX
	list.body.sizeY = data.sizeY
	
	list.maxItems = data.maxItems or 4	-- to be displayed at a time
	assert( list.maxItems > 0 )
	
	if data.visible == nil then
		list.visible = true
	else
		list.visible = data.visible
	end
	
	list.enabled = true
	list.items = {}
	list.itemOffset = 0
	list.focusedItemIndex = nil	-- after clicking it
	
	list.onItemSelected = data.onItemSelected	-- callback when an item is clicked, can be nil
	
	-------------------------------------------------------------------------------------------------------------------------------------------------
	
	function list:addItem( item )
		--print( "list add item", self.id, item )
		table.insert( self.items, item )
	end
	
	
	function list:setItems( items )
		
		local oldOffset = self.itemOffset
		
		self:clearItems()
		
		for i, item in pairs( items ) do
			self:addItem( item )
		end
		
		-- make sure the item offset is not too large
		local nItems = #self.items
		if nItems > self.maxItems
		and oldOffset > nItems - self.maxItems then
			oldOffset = nItems - self.maxItems
		end
		self.itemOffset = oldOffset
		
	end
	
	
	function list:clearItems()
		self.items = {}
		self.itemOffset = 0
		self.focusedItemIndex = nil
	end
	
	
	function list:clearItemFocus()
		self.focusedItemIndex = nil
	end
	
	
	-- basically just returns the body (for mouse interaction)
	function list:getShape()
		return self.body
	end
	
	-------------------------------------------------------------------------------------------------------------------------------------------------
	
	function list:onClickEnd()
		
		local pos = Mouse:getPosition()
		local body = self.body
		if not Tools.isPointInsideRect( pos, body ) then
			return false
		end
		--print( "click at ", pos.x, pos.y )
		
		local itemDisplayHeight = body.sizeY / self.maxItems
		-- convert click pos into position relative to own body
		local y = pos.y - (body.y - 0.5*body.sizeY)
		-- index = count currently visible items
		local index = 1 + math.floor(y / itemDisplayHeight)
		-- index+itemOffset = actuel index used to access self.items
		local itemIndex = index + self.itemOffset
		--print( index, self.items[index+self.itemOffset] or "(empty)" )
		
		-- only call the on select function if there is actually an item
		if self.items[itemIndex] then
			--print( "set new focused item index: ", itemIndex )
			self.focusedItemIndex = itemIndex
			if self.onItemSelected then
				--print( "select item", self.items[itemIndex] )
				self.onItemSelected( self.items[itemIndex] )
			end
		else
			print( "listselect: no item at index", itemIndex )
		end
		
	end
	
	-------------------------------------------------------------------------------------------------------------------------------------------------
	
	-- returns the selected item (text)
	function list:getSelectedItem()
		if not self.items or not self.focusedItemIndex then
			return false
		end
		
		local nItems = #self.items
		if nItems < 1 or self.focusedItemIndex > nItems then
			return false
		end
		
		return self.items[self.focusedItemIndex]
	end
	
	-------------------------------------------------------------------------------------------------------------------------------------------------
	
	function list:onWheelMoved( dx, dy )
		
		-- check if the mouse was inside the list during the wheel move event
		local pos = Mouse:getPosition()
		local body = self.body
		if not Tools.isPointInsideRect( pos, body ) then
			return false
		end
		
		-- no need to scroll when not more than max items are currently in the list
		local nItems = #self.items
		if nItems <= self.maxItems then
			return
		end
		local maxOffset = nItems - self.maxItems
		
		if dy < 0 then
			-- moved down
			self.itemOffset = self.itemOffset + 1
			if self.itemOffset > maxOffset then
				self.itemOffset = maxOffset
			end
		elseif dy > 0 then
			-- moved up
			self.itemOffset = self.itemOffset - 1
			if self.itemOffset < 0 then
				self.itemOffset = 0
			end
		end
		
	end
	
	-------------------------------------------------------------------------------------------------------------------------------------------------
	
	function list:draw()
		
		if not self.visible then
			return
		end
		
		if self.parentId and not (GUImgr:isEnabled(self.parentId) and GUImgr:isVisible(self.parentId)) then
			return
		end
		
		local body = self.body
		local top = body.y - 0.5*body.sizeY
		local left = body.x - 0.5*body.sizeX
		lg.rectangle( "line", left, top, body.sizeX, body.sizeY, 2, 2 )
		
		local itemDisplayHeight = body.sizeY / self.maxItems
		for i = 1, self.maxItems-1 do
			lg.line( left, top + i*itemDisplayHeight, left + body.sizeX, top + i*itemDisplayHeight )
		end
		
		local border = 2
		for i = 1, self.maxItems do
			local itemIndex = i+self.itemOffset
			local item = self.items[itemIndex]
			if item then
				-- highlight currently selected item if any and if currently displayed
				if self.focusedItemIndex and self.focusedItemIndex == itemIndex then
					lg.setColor( 1, 1, 1, 0.1 )
					lg.rectangle( "fill", left, top + (i-1)*itemDisplayHeight, body.sizeX, itemDisplayHeight, 2, 2 )
					lg.setColor( 1, 1, 0, 0.5 )
					lg.rectangle( "line", left, top + (i-1)*itemDisplayHeight, body.sizeX, itemDisplayHeight, 2, 2 )
					lg.setColor( 1, 1, 1, 1 )
				end
				
				lg.printf( item, left+border, top + (i-1)*itemDisplayHeight, body.sizeX-2*border )
			end
		end
		
	end
	
	-------------------------------------------------------------------------------------------------------------------------------------------------
	
	-- initial items, can be empty or nil
	if data.items then
		--for i, item in pairs(data.items) do
		--	list:addItem( item )
		--end
		list:setItems( data.items )
	end
	
	
	Mouse:addListener( "wheelMoved", list )	-- make sure to have onWheelMoved function
	GUImgr:addObject( list )
	UIRenderer:add( list )
	
	return list
end

return List
