local MiniMap = {}
local lg = love.graphics

-- simple, flat depiction of the game world as a mini map

-- size of each tile in pixel
-- this has nothing to do with the global tilesize as defined for the game world!
local tileScale = 0.2
local tileSize = tileScale * TILESIZE


function MiniMap:create( data )
	local self = {}
	self.id = data.id
	self.z = 2
	self.screenId = data.screenId
	self.canHover = true
	-- self.canClick = true	-- TODO! GUImgr?
	
	self.body = {}
	self.body.shape = data.shape or "rect"
	self.body.sizeX = data.sizeX or 100
	self.body.sizeY = data.sizeY or 100
	self.body.x = data.x or 0
	self.body.y = data.y or 0
	self.body.z = data.z or 1
	
	if data.enabled == nil then
		self.enabled = true
	else
		self.enabled = data.enabled
	end
	if data.visible == nil then
		self.visible = true
	else
		self.visible = data.visible
	end
	
	-- map data
	-- self.terrainData = {}
	-- self.entityData = {}
	
	self.cameraScale = 1
	self.cameraOffsetX = 0
	self.cameraOffsetY = 0
	
	----------------------------------------------------------------------------------------------------
	
	function self:getShape()
		return self.body
	end
	
	----------------------------------------------------------------------------------------------------
	
	function self:tick( dt )
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function self:setMapSize( x, y )
		
		self.mapSizeX = x
		self.mapSizeY = y
		
		self.canvasSizeX = x*tileSize
		self.canvasSizeY = y*tileSize
		self.canvas = lg.newCanvas( self.canvasSizeX, self.canvasSizeY )
		-- print( "minimap: set map size", self.canvasSizeX, self.canvasSizeY )
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- called from game
	function self:updateMapData( terrainData, walls, entities )
		
		lg.setCanvas( self.canvas )
		lg.clear( 0, 0, 0 )
		
		lg.setColor( COLORS.white )
		
		-- terrain
		for yi = 1, self.mapSizeY do
			for xi = 1, self.mapSizeX do
				local tileType = terrainData[yi][xi]
				if tileType == 1 then			-- normal ground
					lg.setColor( COLORS.darkgray )
				elseif tileType == 2 then	-- water
					lg.setColor( COLORS.blue )
				elseif tileType == 3 then	-- lava
					lg.setColor( COLORS.orange )
				end
				if tileType then
					lg.rectangle( "fill", (xi-1)*tileSize, (yi-1)*tileSize, tileSize, tileSize )
				end
			end
		end
		
		-- grid
		lg.setColor( 0, 0, 0, 0.2 )
		for yi = 1, self.mapSizeY do
			for xi = 1, self.mapSizeX do
				lg.rectangle( "line", (xi-1)*tileSize, (yi-1)*tileSize, tileSize, tileSize )
			end
		end
		
		-- walls
		for _, wall in pairs( walls ) do
			local wallType = wall.entityType
			if wallType == ENTITIES.solidRock then
				lg.setColor( COLORS.lightgray )
			elseif wallType == ENTITIES.hardRock then
				lg.setColor( COLORS.gray )
			elseif wallType == ENTITIES.looseRock then
				lg.setColor( COLORS.brown )
			end
			lg.rectangle( "fill", (wall.xi-1)*tileSize+1, (wall.yi-1)*tileSize+1, tileSize-2, tileSize-2 )
			if wall.isReinforced then
				lg.setColor( COLORS.yellow )
				lg.rectangle( "line", (wall.xi-1)*tileSize+1, (wall.yi-1)*tileSize+1, tileSize-2, tileSize-2 )
				lg.setColor( COLORS.gray )
			end
		end
		
		-- transform object for all entities (and the camera)
		local transform = love.math.newTransform()
		transform:translate( 0.5*tileSize, 0.5*tileSize )
		transform:scale( tileScale )
		
		-- entities (buildings, mobs, vehicles, Rock Raiders etc.)
		for _, entity in pairs( entities ) do
			local eType = entity.entityType
			local eCat = entity.entityCategory
			local ex, ey = entity:getPosition()
			
			if eCat == ENTITYCATEGORY.building then
				-- buildings represented as smaller rectangles
				lg.setColor( COLORS.yellow )
				local xi, yi = entity.xi, entity.yi
				lg.rectangle( "fill", (xi-1)*tileSize+4, (yi-1)*tileSize+4, tileSize-8, tileSize-8 )
				
			elseif eCat == ENTITYCATEGORY.vehicle
			or eCat == ENTITYCATEGORY.robot then
				-- yellow circles
				lg.setColor( COLORS.yellow )
				local x, y = transform:transformPoint( ex, ey )
				lg.circle( "fill", x, y, 3 )
				
			elseif eCat == ENTITYCATEGORY.monster then
				-- red circles
				lg.setColor( COLORS.red )
				local x, y = transform:transformPoint( ex, ey )
				lg.circle( "fill", x, y, 3 )
				
			elseif eType == ENTITIES.rockRaider then
				-- yellow circles
				lg.setColor( COLORS.yellow )
				local x, y = transform:transformPoint( ex, ey )
				lg.circle( "fill", x, y, 3 )
				
			elseif eType == ENTITIES.energyCrystal then
				-- lightgreen circles
				lg.setColor( COLORS.lightgreen )
				local x, y = transform:transformPoint( ex, ey )
				lg.circle( "fill", x, y, 2 )
				
			elseif eType == ENTITIES.ore then
				-- light brown circles
				lg.setColor( COLORS.lightBrown )
				local x, y = transform:transformPoint( ex, ey )
				lg.circle( "fill", x, y, 2 )
				
			end
		end
		
		
		-- fow: draw discovered but currently not explored tiles a bit darker
		local fow = Game.fow
		lg.setColor( 0, 0, 0, 0.25 )
		for yi = 1, self.mapSizeY do
			for xi = 1, self.mapSizeX do
				if fow:isTileDiscovered( xi, yi ) and not fow:isTileExplored( xi, yi ) then
					lg.rectangle( "fill", (xi-1)*tileSize, (yi-1)*tileSize, tileSize, tileSize )
				end
			end
		end
		
		
		-- camera position
		-- draw a rectangle by using 4 lines between the 4 corner points
		-- (transform each screen edge into a world position)
		lg.setColor( COLORS.white )
		local width = SCREEN.width
		local height = SCREEN.height
		local x1, y1 =  transform:transformPoint( convertTo2D( Camera:getTransform():inverseTransformPoint( 0, 0 ) ) )
		local x2, y2 =  transform:transformPoint( convertTo2D( Camera:getTransform():inverseTransformPoint( width, 0 ) ) )
		local x3, y3 =  transform:transformPoint( convertTo2D( Camera:getTransform():inverseTransformPoint( width, height ) ) )
		local x4, y4 =  transform:transformPoint( convertTo2D( Camera:getTransform():inverseTransformPoint( 0, height ) ) )
		lg.polygon( "line", x1, y1,  x2, y2,  x3, y3,  x4, y4 )
		
		-- debug: center of the screen (camera)
		local cx, cy = transform:transformPoint( convertTo2D( Camera:getTransform():inverseTransformPoint( 0.5*width, 0.5*height ) ) )
		lg.circle( "line", cx, cy, 4 )
		
		lg.setCanvas()
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function self:draw()
		if not self.visible then
			return
		end
		
		if GUImgr:getCurrentScreenId() ~= self.screenId then
			return
		end
		
		local body = self.body
		lg.setColor( 0.1, 0.1, 0.1 )
		lg.rectangle( "fill", body.x - body.sizeX * 0.5, body.y - body.sizeY * 0.5, body.sizeX, body.sizeY )
		lg.setColor( 1, 1, 1, 1 )
		lg.rectangle( "line", body.x - body.sizeX * 0.5, body.y - body.sizeY * 0.5, body.sizeX, body.sizeY )
		
		if not self.canvas then
			return
		end
		
		lg.push()
			local scaleX = body.sizeX / self.canvasSizeX
			local scaleY = body.sizeY / self.canvasSizeY
			local scale = math.min( scaleX, scaleY ) * self.cameraScale
			
			local a = 0.5*self.canvasSizeX
			local b = 0.5*self.canvasSizeY
			local dy = math.sqrt(a*a + b*b) * scale
			lg.translate( self.cameraOffsetX, self.cameraOffsetY - dy )
			local border = 0	-- border size in pixel
			
			-- limit the drawing area to a rectangle
			lg.setScissor( body.x - body.sizeX * 0.5 + border, body.y - body.sizeY * 0.5 + border, body.sizeX - 2*border, body.sizeY - 2*border )
			lg.draw( self.canvas, body.x, body.y, 0.25*math.pi, scale, scale )
			lg.setScissor()
		lg.pop()
		-- extra border, not needed?
		-- lg.rectangle( "line", body.x - body.sizeX * 0.5 + border, body.y - body.sizeY * 0.5 + border, body.sizeX - 2*border, body.sizeY - 2*border )
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function self:onMousePressed( x, y, button )
		
		if not self.visible or GUImgr:getCurrentScreenId() ~= self.screenId
		or not Tools.isPointInsideShape( Mouse:getPosition(), self:getShape() ) then
			return
		end
		
		-- right click to jump to the position
		if button == 2 then
			
			local x = x - self.body.x
			local y = y - self.body.y
			local scaleX = self.body.sizeX / self.canvasSizeX
			local scaleY = self.body.sizeY / self.canvasSizeY
			local scale = math.min( scaleX, scaleY ) * self.cameraScale
			local a = 0.5*self.canvasSizeX
			local b = 0.5*self.canvasSizeY
			local dy = math.sqrt(a*a + b*b) * scale
			
			-- TODO: there is still a little offset
			local transform = love.math.newTransform()
			transform:translate( 0.5*tileSize, 0.5*tileSize )
			-- transform:scale( tileScale )
			-- transform:translate( 0.5*tileSize + self.cameraOffsetX, 0.5*tileSize + self.cameraOffsetY - dy )
			transform:translate( self.cameraOffsetX, self.cameraOffsetY - dy )
			transform:scale( tileScale * scale )	-- it seems all operations of the same type must be combined
			transform:rotate( 0.25*math.pi )
			local wx, wy = transform:inverseTransformPoint( x, y )
			
			-- print( wx, wy )
			Game:minimapJumpTo( wx, wy )
			
		-- middle mouse click to reset the camera offset
		elseif button == 3 then
			self:resetCameraPosition()
		end
		
	end
	
	
	function self:onMouseMoved( dx, dy )
		
		if not self.visible or GUImgr:getCurrentScreenId() ~= self.screenId
		or not Tools.isPointInsideShape( Mouse:getPosition(), self:getShape() ) then
			return
		end
		
		-- left mouse button to move the map
		if love.mouse.isDown( 1 ) then
			self.cameraOffsetX = self.cameraOffsetX + dx
			self.cameraOffsetY = self.cameraOffsetY + dy
		end
		
	end
	
	
	function self:onWheelMoved( dx, dy )
		
		if not self.visible or GUImgr:getCurrentScreenId() ~= self.screenId
		or not Tools.isPointInsideShape( Mouse:getPosition(), self:getShape() ) then
			return
		end
		
		-- camera zoom
		if dy < 0 then
			-- moved down
			self.cameraScale = self.cameraScale - 0.1
		elseif dy > 0 then
			-- moved up
			self.cameraScale = self.cameraScale + 0.1
		end
		self.cameraScale = Tools.clamp( self.cameraScale, 0.8, 2 )
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function self:resetCameraPosition()
		self.cameraOffsetX = 0
		self.cameraOffsetY = 0
	end
	
	----------------------------------------------------------------------------------------------------
	
	GUImgr:addObject( self )
	UIRenderer:add( self )
	
	Mouse:addListener( "mousePressed", self )	-- onMousePressed
	Mouse:addListener( "mouseMoved", self )	-- onMouseMoved
	Mouse:addListener( "wheelMoved", self )	-- onWheelMoved
	
	return self
end


return MiniMap
