local NotificationView = {}
local lg = love.graphics

-- display short messages
-- one at a time, at least for X seconds
-- TODO: optional sound when a new message is displayed (not when it is added...)

function NotificationView:create( data )
	local self = {}
	
	self.id = data.id
	self.z = 2
	self.screenId = data.screenId
	self.canHover = true	-- act as click cather = do not let mouse clicks through to interact with game objects
	self.canClick = true
	
	self.body = {}
	self.body.shape = data.shape or "rect"
	self.body.sizeX = data.sizeX or 100
	self.body.sizeY = data.sizeY or 100
	self.body.x = data.x or 0
	self.body.y = data.y or 0
	self.body.z = data.z or 1
	
	if data.enabled == nil then
		self.enabled = true
	else
		self.enabled = data.enabled
	end
	if data.visible == nil then
		self.visible = true
	else
		self.visible = data.visible
	end
	
	self.bgColor = data.bgColor
	
	-- do not draw at all, when there are no messages to display
	self.hideWhenEmpty = data.hideWhenEmpty
	
	self.maxFontSize = data.fSize or 20
	self.maxTextWidth = 0.9 * self.body.sizeX
	self.maxTextHeight = 0.9 * self.body.sizeY
	
	-- store messages
	self.messageBuffer = {}
	
	-- each message will be displayed for at least X seconds
	self.minDisplayTime = data.minDisplayTime or 2
	
	-- maximum duration a message is displayed if no newer message arrives
	self.maxDisplayTime = data.maxDisplayTime or 10
	
	----------------------------------------------------------------------------------------------------
	
	function self:getShape()
		return self.body
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- note: default to left alignment
	function self:setText( _text, alignMode )
		-- create a new text object to reset the font size
		self.text = lg.newText( lg.newFont( self.maxFontSize ), "" )
		-- self.text = self.text or lg.newText( lg.newFont( self.maxFontSize ), "" )
		self.text:setf( _text or "", self.maxTextWidth, alignMode or self.alignMode or "left" )
		self:autoAdjustFontSize( self.text, self.maxTextWidth, self.maxTextHeight )
	end
	
	
	function self:autoAdjustFontSize( text, sizeX, sizeY )
		local fSize = self.maxFontSize
		
		while text:getWidth() >= sizeX
		or text:getHeight() >= sizeY do
			fSize = fSize - 2
			if fSize < 8 then
				print( "warning: UI element too small for text!", self.id )
				break
			end
			text:setFont( lg.newFont( fSize ) )
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- add a new message (string) to be displayed
	-- optional: overwrite default maximum duration for this message only
	function self:newMessage( text, duration )
		
		local entry = {
			text = text,
			displayTime = 0,
			maxDisplayTime = duration or self.maxDisplayTime,
		}
		table.insert( self.messageBuffer, entry )
		
		self:updateCurrentText()
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function self:updateCurrentText()
		local text = self:getCurrentText()
		if text then
			self:setText( text )
			-- print( self.id, "update text:", text )
		else
			self:setText( "" )
		end
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- return the most current message (if any)
	function self:getCurrentText()
		local msg = self.messageBuffer[1]
		if msg then
			return msg.text
		else
			return nil
		end
	end
	
	----------------------------------------------------------------------------------------------------
	
	function self:tick( dt )
		
		-- currently displaying a message?
		local msg = self.messageBuffer[1]
		if msg then
			msg.displayTime = msg.displayTime + dt
			if msg.displayTime >= msg.maxDisplayTime then
				table.remove( self.messageBuffer, 1 )
				self:updateCurrentText()
			end
			
			-- more messages to be displayed?
			-- only show the current message for the minumum duration
			if self.messageBuffer[2] and msg.displayTime >= self.minDisplayTime then
				table.remove( self.messageBuffer, 1 )
				self:updateCurrentText()
			end
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function self:draw()
		
		if not self.visible then
			return
		end
		
		if self.parentId ~= nil
		and not (GUImgr:isEnabled(self.parentId) and GUImgr:isVisible(self.parentId)) then
			return
		end
		
		-- note that this still catches mouse clicks! (see GUImgr)
		if self.hideWhenEmpty and self:getCurrentText() == nil then
			return
		end
		
		local body = self.body
		
		if self.bgColor then
			lg.setColor( self.bgColor )
			love.graphics.rectangle( "fill", body.x - body.sizeX/2, body.y - body.sizeY/2, body.sizeX, body.sizeY, 4, 4 )
		end
		lg.setColor( COLORS.white )
		love.graphics.rectangle( "line", body.x - body.sizeX/2, body.y - body.sizeY/2, body.sizeX, body.sizeY, 4, 4 )
		
		lg.draw( self.text, body.x, body.y, 0, 1, 1, self.maxTextWidth * 0.5, self.text:getHeight() * 0.5 )
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	GUImgr:addObject( self )
	Mainloop:add( self )
	UIRenderer:add( self )
	
	self:updateCurrentText()
	
	return self
end


return NotificationView
