local Slider = {}
local lg = love.graphics

local margin = 4

-- TODO: right click to reset to default value


function Slider:create( data )
	local slider = {}
	slider.id = data.id
	slider.z = 2
	
	slider.screenId = data.screenId	-- parent screen (must be!)
	slider.parentId = data.parentId	-- parent widget (if any)
	slider.canHover = true
	slider.canClick = true
	
	slider.body = {}
	slider.body.shape = "rect"
	slider.body.x = data.x
	slider.body.y = data.y
	slider.body.sizeX = data.sizeX or SCREEN.width * 0.2
	slider.body.sizeY = data.sizeY or SCREEN.height * 0.2
	
	slider.roundCorners = data.roundCorners
	
	slider.enabled = data.enabled ~= false
	slider.visible = data.visible ~= false
	
	slider.onValueChanged = data.onValueChanged or function(v) end
	
	slider.horizontal = data.horizontal
	slider.minValue = data.minValue or 0
	slider.maxValue = data.maxValue or 100
	-- slider.value = data.value or slider.minValue	-- current value, see setValue
	if slider.minValue >= slider.maxValue then
		print( "slider ERROR: invalid min/max values!", slider.id )
		return false
	end
	
	if slider.horizontal then
		slider.minX = slider.body.x - 0.5 * slider.body.sizeX + margin
		slider.maxX = slider.body.x + 0.5 * slider.body.sizeX - margin
		slider.knobSize = data.knobSize or 4 --0.1 * slider.body.sizeY
		slider.sliderPosition = slider.body.x - margin	-- initial position only
	else
		slider.minY = slider.body.y - 0.5 * slider.body.sizeY + margin
		slider.maxY = slider.body.y + 0.5 * slider.body.sizeY - margin
		slider.knobSize = data.knobSize or 4 --0.1 * slider.body.sizeX
		slider.sliderPosition = slider.body.y + margin
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- basically just returns the body (for mouse interaction)
	function slider:getShape()
		return self.body
	end
	
	
	function slider:setPosition( x, y )
		self.body.x = x or self.body.x
		self.body.y = y or self.body.y
	end
	
	----------------------------------------------------------------------------------------------------
	
	function slider:onVisibilityChanged()
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function slider:onClickStart()
		self.active = true
	end
	
	-- on mouse (left?) click event
	function slider:onClickEnd()
		-- if self.callback ~= nil
		-- and self.enabled then
			-- self.callback()
		-- end
		self.active = false
	end
	
	-- mouse button was released but not while hovering the slider
	function slider:onClickCancel()
		self.active = false
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- value is either x (horizontal) or y (vertical)
	function slider:setSliderPosition( v )
		self.sliderPosition = v
	end
	
	
	-- the slider has been moved by the user
	-- now update the corresponding value
	function slider:updateValue()
		local oldValue = self.value
		local valueRange = self.maxValue - self.minValue
		local sliderRange
		local sliderValue
		if self.horizontal then
			sliderRange = self.maxX - self.minX
			sliderValue = self.sliderPosition - self.minX
		else
			sliderRange = self.maxY - self.minY
			sliderValue = self.maxY - self.sliderPosition	-- note: bottom = lowest value!
		end
		-- sliderValue / sliderRange = result / valueRange
		self.value = self.minValue + valueRange * sliderValue / sliderRange
		
		-- only call an update if the value actually changed! might not hold true when rounding values...
		if self.value ~= oldValue then
			-- print( self.value )
			self.onValueChanged( self.value )
		end
	end
	
	
	-- called when creating the slider or when setting the value by script
	-- update the slider position accordingly
	function slider:setValue( value )
		local oldValue = self.value
		self.value = value
		
		self:updateSliderPosition()
		
		if self.value ~= oldValue then
			self.onValueChanged( self.value )
		end
	end
	
	
	function slider:updateSliderPosition()
		self.sliderPosition = self:valueToSliderPosition( self.value )
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- catch mouse click events
	function slider:onMousePressed( x, y, button )
		if not self.visible or not self.enabled then
			return
		end
		
		-- only care about right click
		if button ~= 2 then
			return
		end
		
		-- has a default value at all?
		if not self.defaultValue then return end
		
		-- check if the mouse was inside the element during the click event
		local pos = Mouse:getPosition()
		local body = self.body
		if not Tools.isPointInsideRect( pos, body ) then
			return
		end
		
		-- reset value to default
		self:setValue( self.defaultValue )
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function slider:tick( dt )
		if not self.visible or not self.active or not self.enabled then
			return
		end
		
		if love.mouse.isDown( 1 ) then
			local mouse_x, mouse_y = love.mouse.getPosition()
			local v = 0
			if self.horizontal then
				-- local x = mouse_x - self.minX
				v = Tools.clamp( mouse_x, self.minX, self.maxX )
			else
				-- local y = mouse_y - self.minY
				v = Tools.clamp( mouse_y, self.minY, self.maxY )
			end
			-- print( v )
			
			-- discrete values only
			if self.sliderPositions then
				local bestD = math.huge
				local closestV = v
				for i = 1, #self.sliderPositions do
					local d = math.abs( v - self.sliderPositions[i] )
					if d < bestD then
						bestD = d
						closestV = self.sliderPositions[i]
					end
				end
				v = closestV
			end
			
			self:setSliderPosition( v )
			self:updateValue()
		end
	end
	
	----------------------------------------------------------------------------------------------------
	
	function slider:draw()
		if not self.visible then
			return
		end
		
		if GUImgr:getCurrentScreenId() ~= self.screenId then
			return
		end
		
		if self.parentId ~= nil
		and not (GUImgr:isEnabled(self.parentId) and GUImgr:isVisible(self.parentId)) then
			return
		end
		
		local body = self.body
		
		if not self.enabled then
			lg.setColor( COLORS.darkgray )
		end
		if self.roundCorners then
			lg.rectangle( "line", body.x - body.sizeX * 0.5, body.y - body.sizeY * 0.5, body.sizeX, body.sizeY, 4, 4 )
		else
			lg.rectangle( "line", body.x - body.sizeX * 0.5, body.y - body.sizeY * 0.5, body.sizeX, body.sizeY )
		end
		lg.setColor( COLORS.white )
		
		-- small circles or small lines?
		if self.sliderPositions then
			lg.setColor( 1, 1, 1, 0.5 )
			local lineSize = 4
			if self.horizontal then
				for i = 1, #self.sliderPositions do
					local x = self.sliderPositions[i]
					-- lg.circle( "fill", x, body.y, 2 )
					lg.line( x, body.y - lineSize, x, body.y + lineSize )
				end
			else
				for i = 1, #self.sliderPositions do
					local y = self.sliderPositions[i]
					-- lg.circle( "fill", body.x, y, 2 )
					lg.line( body.x - lineSize, y, body.x + lineSize, y )
				end
			end
			lg.setColor( COLORS.white )
		end
		
		-- the actual "slider"
		if not self.enabled then
			lg.setColor( COLORS.darkgray )
		end
		if self.horizontal then
			lg.line( body.x - 0.5 * body.sizeX + margin, body.y, body.x + 0.5 * body.sizeX - margin, body.y )
			lg.circle( "fill", self.sliderPosition, body.y, self.knobSize )
		else	--> vertical
			-- note: bottom = lowest value
			lg.line( body.x, body.y - 0.5 * body.sizeY + margin, body.x, body.y + 0.5 * body.sizeY - margin )
			lg.circle( "fill", body.x, self.sliderPosition, self.knobSize )
		end
		lg.setColor( COLORS.white )
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function slider:valueToSliderPosition( value )
		local valueRange = self.maxValue - self.minValue
		local sliderRange
		if self.horizontal then
			sliderRange = self.maxX - self.minX
			return self.minX + sliderRange * (value-self.minValue) / valueRange
		else
			sliderRange = self.maxY - self.minY
			return self.minY + sliderRange * value / valueRange
		end
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- continuous or discrete values allowed?
	local step = data.step
	if step then
		local valueRange = slider.maxValue - slider.minValue
		local nSegments = valueRange / step + 1
		if nSegments ~= Tools.round( nSegments ) then
			print( "slider ERROR: value range must be a multiple of step size", slider.id )
			return false
		end
		
		slider.sliderPositions = {}
		for i = 1, nSegments do
			slider.sliderPositions[i] = slider:valueToSliderPosition( slider.minValue + (i-1) * step )
			-- print( "slider pos", i, slider.sliderPositions[i] )
		end
	end
	
	if data.value then
		slider.defaultValue = data.value
	end
	slider:setValue( data.value or slider.minValue )
	
	Mouse:addListener( "mousePressed", slider )
	Mainloop:add( slider )
	GUImgr:addObject( slider )
	UIRenderer:add( slider )
	
	return slider
end


return Slider
