local TextScroll = {}
local lg = love.graphics

-- like a label but suited for more text + vertical scroll (if needed)

-- space around the text (top and bottom)
local textBorder = 16

function TextScroll:create( data )
	local self = {}
	
	self.id = data.id
	self.z = 2
	self.screenId = data.screenId
	self.parentId = data.parentId
	self.canHover = true
	self.canClick = true
	
	self.body = {}
	self.body.shape = "rect"
	self.body.x = data.x
	self.body.y = data.y
	self.body.sizeX = data.sizeX
	self.body.sizeY = data.sizeY
	
	-- text...
	self.maxFontSize = data.fSize or 40
	self.maxTextWidth = self.body.sizeX - 2 * textBorder	--* 0.9
	-- self.maxTextHeight = self.body.sizeY * 0.9
	self.alignMode = data.alignMode or "left"
	self.offsetY = -textBorder
	
	self.drawFrame = true
	if data.drawFrame == false then
		self.drawFrame = false
	end
	
	----------------------------------------------------------------------------------------------------
	
	function self:getShape()
		return self.body
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- manage mouseover
	function self:onMouseHoverStart()
		self.hover = true
	end
	
	function self:onMouseHoverEnd()
		self.hover = false
	end
	
	----------------------------------------------------------------------------------------------------
	
	function self:onWheelMoved( dx, dy )
		
		-- check if the mouse was inside the element during the wheel move event
		local pos = Mouse:getPosition()
		local body = self.body
		if not Tools.isPointInsideRect( pos, body ) then
			return false
		end
		
		local scrollSpeed = 10
		if dy < 0 then
			-- moved down
			self.offsetY = self.offsetY + scrollSpeed
		elseif dy > 0 then
			-- moved up
			self.offsetY = self.offsetY - scrollSpeed
		end
		
		-- maximum offset can be calculated from the text size and UI element height
		local maxOffset = self.text:getHeight() - body.sizeY
		self.offsetY = Tools.clamp( self.offsetY, -textBorder, maxOffset + textBorder )
		
		self:onOffsetChanged()
		
	end
	
	
	function self:onOffsetChanged()
		
		-- maximum offset can be calculated from the text size and UI element height
		local maxOffset = self.text:getHeight() - self.body.sizeY
		
		self.drawScrollHintTop = self.offsetY > -textBorder
		self.drawScrollHintBottom = self.offsetY < maxOffset + textBorder
		
	end
	
	
	function self:resetOffset()
		self.offsetY = -textBorder
		self:onOffsetChanged()
	end
	
	----------------------------------------------------------------------------------------------------
	
	function self:draw()
		
		if not self.visible then
			return
		end
		
		if self.parentId and not (GUImgr:isEnabled(self.parentId) and GUImgr:isVisible(self.parentId)) then
			return
		end
		
		local body = self.body
		local left = body.x - 0.5 * body.sizeX
		local top = body.y - 0.5 * body.sizeY
		-- if self.hover then
			-- lg.setColor( COLORS.darkgray )
		-- end
		if self.drawFrame then
			lg.setColor( COLORS.black )
			lg.rectangle( "fill", left, top, body.sizeX, body.sizeY, 2, 2 )
			lg.setColor( COLORS.white )
			lg.rectangle( "line", left, top, body.sizeX, body.sizeY, 2, 2 )
		end
		
		-- limit the drawing area to a rectangle
		-- lg.setScissor( left, top, body.sizeX, body.sizeY )
		lg.setScissor( left + textBorder, top + textBorder, body.sizeX - 2 * textBorder, body.sizeY - 2 * textBorder )
		-- use the defined text width, not the current width of the text object for the horizontal axis
		-- but use the current text height for the y coordinate
		-- plus for the y position: start from the top + add half of the (true) text height
		-- the offsetY already includes a border around the text
		lg.draw( self.text, body.x, top + self.text:getHeight() * 0.5 - self.offsetY, 0, 1, 1, self.maxTextWidth * 0.5, self.text:getHeight() * 0.5 )
		lg.setScissor()
		
		-- give a visual hint that the user can scroll up and down
		do
			local x = body.x
			local y = body.y
			local top = top + 4
			local bottom = y + 0.5 * body.sizeY - 4
			local size = 8
			if self.drawScrollHintTop then
				lg.polygon( "fill", x-size, top+size, x+size, top+size, x, top )
			end
			if self.drawScrollHintBottom then
				lg.polygon( "fill", x-size, bottom-size, x+size, bottom-size, x, bottom )
			end
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function self:setText( _text, alignMode )
		self.text = self.text or lg.newText( lg.newFont( self.maxFontSize ), "" )
		self.text:setf( _text or "", self.maxTextWidth, alignMode or self.alignMode or "left" )
		self:autoAdjustFontSize( self.text, self.maxTextWidth )
		
		self:onOffsetChanged()
	end
	
	-- no height limitation here!
	function self:autoAdjustFontSize( text, sizeX )
		local fSize = self.maxFontSize or 40
		
		while text:getWidth() >= sizeX do
		-- or text:getHeight() >= sizeY do
			fSize = fSize - 2
			if fSize < 8 then
				print( "textscroll warning: UI element too small for text!", self.id )
				break
			end
			text:setFont( lg.newFont( fSize ) )
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	self:setText( data.text )
	
	Mouse:addListener( "wheelMoved", self )	-- make sure to have onWheelMoved function
	GUImgr:addObject( self )
	UIRenderer:add( self )
	
	return self
end

return TextScroll
