-- simple input field to select a (number) value by using the mouse wheel
-- scroll up and down to increase and decrease the value

local ValuePicker = {}
local lg = love.graphics

function ValuePicker:create( data )
	local vp = {}
	
	vp.id = data.id
	vp.z = 2
	vp.screenId = data.screenId
	vp.parentId = data.parentId
	vp.canHover = true
	vp.canClick = true
	
	vp.body = {}
	vp.body.shape = "rect"
	vp.body.x = data.x
	vp.body.y = data.y
	vp.body.sizeX = data.sizeX
	vp.body.sizeY = data.sizeY
	
	vp.minValue = data.minValue or 0
	vp.maxValue = data.maxValue or 100
	vp.increment = data.increment or 1
	vp.value = data.value or vp.minValue	-- current value
	
	-- text...
	vp.maxFontSize = data.fSize or 40
	vp.maxTextWidth = vp.body.sizeX * 0.9
	vp.maxTextHeight = vp.body.sizeY * 0.9
	vp.alignMode = data.alignMode or "center"
	-- optional caption
	--> caption: value
	vp.caption = data.caption
	
	if data.visible == nil then
		vp.visible = true
	else
		vp.visible = data.visible
	end
	
	vp.onValueUpdate = data.onValueUpdate	-- callback can be nil
	
	----------------------------------------------------------------------------------------------------
	
	-- basically just returns the body (for mouse interaction)
	function vp:getShape()
		return self.body
	end
	
	----------------------------------------------------------------------------------------------------
	
	function vp:getValue()
		return self.value
	end
	
	----------------------------------------------------------------------------------------------------
	
	function vp:onWheelMoved( dx, dy )
		
		-- check if the mouse was inside the element during the wheel move event
		local pos = Mouse:getPosition()
		local body = self.body
		if not Tools.isPointInsideRect( pos, body ) then
			return false
		end
		
		if dy < 0 then
			-- moved down
			self.value = self.value - self.increment
		elseif dy > 0 then
			-- moved up
			self.value = self.value + self.increment
		end
		
		self.value = Tools.clamp( self.value, self.minValue, self.maxValue )
		self:updateText()
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- manage mouseover
	function vp:onMouseHoverStart()
		self.hover = true
	end
	
	function vp:onMouseHoverEnd()
		self.hover = false
	end
	
	----------------------------------------------------------------------------------------------------
	
	function vp:updateText()
		local _text = self.value
		if self.caption then
			_text = string.format( "%s: %d", self.caption, self.value )
		end
		self.text = self.text or lg.newText( lg.newFont( self.maxFontSize ), "" )
		self.text:setf( _text or "", self.maxTextWidth, self.alignMode or "center" )
		self:autoAdjustFontSize( self.text, self.maxTextWidth, self.maxTextHeight )
	end
	
	
	function vp:autoAdjustFontSize( text, sizeX, sizeY )
		local fSize = self.maxFontSize or 40
		
		while text:getWidth() >= sizeX
		or text:getHeight() >= sizeY do
			fSize = fSize - 2
			if fSize < 8 then
				print( "ValuePicker warning: UI element too small for text!", self.id )
				break
			end
			text:setFont( lg.newFont( fSize ) )
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function vp:draw()
		
		if not self.visible then
			return
		end
		
		if self.parentId and not (GUImgr:isEnabled(self.parentId) and GUImgr:isVisible(self.parentId)) then
			return
		end
		
		local body = self.body
		local top = body.y - 0.5*body.sizeY
		local left = body.x - 0.5*body.sizeX
		if self.hover then
			lg.setColor( 0.15, 0.9, 0, 1 )
		end
		lg.rectangle( "line", left, top, body.sizeX, body.sizeY, 2, 2 )
		lg.setColor( COLORS.white )
		
		-- use the defined text width, not the current width of the text object for the horizontal axis
		-- but use the current text height for the y coordinate
		lg.draw( self.text, body.x, body.y, 0, 1, 1, self.maxTextWidth * 0.5, self.text:getHeight() * 0.5 )
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	vp:updateText()
	
	Mouse:addListener( "wheelMoved", vp )	-- make sure to have onWheelMoved function
	GUImgr:addObject( vp )
	UIRenderer:add( vp )
	
	return vp
end

return ValuePicker
