
-- A* pathfinding algorithm

-- DEBUG
-- how often does caching the paths actually help?
local pathsCalculated = 0
local pathReadsFromCache = 0

local Astar = { maps = {}, paths = {}, }

-- add or update map data for a specific map
-- e.g. a mapData labeled "walking" marks water as not traversable, for "flying" it is traversable
function Astar.setMap( mapLabel, mapData )
	if Astar.maps[mapLabel] then
		-- print( "A* update map: ", mapLabel )
	else
		print( "A* new map: ", mapLabel )
	end
	
	Astar.maps[mapLabel] = mapData
end


-- convert start and target coordinates into a string used as index for cached paths
-- note: do not forget the map label!
function Astar.buildPathIndex( start_x, start_y, target_x, target_y, mapLabel )
	-- returns "mapLabel_x_y_x_y"
	return string.format( "%s_%d_%d_%d_%d", mapLabel, start_x, start_y, target_x, target_y )
end


-- remember a path that was just calculated
function Astar.storePath( start_x, start_y, target_x, target_y, mapLabel, path )
	Astar.paths[mapLabel] = Astar.paths[mapLabel] or {}
	local index = Astar.buildPathIndex( start_x, start_y, target_x, target_y, mapLabel )
	Astar.paths[mapLabel][index] = path
end


-- clear all cached paths for a given map label (to be called when the blocking changes)
function Astar.clearPaths( mapLabel )
	Astar.paths[mapLabel] = {}
end


-- try to find a path from start to taget tile; lookup in "cache" first
-- if this exact path is already stored in temporary cache, do not need to run the algorithm again!
function Astar.getPath( start_x, start_y, target_x, target_y, mapLabel )
	local index = Astar.buildPathIndex( start_x, start_y, target_x, target_y, mapLabel )
	if Astar.paths[mapLabel] and Astar.paths[mapLabel][index] ~= nil then
		local pathOrFalse = Astar.paths[mapLabel][index]
		pathReadsFromCache = pathReadsFromCache + 1	-- DEBUG
		-- print( "A* - pathsCalculated =", pathsCalculated )
		-- print( "A* - pathReadsFromCache =", pathReadsFromCache )
		
		if pathOrFalse then
			-- return a copy! the path gets shorter when the unit follows it
			return Tools.copyTable( Astar.paths[mapLabel][index] )
		else
			-- value "false" has been stored = there is no path between the two points!
			return false
		end
	end
	
	pathsCalculated = pathsCalculated + 1	-- DEBUG
	-- print( "A* - pathsCalculated =", pathsCalculated )
	-- print( "A* - pathReadsFromCache =", pathReadsFromCache )
	local pathOrFalse = Astar.run( { x = start_x, y = start_y }, { x = target_x, y = target_y }, mapLabel )
	Astar.storePath( start_x, start_y, target_x, target_y, mapLabel, pathOrFalse )
	if pathOrFalse then
		-- return a copy! the path gets shorter when the unit follows it
		return Tools.copyTable( pathOrFalse )
	else
		return false
	end
end


-- actually run the algorithm
function Astar.run( start, goal, mapLabel )
	
	-- check that map is available
	if not Astar.maps[mapLabel] then
		print("A* error: invalid or no map!")
		return false
	end
	
	Astar.map = Astar.maps[mapLabel]
	
	Astar.open		= {}
	Astar.closed	= {}
	Astar.goal		= goal
	
	-- select a heuristic for the distance estimation
	-- note: using euclidean heuristic entities tend to go diagonal (or zig-zag depending on movement code)
	-- while for manhattan heuristic entities walk straight horizontal / vertical
	Astar.heuristic = Astar.hManhattan
	-- alternative:
	-- Astar.heuristic = Astar.hEuclidean
	
	-- check that start and goal are inside the map and not blocked itself
	if (start.x < 1) or (start.x > Astar.map.width) or (start.y < 1) or (start.y > Astar.map.height)
	or (goal.x < 1) or (goal.x > Astar.map.width) or (goal.y < 1) or (goal.y > Astar.map.height) then
		-- print("a* error: position check failed!")
		return false
	end
	
	-- check if the starting and target position is traversable
	if not Astar.map.blockingData
	or not Astar.map.blockingData[start.y]
	or not Astar.map.blockingData[start.y][start.x]
	or not Astar.map.blockingData[goal.y]
	or not Astar.map.blockingData[goal.y][goal.x] then
		-- print("a* warning: initial blocking check failed!")
		return false
	end
	
	-- open set / "priority queue"
	table.insert(Astar.open, start)
	
	start.g = 0
	start.f = start.g + Astar.heuristic(start, goal)
	--> f = total cost of current node; g = distance from current to start node
	
	while #Astar.open > 0 do
	
		--print("a* open nodes: " .. #Astar.open)
		--> searches for the neighbour node that has the best chance to provide improvement towards the goal
		local current = Astar.getElementWithLowestF()
		if not current then
			print("A* error: cannot find element with lowest F")
			return false
		end
		--> we found our target!
		if Astar.isNodeEqual(current, goal) then
			return Astar.constructPath(current)
		end
		-- otherwise...
		Astar.removeNode(Astar.open, current)
		table.insert(Astar.closed, current)
		
		for _, neighbour in pairs(Astar.getNeighbours(current)) do
			--print("neighbour parent: [" .. neighbour.parent.x .. ":" .. neighbour.parent.y .. "]")
			
			local cost = current.g + 1-- + neighbour.penalty	-- movement cost from current to neighbour
			
			if Astar.isNodeInList(neighbour, Astar.open)
			and cost < neighbour.g then
				Astar.removeNode(Astar.open, neighbour)
			end
			
			if Astar.isNodeInList(neighbour, Astar.closed)
			and cost < neighbour.g then
				Astar.removeNode(Astar.closed, neighbour)
			end
			
			if not Astar.isNodeInList(neighbour, Astar.open)
			and not Astar.isNodeInList(neighbour, Astar.closed) then
				neighbour.g = cost
				table.insert(Astar.open, neighbour)
				neighbour.f = neighbour.g + Astar.heuristic(neighbour, goal)
				neighbour.parent = current
			end
			
		end
		
	end
	
	return false

end


function Astar.isNodeInList(node, list)
	for _, n in pairs(list) do
		if (n.x == node.x) and (n.y == node.y) then
			return true
		end
	end
	return false
end


function Astar.hManhattan(start, goal)
	-- manhattan
	local d = math.abs(goal.x - start.x) + math.abs(goal.y - start.y)
	--print("A*:", start.x, start.y, goal.x, goal.y, d)
	return d
end

function Astar.hEuclidean(start, goal)
	-- "euclidean" no square root needed! not for measurement, just comparison
	local d = (goal.x - start.x) * (goal.x - start.x) + (goal.y - start.y) * (goal.y - start.y)
	--print("A*:", start.x, start.y, goal.x, goal.y, d)
	return d
end


function Astar.getElementWithLowestF()
	--> for initial max value use AT LEAST the max distance depending on the HEURISTIC!
	--local bestF, node = Astar.map.width + Astar.map.height, nil
	local bestF, node = Astar.map.width * Astar.map.width + Astar.map.height * Astar.map.height, nil
	for i = 1, #Astar.open do
		if Astar.open[i].f < bestF then
			bestF = Astar.open[i].f
			node = Astar.open[i]
		end
	end
	
	return node
end


function Astar.isNodeEqual(node1, node2)
	return (node1.x == node2.x) and (node1.y == node2.y)
end


function Astar.removeNode(_table, node)
	for i = #_table, 1, -1 do
		if Astar.isNodeEqual(_table[i], node) then
			--print("a* remove node from open table: " .. node.x .. " " .. node.y)
			table.remove(_table, i)
			break
		end
	end
end

--[[
map.blockingData should be a 2D data structure:
[0,0][1,0][2,0]...
[0,1][1,1][2,1]
[0,2][1,2][2,2]
...            [x,y]
where 0 = not blocking and > 0 can be blocking depending on the movement type
--]]

function Astar.getNeighbours(node)
	local neighbours = {}
	local width, height = Astar.map.width, Astar.map.height
	--local nodeIndex = (node.x + node.y * width) + 1	-- old
	
	-- top neighbour
	if node.y > 1 then
		local neighbourNode = {x = node.x, y = node.y - 1}
		if Astar.map.blockingData[neighbourNode.y][neighbourNode.x] then
		--or Astar.isNodeEqual(neighbourNode, Astar.goal) then
			neighbourNode.g = node.g + 1
			-- neighbourNode.penalty = Astar.map.blockingData[neighbourNode.y][neighbourNode.x].penalty
			table.insert(neighbours, neighbourNode)
		end
	end
	
	-- bottom neighbour
	if node.y < height then
		local neighbourNode = {x = node.x, y = node.y + 1}
		if Astar.map.blockingData[neighbourNode.y][neighbourNode.x] then
		--or Astar.isNodeEqual(neighbourNode, Astar.goal) then
			neighbourNode.g = node.g + 1
			-- neighbourNode.penalty = Astar.map.blockingData[neighbourNode.y][neighbourNode.x].penalty
			table.insert(neighbours, neighbourNode)
		end
	end
	
	-- left neighbour
	if node.x > 1 then
		local neighbourNode = {x = node.x - 1, y = node.y}
		if Astar.map.blockingData[neighbourNode.y][neighbourNode.x] then
		--or Astar.isNodeEqual(neighbourNode, Astar.goal) then
			neighbourNode.g = node.g + 1
			-- neighbourNode.penalty = Astar.map.blockingData[neighbourNode.y][neighbourNode.x].penalty
			table.insert(neighbours, neighbourNode)
		end
	end
	
	-- right neighbour
	if node.x < width then
		local neighbourNode = {x = node.x + 1, y = node.y}
		if Astar.map.blockingData[neighbourNode.y][neighbourNode.x] then
		--or Astar.isNodeEqual(neighbourNode, Astar.goal) then
			neighbourNode.g = node.g + 1
			-- neighbourNode.penalty = Astar.map.blockingData[neighbourNode.y][neighbourNode.x].penalty
			table.insert(neighbours, neighbourNode)
		end
	end
	
	return neighbours
end


function Astar.constructPath(node)
	local path = {node}
	while node.parent ~= nil do
		--print("path node: " .. node.x .. " : " .. node.y)
		node = node.parent
		table.insert(path, node)
	end
	return path
end


return Astar
