
local tools = {}

-- tools.transforms = {}
-- tools.transforms[45] = love.math.newTransform(0, 0, math.pi*7/4, 1, 1, 0, 0)	-- NE
-- tools.transforms[135] = love.math.newTransform(0, 0, math.pi/4, 1, 1, 0, 0)		-- NW
-- tools.transforms[225] = love.math.newTransform(0, 0, math.pi*7/4, -1, -1, 0, 0)	-- SW
-- tools.transforms[315] = love.math.newTransform(0, 0, math.pi/4, -1, -1, 0, 0)	-- SE

----------------------------------------------------------------------------------------------------

-- from https://stackoverflow.com/questions/640642/how-do-you-copy-a-lua-table-by-value
function tools.copyTable(obj, seen)
	if type(obj) ~= 'table' then return obj end
	if seen and seen[obj] then return seen[obj] end
	local s = seen or {}
	local res = setmetatable({}, getmetatable(obj))
	s[obj] = res
	for k, v in pairs(obj) do res[tools.copyTable(k, s)] = tools.copyTable(v, s) end
	return res
end

----------------------------------------------------------------------------------------------------

function tools.round(value)
	return math.floor(value + 0.5)
end

function tools.roundTo ( number, multiple )
  return math.floor ( ( number / multiple ) + 0.5 ) * multiple
end

-- round a given number with digits decimal places
function tools.roundEx(number, digits)
	if not digits then
		return tools.round(number)
	end
	
	assert(digits > 0, "tools: roundEx invalid digits count")
	
	local shift = math.pow(10, digits)
	local val = number * shift
	return tools.round(val) / shift
end

----------------------------------------------------------------------------------------------------

function tools.getSign(number)
	if number < 0 then
		return -1
	else
		return 1
	end
end

function tools.clamp( value, min, max )
	return tools.max(tools.min(value, max), min)
end

function tools.min(a, b)
	if a > b then
		return b
	else
		return a
	end
end

function tools.max(a, b)
	if a < b then
		return b
	else
		return a
	end
end

----------------------------------------------------------------------------------------------------

-- position table version
function tools.getDistance(pos1, pos2)
	return math.sqrt((pos2.x - pos1.x) * (pos2.x - pos1.x) + (pos2.y - pos1.y) * (pos2.y - pos1.y))
end

-- 4x number version
function tools.getDistance2( x1, y1, x2, y2 )
	return math.sqrt((x2 - x1) * (x2 - x1) + (y2 - y1) * (y2 - y1))
end

-- a little less handy but faster as it does not require a root
function tools.getDistanceSquared(pos1, pos2)
	return (pos2.x - pos1.x) * (pos2.x - pos1.x) + (pos2.y - pos1.y) * (pos2.y - pos1.y)
end

function tools.getManhattanDistance( pos1, pos2 )
	return math.abs(pos1.x-pos2.x) + math.abs(pos1.y-pos2.y)
end

-- 4x number version
function tools.getManhattanDistance2( x1, y1, x2, y2 )
	return math.abs(x1-x2) + math.abs(y1-y2)
end

----------------------------------------------------------------------------------------------------

function tools.vectorGetLength(vector)
	return math.sqrt(vector.x * vector.x + vector.y * vector.y)
end

function tools.vectorSetLength(vector, length)
	local newLen = length or 1
	local currLen = tools.vectorGetLength(vector)
	local vx, vy
	vx = vector.x * newLen/currLen
	vy = vector.y * newLen/currLen
	return {x = vx, y = vy}
end

function tools.vectorTrim(vector, maximumLength)

	local maxLen = maximumLength or 1
	local currLen = tools.vectorGetLength(vector)
	
	if currLen <= maxLen then
		-- vector already short enough
		return vector
	else
		-- vector norm too big
		local vx, vy
		vx = vector.x * maxLen/currLen
		vy = vector.y * maxLen/currLen
		return {x = vx, y = vy}
	end
	
end

function tools.vectorSub(vec_1, vec_2)
	return {x = vec_2.x - vec_1.x, y = vec_2.y - vec_1.y}
end


function tools.vectorAdd(vec_1, vec_2)
	return {x = vec_2.x + vec_1.x, y = vec_2.y + vec_1.y}
end

-- vector pointing towards x -> angle = 0
-- moving clockwise the angle increases until pi (vector facing negative x)
-- facing negative y -> angle = -0.5*pi
function tools.getAngleFromVector(vec2)
	return math.atan2(vec2.y, vec2.x)	-- + math.pi * 0.5
end

function tools.getVectorFromAngle(_rotation)
	return {x = math.cos(_rotation), y = math.sin(_rotation)}
end

----------------------------------------------------------------------------------------------------

-- TODO: are two shapes (rect, circle, point?) colliding?
function tools.areBodiesColliding(body1, body2)
	
	if body1.shape == "rect" then
		
		if body2.shape == "rect" then
			return tools.areRectsColliding(body1, body2)
			
		elseif body2.shape == "circle" then
			return tools.areCircleAndRectColliding(body2, body1)
			
		elseif body2.shape == "point" then
			return tools.isPointInsideRect(body2, body1)
			
		else
			print("ERROR in tools: invalid shape (2)", body2.shape)
			return false
		end
		
	elseif body1.shape == "circle" then
		
		if body2.shape == "rect" then
			return tools.areCircleAndRectColliding(body1, body2)
			
		elseif body2.shape == "circle" then
			return tools.areCirclesColliding(body1, body2)
			
		elseif body2.shape == "point" then
			return tools.isPointInsideCircle(body2, body1)
			
		else
			print("ERROR in tools: invalid shape (2)", body2.shape)
			return false
		end
		
	elseif body1.shape == "point" then
		
		if body2.shape == "rect" then
			return tools.isPointInsideRect(body1, body2)
			
		elseif body2.shape == "circle" then
			return tools.isPointInsideCircle(body1, body2)
			
		elseif body2.shape == "point" then
			
			return body1.x == body2.x and body2.y == body2.y
			
		else
			print("ERROR in tools: invalid shape (2)", body2.shape)
			return false
		end
		
	else
		print("ERROR in tools: invalid shape (1)", body1.shape)
		return false
	end
	
end

function tools.isPointInsideShape(point, shape)
	if not shape or not shape.shape then
		print("tools: invalid shape!", shape.shape)
		for k,v in pairs( shape ) do
			print( k,v )
		end
		return false
	end
	if shape.shape == "rect" then
		return tools.isPointInsideRect(point, shape)
	elseif shape.shape == "circle" then
		return tools.isPointInsideCircle(point, shape)
	else
		print("tools: unknown body type!", shape.shape)
		return false
	end
	
end

-- rectangles without rotation!
function tools.isPointInsideRect(point, rect)
	
	local x, y = rect.x, rect.y
	
	if point.x > x - rect.sizeX/2
	and point.x < x + rect.sizeX/2
	and point.y > y - rect.sizeY/2
	and point.y < y + rect.sizeY/2 then
		return true
	end
	
	return false
	
end

-- note: here only non rotated rectangles... and all rects have their origin at the CENTER
function tools.areRectsColliding(rect_1, rect_2)
	-- use the top left corner internally
	local x1, y1 = rect_1.x - rect_1.sizeX/2, rect_1.y - rect_1.sizeY/2
	local x2, y2 = rect_2.x - rect_2.sizeX/2, rect_2.y - rect_2.sizeY/2
	
	if x1 + rect_1.sizeX > x2 and
	   x1 < x2 + rect_2.sizeX and
	   y1 + rect_1.sizeY > y2 and
	   y1 < y2 + rect_2.sizeY then
	   	return true
	end

	return false
end


function tools.isPointInsideCircle(point, circle)
	return tools.getDistanceSquared({x = circle.x, y = circle.y}, {x = point.x, y = point.y}) <= circle.radius * circle.radius
end


function tools.areCirclesColliding(circle1, circle2)
	return (circle1.radius + circle2.radius) * (circle1.radius + circle2.radius) >= tools.getDistanceSquared({x = circle1.x, y = circle1.y}, {x = circle2.x, y = circle2.y})
end


function tools.getRotatedPoint(point, angle, center)
	local oldX, oldY = point.x - center.x, point.y - center.y
	local newX = oldX * math.cos(angle) - oldY * math.sin(angle) + center.x
	local newY = oldY * math.cos(angle) + oldX * math.sin(angle) + center.y
	return newX, newY
end

-- rectangles without rotation!
function tools.areCircleAndRectColliding(circle, rect)
	local rx, ry, dx, dy
	rx = tools.clamp(circle.x, rect.x - rect.sizeX/2, rect.x + rect.sizeX/2)
	ry = tools.clamp(circle.y, rect.y - rect.sizeY/2, rect.y + rect.sizeY/2)
	dx = circle.x - rx
	dy = circle.y - ry
	return (dx * dx + dy * dy) < (circle.radius * circle.radius), rx, ry
end

----------------------------------------------------------------------------------------------------

-- returns true if a given value can be found inside the table, false otherwise
function tools.isValueInTable(_value, _table)
	for index, value in ipairs(_table) do
		if value == _value then
			return true
		end
	end

	return false
end

----------------------------------------------------------------------------------------------------

-- get RGB relative to given health: turn from green to yellow to red (form 100% to 0% health)
function tools.getRGBbyHealth(health)
	local R, G, B
	--[[
	if health == 100 then
		R, G, B = 0, 255, 0
	elseif health > 50 then
		R = 255 - health * 2.55
		G = 255
		B = 0
	else
		R = 255
		G = health * 5.1
		B = 0
	end
	return R/255, G/255, B/255
	--]]
	
	if health > 0.5 then
		R = 1 - health + 0.5
		G = 1
		B = 0
	else
		R = 1
		G = health * 2
		B = 0
	end
	return R, G, B
end

----------------------------------------------------------------------------------------------------

-- note: two separators next to each other = no table entry (careful when passing "empty" parameters)
function tools.string_split(strng, sep)
	-- see: https://stackoverflow.com/questions/1426954/split-string-in-lua
	local t = {}
	local i = 1
	for str in string.gmatch(strng, "([^"..sep.."]+)") do
		t[i] = str
		i = i + 1
	end
	return t
end

----------------------------------------------------------------------------------------------------

return tools
